#ifndef CLIENT_H
#define CLIENT_H

#define AttackTypeBeam		1
#define AttackTypeBall		2
#define AttackTypeKiblast	3
#define AttackTypeOther		4
#define AttackTypeUnknown	5

#define MaxOldPositions			5
#define MaxOldPositionLifeTime	15.0
#define MinNewPositionDistance	200.0

//#define usestats

#ifndef __linux__

#define PL_OFFSET			460
#define ACTUAL_PL_OFFSET	461
#define TURBO_OFFSET		196
#define MELEE_OFFSET		301

#else

#define PL_OFFSET			425
#define ACTUAL_PL_OFFSET	426
#define TURBO_OFFSET		196
#define MELEE_OFFSET		290

#endif

// Combo charge buttons
enum EComboChargeButton
{
	COMBO_BUTTON_PUNCH_LEFT = 2,
	COMBO_BUTTON_PUNCH_RIGHT = 6,
	COMBO_BUTTON_PUNCH_UP = 0,
	COMBO_BUTTON_PUNCH_DOWN = 4,
	COMBO_BUTTON_PUNCH_LEFT_UP = 1,
	COMBO_BUTTON_PUNCH_RIGHT_UP = 7,
	COMBO_BUTTON_PUNCH_LEFT_DOWN = 3,
	COMBO_BUTTON_PUNCH_RIGHT_DOWN = 5,

	COMBO_BUTTON_KICK_LEFT = 10,
	COMBO_BUTTON_KICK_RIGHT = 14,
	COMBO_BUTTON_KICK_UP = 8,
	COMBO_BUTTON_KICK_DOWN = 12,
	COMBO_BUTTON_KICK_LEFT_UP = 9,
	COMBO_BUTTON_KICK_RIGHT_UP = 15,
	COMBO_BUTTON_KICK_LEFT_DOWN = 11,
	COMBO_BUTTON_KICK_RIGHT_DOWN = 13,

	COMBO_BUTTON_RANDOM = 16, // Random for defender, idle punch for attacker

	COMBO_BUTTON_INVALID = -1
};

extern CLog clog;

typedef struct
{
	Vector Place;
	float CreateTime;
} OldPosition_t;

#ifdef useclassstuff

enum ESwitch;
class CClientManager;

class CClient
{
protected:
	// Status vars
	short m_sKi;
	long m_lPL;
    bool m_bPowerUps;
	byte m_bChargeValue;
	float m_fAscendTime;
	short m_sPSVal;
	float m_fBlindedTime;
	short m_sMaxHealth;

	bool m_bIsAutoPunching;
	bool m_bGetAutoPunched;
	bool m_bIsInMeleeMode;
	bool m_bAscends;
	bool m_bPowerstruggles;
	bool m_bChargesAttack;
	bool m_bFlysAfterHit;
	bool m_bChargesWallJump;
	bool m_bWallJumps;
	bool m_bSwoops;
	bool m_bMeleeStruggles;
	bool m_bIsStuckAgainstWall;
	bool m_bLaysOnGround;
	bool m_bControlsAttack;
	bool m_bBeamJumps;
	bool m_bIsBlinded;
	bool m_bBlocks;
	bool m_bBlockStruggles;

	// for super bots
	float m_fFrozenTime;

	float m_fLastSoundTime; // Time that this player made a sound
	char m_szSoundName[64]; // The name of the sound

	// Locations where this player moved in the past(so that bots can hunt him down)
	float m_fLastOldPosCheck;
	OldPosition_t m_OldPositions[MaxOldPositions];

	bool m_bIsIngame; // Is this player connected to the server?

	friend CClientManager;

public:
	// Public var(s)
	edict_t *pEdict;

	CClient(edict_t *pPlayer); // Construction
	virtual ~CClient(void) {}; // Destruction

	// Member functions
	virtual void Think(void);
	virtual bool IsBot(void){ return false; };

	bool IsAutoPunching(void)		{ return m_bIsAutoPunching; };
	bool GetAutoPunched(void)		{ return m_bGetAutoPunched; };
	bool IsInMeleeMode(void)		{ return m_bIsInMeleeMode; };
	bool Ascends(void)				{ return m_bAscends; };
	bool Powerstruggles(void)		{ return m_bPowerstruggles; };
	bool ChargesAttack(void)		{ return m_bChargesAttack; };
	bool FlysAfterHit(void)			{ return m_bFlysAfterHit; };
	bool ChargesWallJump(void)		{ return m_bChargesWallJump; };
	bool WallJumps(void)			{ return m_bWallJumps; };
	bool Swoops(void)				{ return m_bSwoops; };
	bool MeleeStruggles(void)		{ return m_bMeleeStruggles; };
	bool IsStuckAgainstWall(void)	{ return m_bIsStuckAgainstWall; };
	bool LaysOnGround(void)			{ return m_bLaysOnGround; };
	bool ControlsAttack(void)		{ return m_bControlsAttack; };
	bool BeamJumps(void)			{ return m_bBeamJumps; };
	bool IsBlinded(void)			{ return m_bIsBlinded; };
	bool Blocks(void)				{ return m_bBlocks; };
	bool BlockStruggles(void)		{ return m_bBlockStruggles; };
	bool PowerUps(void)				{ return m_bPowerUps; };
	bool MadeASound(void)			{ return (m_fLastSoundTime > (gpGlobals->time - 1.0)); };
	bool HasTurboOn(void)			{ return (*((int *)pEdict->pvPrivateData + TURBO_OFFSET) == 1); };
	bool IsAlive(void);

	short GetHealth(void)			{ return static_cast<short>(pEdict->v.health); };
	short GetMaxHealth(void)		{ return m_sMaxHealth; };
	long GetPl(void)				{ return m_lPL; };
	short GetKi(void)				{ return static_cast<short>(pEdict->v.fuser4); };
	byte GetTeam(void)				{ return static_cast<byte>(pEdict->v.team); };
	byte GetClass(void)				{ return static_cast<byte>(pEdict->v.playerclass); };
	char *GetSoundName()			{ return m_szSoundName; };
	float GetVelocity(void)			{ return pEdict->v.velocity.Length(); };
	float GetMaxSpeed(void)			{ return pEdict->v.maxspeed; };
	Vector GetGunPosition(void)		{ return (pEdict->v.origin + pEdict->v.view_ofs); };
	Vector GetOrigin(void)			{ return pEdict->v.origin; };
	float GetDistance(Vector v_end)	{ return (GetOrigin() - v_end).Length(); };
	void GetModel(char *Output);

	void PlayerClassify(char *ClassName);

	void SetPl(long NewPl)			{ *((long *)pEdict->pvPrivateData + PL_OFFSET) = NewPl; };
	void SetRealPl(long NewPl)		{ *((long *)pEdict->pvPrivateData + ACTUAL_PL_OFFSET) = NewPl; };
	void SetKi(short NewKi)			{ pEdict->v.fuser4 = static_cast<float>(NewKi); };
	void SetHealth(short NewHealth)	{ pEdict->v.health = static_cast<float>(NewHealth); };
	void ToggleTurbo(ESwitch OnOff);
	virtual void SetTeam(short Team);

	bool FInViewCone(Vector *pOrigin);
	bool FVisible(const Vector &vecOrigin);

	void ClearOldPositions(void);
	void UpdateOldPositions(void);
	void UpdateStatus(void);
	void UpdateFrozenStatus(void); // sbot stuff

	// Client message hook functions
	virtual void ClientMsgChargeValue(void *pValue)		{ m_bChargeValue = *(byte *)pValue; };
	virtual void ClientMsgPowerstruggle(void *pValue)	{ m_sPSVal = *(short *)pValue; };
	virtual void ClientMsgScreenFade(void *pValue);
	virtual void ClientMsgVGUI(void *pValue);
	virtual void ClientMsgMaxHealth(void *pValue)		{ m_sMaxHealth = *(short *)pValue; };

	void ActivatePlayer(void) { m_bIsIngame = true; };
};

// Thrown when a client needs to be removed from the client list
class CBadEdictException
{
public:
	CClient *pClient;
	CBadEdictException(CClient *p) : pClient(p) {};
};

#endif

#ifdef usestats
struct SWeaponStatData
{
	char szWeaponName[32];
	unsigned long DamageCaused; // How much damage this weapon caused last minute.
	unsigned long DamageTaken; // How much damage this weapon has caused to this player last minute.
	unsigned long KillsCaused; // How much kills this weapon caused last minute.
	unsigned long KilledBy; // How much kills this weapon has caused to this player last minute.
	short WeaponScore; // The score of this weapon(0 - 100)
	short WeaponDanger; // The danger value of this weapon(0 - 100)

	SWeaponStatData(const char *str) : DamageCaused(0), DamageTaken(0), KillsCaused(0), KilledBy(0),
									   WeaponScore(0), WeaponDanger(0) { strcpy(szWeaponName, str); };
	void ClearTempData(void) { DamageCaused = DamageTaken = KillsCaused = KilledBy = 0; };
	void AddDamageCaused(unsigned long ul)  { DamageCaused += ul; };
	void AddDamageTaken(unsigned long ul)	{ DamageTaken += ul; };
	void AddKillsCaused(void)				{ KillsCaused++; };
	void AddKilledBy(void)					{ KilledBy++; };
	void AddWeaponScore(void);
	void AddWeaponDanger(void);

	// Debug stuff
	void LogData(void);
};

class CWeaponStatistics
{
	TLinkedList<SWeaponStatData*> WeaponDataList;

public:
	~CWeaponStatistics(void) { while (!WeaponDataList.Empty()) delete WeaponDataList.Pop(); };
	void MakeWeaponStats(void);
	void ResetVars(void);
	void ClearAll(void) { while (!WeaponDataList.Empty()) delete WeaponDataList.Pop(); };
	void AddWeaponDamageCaused(const char *szAttackName, int Damage);
	void AddWeaponDamageHit(const char *szAttackName, int Damage);
	void AddWeaponKillsCaused(const char *szAttackName);
	void AddWeaponKilledBy(const char *szAttackName);
	int GetAttackThatMostDamageCaused(void);
	int GetAttackThatMostHit(void);

	// List operation functions
	bool FindAttackByName(SWeaponStatData *pNode, const char *szWeaponName) { return (!stricmp(pNode->szWeaponName, szWeaponName)); };

	// Debug stuff
	void LogWeaponStats(void);
};
#endif

typedef struct
{
   int  iId;     // weapon ID
   int  iClip;   // amount of ammo in the clip
   int  iAmmo1;  // amount of ammo in primary reserve
   int  iAmmo2;  // amount of ammo in secondary reserve
} SCurrentWeapon;

typedef struct 
{
	edict_t *pEdict;
	SCurrentWeapon CurrentWeapon;
	int Health;
	int ki;
	long int PL;
    bool PowerUp;
	int ChargeBeamVal;
	float AscendTime;
	short PSVal;
	float BlindedTime;

	// for sbots
	float FrozenTime;

	float LastSoundTime;
	char SoundName[64];

	OldPosition_t OldPositions[MaxOldPositions];
	float LastOldPosCheck;

	bool IsAutoPunching;
	bool GetAutoPunched;
	bool IsInMeleeMode;
	bool IsStunnedByMelee;
	bool HasThrowLanded;
	bool Ascends;
	bool Powerstruggles;
	bool ChargesAttack;
	bool FlysAfterHit;
	bool ChargesWallJump;
	bool WallJumps;
	bool Swoops;
	bool MeleeStruggles;
	bool DoesComboStruggle;
	bool IsStuckAgainstWall;
	bool LaysOnGround;
	bool ControlsAttack;
	bool BeamJumps;
	bool IsBlinded;
	bool Blocks;
	bool BlockStruggles;
	bool HasDB;
	bool GetSwung;

	// Advanced melee stuff
	edict_t *pAttachedPlayer; // Player this client is attached to
	TLinkedList<EComboChargeButton> DefendButtonsList; // List of buttons the player should defend to
	bool Attacker; // Is this player the attacker(charging the combo)?
	bool Defender; // Is this player the defender(defend the charged buttons)
	short ButtonsRandom; // DEBUG - how many buttons where random in the previous defender session
	short ButtonsMissed; //	DEBUG - how many buttons missed this player
	short DefendButtonCount; // How much buttons had this player to defend
	float MaxComboChargeTime; // How long the attacker can charge a combo
	float WeakButtonChargeTime; // Max time for a weak button to charge(in sec)
	float MediumButtonChargeTime; // Max time for a medium button to charge(in sec)
	float StrongButtonChargeTime; // Max time for a strong button to charge(in sec)
	byte PrepunchHits; // Amount of prepunches this player has caused

#ifdef usestats
	// Statistics
	unsigned long DamageCausedLastMinut;
	unsigned long Deaths;
	unsigned long Kills;
	float flRefreshStatTime;
	CWeaponStatistics WeaponStatistics;
#endif

} ClientInfo_t;

/* 
struct AttackList_t
{
	edict_t *pAttack;
	short type;
	float CreateTime;

	// for sbots
	bool Attacked;
	float AttackedTime;
	Vector OldVelocity;

	AttackList_t *next;
};
*/
bool PlayerIsAutoPunching(edict_t *pEdict);
bool PlayerGetAutoPunched(edict_t *pEdict);
bool PlayerIsInMeleeMode(edict_t *pEdict);
bool PlayerAscends(edict_t *pEdict);
bool PlayerPowerstruggles(edict_t *pEdict);
bool PlayerChargesAttack(edict_t *pEdict);
bool PlayerFlysAfterHit(edict_t *pEdict);
bool PlayerChargesWallJump(edict_t *pEdict);
bool PlayerWallJumps(edict_t *pEdict);
bool PlayerSwoops(edict_t *pEdict);
bool PlayerMeleeStruggles(edict_t *pEdict);
bool PlayerIsStuckAgainstWall(edict_t *pEdict);
bool PlayerLaysOnGround(edict_t *pEdict);
bool PlayerControlsAttack(edict_t *pEdict);
bool PlayerBeamJumps(edict_t *pEdict);
bool PlayerIsBlinded(edict_t *pEdict);
bool PlayerBlocks(edict_t *pEdict);
bool PlayerBlockStruggles(edict_t *pEdict);
bool PlayerHasADB(edict_t *pEdict);

long GetPlayerPl(edict_t *pEdict);
long GetPlayerActualPl(edict_t *pEdict);

void ClientThink(ClientInfo_t *pClient);

Vector PredictEntityPos(edict_t *pEntity, float Time, bool UseTraceLine = true);
/*
void AddAttackToList(edict_t *pent);
void RemoveAttackFromList(AttackList_t *AL);

edict_t *GetNearestHostileAttack(edict_t *pPlayer, float CheckRadius, bool CheckViewCone);
short GetSurroundingHostileAttacks(edict_t *pPlayer, float CheckRadius, bool CheckViewCone);

void PrintAttackList();

void DeleteAttackList();
void RemoveDeletedAttacks();

void AttackListThink();

edict_t *GetAttackFromPlayer(edict_t *pPlayer);
edict_t *SearchFriezaDisc(edict_t *pPlayer);
*/
void PlayerClient_ESF_PL(void *p, int player_ind);
void PlayerClient_ESF_Ki(void *p, int player_ind);
void PlayerClient_ESF_StartPowerUp(void *p, int player_ind);
void PlayerClient_ESF_StopPowerUp(void *p, int player_ind);
void PlayerClient_ESF_Ki(void *p, int player_ind);
void PlayerClient_ESF_ChargeBeam(void *p, int bot_index);
void PlayerClient_ESF_TransformFX(void *p, int player_ind);
void PlayerClient_ESF_StopTransFX(void *p, int player_ind);
void PlayerClient_ESF_Powerstruggle(void *p, int player_ind);
void PlayerClient_ESF_ScreenFade(void *p, int player_ind);
void PlayerClient_ESF_VGUI(void *p, int player_index);
void PlayerClient_ESF_MeleeHud(void *p, int player_index);
void PlayerClient_ESF_MeleeHudTxt(void *p, int player_index);
void PlayerClient_ESF_ManipModel(void *p, int player_index);
void PlayerClient_ESF_SayTextMsg(void *p, int player_index);
void PlayerClient_ESF_CurrentWeapon(void *p, int player_index);
void PlayerClient_ESF_Health(void *p, int player_index);
void PlayerClient_ESF_DeathMsg(void *p, int bot_index);

bool EntityHeadsToPlayer(edict_t *pPlayer, edict_t *pEntity);

void PlayerClassifyForSound(edict_t *pEdict, char *ClassName);
void PlayerClassify(edict_t *pEdict, char *ClassName);

void SetCameraToPlayer(edict_t *pPlayer, edict_t *pEntity);
void DestroyPlayerCamera(edict_t *pPlayer);
void ActivateBeamCam(edict_t *pEdict);
void DeactivateBeamCam(edict_t *pEdict);

bool EntIsInFieldOfView(edict_t *pPlayer, edict_t *pEnt);
bool EntIsInTriggerHurt(edict_t *pEdict);

void ExplodePlayer(edict_t *pEdict);

EComboChargeButton GetComboChargeButtonFromInt(int Value);
bool IsComboChargeButton(int Value);

#ifdef usestats
void LogAllWeaponStatistics(void);
#endif

#ifdef classstuff
void LogCLList(CClient *pClient);

class CClientManager
{
	CClient *pHost;
	short PlayerCount;

public:
	TLinkedList<CClient*> *pClientList;

	void FilterBadClients(CClient *pClient);

	void LetClientsThink(CClient *pClient);
	void ConnectPlayer(edict_t *pEdict);
	void ConnectBot(edict_t *pEdict);
	void DisconnectPlayer(edict_t *pEdict);

	short GetPlayerCount(void)		{ return PlayerCount; };
	CClient *GetClientFromEdict(edict_t *pEdict);
	bool EdictIsFromClient(CClient *pClient, edict_t *pEdict);

	void SetHost(CClient *pClient)  { if (pClient) pHost = pClient; };
	CClient *GetHost(void)			{ return pHost; };

	//Construction/destruction
	CClientManager(void);
	~CClientManager(void);

	// Functions to hook global network messages
	void GlobalNetworkMsgPL(void *pValue);
	void GlobalNetworkMsgStartPowerUp(void *pValue);
	void GlobalNetworkMsgStopPowerUp(void *pValue);
	void GlobalNetworkMsgTransformFX(void *pValue);
	void GlobalNetworkMsgStopTransFX(void *pValue);

	// Debug functions
	void LogClientList(void) { /*pClientList->ForEach(&LogCLList);*/ };

	// List operation functions
	void CountTeams(CClient *pClient, byte *GoodCount, byte *EvilCount);

	// Class functions
	class CCountTeams
	{
		byte GoodCount;
		byte EvilCount;
	public:
		CCountTeams(void) : GoodCount(0), EvilCount (0) {};

		byte GetGoodCount(void) { return GoodCount; };
		byte GetEvilCount(void) { return EvilCount; };

		operator ()(CClient *pClient)
		{
			if (pClient->GetTeam() == 1)
				GoodCount++;
			else
				EvilCount++;
		};
	};
};
#endif

#endif