#include "general_header.h"
#include "studio.h"

extern int num_waypoints;
extern edict_t *listenserver_edict;

bool DrawDBSpots = false;
bool CheckForDBSpot = true;
Vector GatherDBPosForTeamGood = g_vecZero;
Vector GatherDBPosForTeamEvil = g_vecZero;

// Custom made gather spots
SDBSpot GatherSpotsForTeamGood[MAX_CUSTOM_DB_GATHER_SPOTS];
SDBSpot GatherSpotsForTeamEvil[MAX_CUSTOM_DB_GATHER_SPOTS];

int SearchNearestWPForDBSpot(bot_t *pBot)
{
	if ((GatherDBPosForTeamGood == g_vecZero) || (GatherDBPosForTeamEvil == g_vecZero))
		return -1;

	Vector v_GatherVec;

	if (UTIL_GetTeam(pBot->pEdict) == 0)
		v_GatherVec = GatherDBPosForTeamGood;
	else
		v_GatherVec = GatherDBPosForTeamEvil;

	int i = WaypointFindNearest(v_GatherVec, pBot->pEdict, 1000, UTIL_GetTeam(pBot->pEdict));
	if (i == -1)
		cout << "No WP to DB spot\n";
	return i;
}

// returns true if the bot thrown the DB
bool CDBTask::CheckIfThrowToDBSpot()
{
	Vector v_GatherVec;

	if (UTIL_GetTeam(pBot->pEdict) == 0)
		v_GatherVec = GatherDBPosForTeamGood;
	else
		v_GatherVec = GatherDBPosForTeamEvil;
	
	float DistToSpot = UTIL_Distance(pEdict->v.origin, v_GatherVec);

	if (DistToSpot < 150) // Is the bot close?
	{
		// Is the bot on the ground and not flying?
		if ((pEdict->v.gravity > 0.1) && (pEdict->v.flags & FL_ONGROUND))
		{
			// Just throw the db away
			pEdict->v.button |= IN_ATTACK2;
			return true;
		}
		else if (DistToSpot < 125.0f) // Is the bot to close?
			SetMoveDir(pBot, dir_back);
		else
		{
			ResetMoveSpeed(pBot);

			// Make sure we are between 150 - 200 units higher then the spot
			if (v_GatherVec.z <= (pEdict->v.origin.z - 175)) // if not higher then 150 units...
			{
				SetMoveDir(pBot, (dir_up)); // ... go up
			}
			else if (v_GatherVec.z <= (pEdict->v.origin.z - 200)) // else if higher then 200 units...
			{
				SetMoveDir(pBot, (dir_down)); // ... go down
			}
			else // else we can throw :)
			{
				if (flDBThrowDelay == 0.0f)
					flDBThrowDelay = gpGlobals->time + 0.25f;
				else if (flDBThrowDelay < gpGlobals->time)
				{
					pEdict->v.button |= IN_ATTACK2;
					return true;
				}
			}
		}
	}
	return false;
}

bool CDBTask::GoToGatheredSpot()
{
	if (pBot->MadeWish)
		return false;

	if (IsNull(pBot->pGoalDB))
		return false;

	Vector v_dest = pBot->pGoalDB->v.origin;

	if (FVisible(v_dest, pBot->pEdict))
	{
		if (pBot->curr_waypoint_index != -1)
			pBot->curr_waypoint_index = -1; // Reset

		if (UTIL_Distance(pEdict->v.origin, v_dest) < 100.0f)
		{
			if (RANDOM_LONG(1, 100) <= 50)
				pEdict->v.button |= IN_USE;
			ResetMoveSpeed(pBot);
		}
		else
			pBot->AimVec = WrapAngles(UTIL_VecToAngles(v_dest - pEdict->v.origin));

		return true;
	}

	// DB spot is NOT visible, see if we have a goal WP so we can reach it
	if (pBot->waypoint_goal != -1)
	{
		HeadToNextDBWP();
		return true;
	}

	return false;
}

bool CDBTask::GoToDBSpot()
{
	if ((GatherDBPosForTeamGood == g_vecZero) || (GatherDBPosForTeamEvil == g_vecZero))
		return false;

	if (UTIL_GetTeam(pEdict) == 0)
	{
		if (FVisible(GatherDBPosForTeamGood, pBot->pEdict))
		{
			if (pBot->curr_waypoint_index != -1)
				pBot->curr_waypoint_index = -1; // Reset

			if (CheckIfThrowToDBSpot())
				BotTurnAround(pBot); // Turn away from the spot otherwise might pick the db again
			else
				pBot->AimVec = WrapAngles(UTIL_VecToAngles(GatherDBPosForTeamGood - pEdict->v.origin));

			return true;
		}
	}
	else
	{
		if (FVisible(GatherDBPosForTeamEvil, pBot->pEdict))
		{
			if (pBot->curr_waypoint_index != -1)
				pBot->curr_waypoint_index = -1; // Reset

			if (CheckIfThrowToDBSpot())
				BotTurnAround(pBot); // Turn away from the spot otherwise might pick the db again
			else
				pBot->AimVec = WrapAngles(UTIL_VecToAngles(GatherDBPosForTeamEvil - pEdict->v.origin));

			return true;
		}
	}

	// DB spot is NOT visible, see if we have a goal WP so we can reach it
	if (pBot->waypoint_goal != -1)
	{
		HeadToNextDBWP();
		return true;
	}

	return false;
}

void CDBTask::GoToDB()
{
	if (IsNull(pBot->pGoalDB))
	{
		TaskState = done;
		return;
	}

	if ((DBSAreGathered()) || (DBSAreActive() == false))
	{
		TaskState = done;
		return; // Can't collect dragonballs
	}

	// Is the DB visible?
	if (FVisible(pBot->pGoalDB->v.origin, pEdict))
	{
		// Just aim at the DB when visible
		pBot->AimVec = WrapAngles(UTIL_VecToAngles(pBot->pGoalDB->v.origin - pEdict->v.origin));
		if (pBot->curr_waypoint_index != -1)
			pBot->curr_waypoint_index = -1; // Reset
		return;
	}

	// DB is NOT visible see if we have a goal WP so we can reach it
	if (pBot->waypoint_goal != -1)
	{
		HeadToNextDBWP();
	}
	else
		TaskState = done; // Can't reach the DB anymore :(
}

// returns true if the has found a db
bool SearchDB(bot_t *pBot)
{
	if ((IsTeamPlay() == false) || (CVAR_GET_FLOAT("mp_gamemode") != 2.0f))
		return false; // No DB's when no DB mode...

	if (pBot->pBotClient->HasDB)
		return true; // Don't search a db when we already got one :)

	if ((DBSAreGathered()) || (DBSAreActive() == false))
		return false; // Can't collect dragonballs

	edict_t *pent = NULL;
	edict_t *pDragonBall = NULL;
	float Dist = 0.0f;
	float NearestDBDist = 9999.9f;
	int WptInd = -1;
	int NearestWptInd = -1;
	int WptIndNearBot = -1;
	int BotTeam = UTIL_GetTeam(pBot->pEdict);
	Vector v_GatherVec;

	if (UTIL_GetTeam(pBot->pEdict) == 0)
		v_GatherVec = GatherDBPosForTeamGood;
	else
		v_GatherVec = GatherDBPosForTeamEvil;

	if (num_waypoints)
	{
		// Check if there is a valid waypoint near the bot
		WptIndNearBot = WaypointFindNearest(pBot->pEdict->v.origin, pBot->pEdict, 1750.0f,
		                                    BotTeam);
	}

	// Search a dragonball
   	while ((pent = UTIL_FindEntityByClassname(pent, "dragonball")) != NULL)
	{
		if (IsNull(pent))
			continue;

		Dist = UTIL_Distance(pBot->pEdict->v.origin, pent->v.origin);

		if (Dist >= NearestDBDist)
			continue;

		if (UTIL_Distance(v_GatherVec, pent->v.origin) < 1000.0)
			continue;

		// If the dragonball is not visible look the bot can use a waypoint to reach it
		if (!FVisible(pent->v.origin, pBot->pEdict))
		{
			// Is there a waypoint near the bot?
			if (WptIndNearBot == -1)
				continue; // Nope, skip the DB(Can't reach it)

			// Is there is a valid waypoint near the db?
			WptInd = WaypointFindNearest(pent->v.origin, pent, 750.0f, BotTeam);
			if (WptInd == -1)
				continue; // Nope, skip the DB(Can't reach it)

			// See if there is a route to the waypoint near the DB waypoint
			if (WaypointRouteFromTo(WptIndNearBot, WptInd, BotTeam) == WAYPOINT_UNREACHABLE)
				continue; // Nope, skip the DB(Can't reach it)
		}

		// If we get here the DB is reachable and is nearer then the nearest found DB
		NearestDBDist = Dist;
		pDragonBall = pent;

		if (WptInd != -1)
			NearestWptInd = WptInd;

		WptInd = -1;
	}

	if (pDragonBall)
	{
		pBot->pGoalDB = pDragonBall;

		if (NearestWptInd != -1) // Should the bot use a waypoint?
			pBot->waypoint_goal = NearestWptInd; // Set the wp index as goal
		else // else just aim at the DB
			pBot->AimVec = WrapAngles(UTIL_VecToAngles(pDragonBall->v.origin - pBot->pEdict->v.origin));

		return true;
	}

	return false; // No db's found
}

// Returns true when found a DB gather spot
bool SearchDBGatherSpot(bot_t *pBot)
{
	if ((GatherDBPosForTeamGood == g_vecZero) || (GatherDBPosForTeamEvil == g_vecZero))
		return false;

	if (pBot->pBotClient->HasDB == false)
		return false; // Why search a db spot when we don't have a db...

	short Team = UTIL_GetTeam(pBot->pEdict);
	if (Team == 0)
	{
		if (FVisible(GatherDBPosForTeamGood, pBot->pEdict))
		{
			return true;
		}
	}
	else
	{
		if (FVisible(GatherDBPosForTeamEvil, pBot->pEdict))
		{
			return true;
		}
	}

	int GoalWPIndex = -1;

	// Any waypoints?
	if (num_waypoints != 0)
		GoalWPIndex = SearchNearestWPForDBSpot(pBot);

	if (GoalWPIndex != -1)
	{
		// Check if there is a valid waypoint near the bot
		int WPNearBotIndex = WaypointFindNearest(pBot->pEdict->v.origin, pBot->pEdict, 1750.0f,
	 	 	                                     Team);

		// if so check if there is a route from that WP to the WP near the gatherspot
		if ((WPNearBotIndex != -1) && 
			(WaypointRouteFromTo(WPNearBotIndex, GoalWPIndex, Team) != WAYPOINT_UNREACHABLE))
		{
			pBot->waypoint_goal = GoalWPIndex;
			return true;
		}
	}

	return false;
}

bool SearchGatheredSpot(bot_t *pBot)
{
	if (pBot->MadeWish)
		return false;

	short Team = UTIL_GetTeam(pBot->pEdict);
	edict_t *pDB = NULL;

	pDB = UTIL_FindEntityByClassname(pDB, "dragonball");
	if (IsNull(pDB))
		return false;

	Vector v_dest = pDB->v.origin;

	if (FVisible(v_dest, pBot->pEdict))
	{
		pBot->pGoalDB = pDB;
		return true;
	}

	int GoalWPIndex = -1;

	// Any waypoints?
	if (num_waypoints != 0)
		GoalWPIndex = WaypointFindNearest(v_dest, pBot->pEdict, 1750.0f, Team);

	if (GoalWPIndex != -1)
	{
		// Check if there is a valid waypoint near the bot
		int WPNearBotIndex = WaypointFindNearest(pBot->pEdict->v.origin, pBot->pEdict, 1750.0f,
	 	 	                                     Team);

		// if so check if there is a route from that WP to the WP near the gatherspot
		if ((WPNearBotIndex != -1) && 
			(WaypointRouteFromTo(WPNearBotIndex, GoalWPIndex, Team) != WAYPOINT_UNREACHABLE))
		{
			pBot->waypoint_goal = GoalWPIndex;
			pBot->pGoalDB = pDB;
			return true;
		}
	}

	return false;
}

// Returns true when the DB task should be made
bool CheckDBS(bot_t *pBot)
{
	if (pBot->DBCheckDelay > gpGlobals->time)
		return false;

	pBot->DBCheckDelay = gpGlobals->time + 0.75f;

	if (DBSAreGathered())
		return SearchGatheredSpot(pBot);
	else if (pBot->pBotClient->HasDB)
		return SearchDBGatherSpot(pBot);
	else if (!pBot->pBotEnemy)
		return SearchDB(pBot);

	return false;
}

bool DBSAreGathered()
{
	edict_t *pent = NULL;
	// Only need to examine the first DB because if one DB is glowing, all are
	pent = UTIL_FindEntityByClassname(pent, "dragonball");

	if (IsNull(pent))
		return false;

	return (pent->v.renderfx == kRenderFxGlowShell); // DB is glowing so they must be gathered
}

bool DBSAreActive()
{
	edict_t *pent = NULL;
	// Only need to examine the first DB because if one DB is active is, all are
	pent = UTIL_FindEntityByClassname(pent, "dragonball");

	if (IsNull(pent))
		return false;

	UTIL_PrintEntity(pent, "dragonball.txt");

	void *pModel = GET_MODEL_PTR(pent);
	studiohdr_t *pstudiohdr;
	pstudiohdr = (studiohdr_t *)pModel;

	return ((pent->v.solid != SOLID_NOT) && (pent->v.movetype != MOVETYPE_NOCLIP) &&
			(pent->v.skin == 0));
}

/*
void RefreshDBSpots()
{
	edict_t *pent1 = NULL;
	edict_t *pent2 = NULL;
	byte DBCounter = 0;
	byte MostDBSAtSpot = 0;
	byte DBSAtCurrentSpot = 0;
	bool Found2NearDBs = false;
	edict_t *pBestGatherSpot;

	// Count the near DBs
   	while ((pent1 = UTIL_FindEntityByClassname(pent1, "dragonball")) != NULL)
	{
		if (IsNull(pent1))
			continue;

		if (UTIL_Distance(pGatherDBForTeamGood->v.origin, pent2->v.origin) < 75.0)
			DBSAtCurrentSpot++;
	}

	pent1 = NULL;

	// Check a spot for team good
   	while ((pent1 = UTIL_FindEntityByClassname(pent1, "item_dragonball")) != NULL)
	{
		if (IsNull(pent1))
			continue;

		DBCounter = 0;
		pent2 = NULL;

		// Count the near DBs
	   	while ((pent2 = UTIL_FindEntityByClassname(pent2, "dragonball")) != NULL)
		{
			if (IsNull(pent2))
				continue;

			if (UTIL_Distance(pent1->v.origin, pent2->v.origin) < 75.0)
				DBCounter++;
		}

		if (DBCounter > MostDBSAtSpot)
		{
			pBestGatherSpot = pent1;
			MostDBSAtSpot = DBCounter;
		}
	}

	if (MostDBSAtSpot > (DBSAtCurrentSpot + 1))
	{
		pGatherDBForTeamGood = pBestGatherSpot;
	}

	// Count the near DBs for team evil
   	while ((pent1 = UTIL_FindEntityByClassname(pent1, "dragonball")) != NULL)
	{
		if (IsNull(pent1))
			continue;

		if (UTIL_Distance(pGatherDBForTeamEvil->v.origin, pent2->v.origin) < 75.0)
			DBSAtCurrentSpot++;
	}

	pent1 = NULL;
	MostDBSAtSpot = 0;

	// Check a spot for team evil
   	while ((pent1 = UTIL_FindEntityByClassname(pent1, "item_dragonball")) != NULL)
	{
		if (IsNull(pent1))
			continue;

		if (pent1 == pGatherDBForTeamGood)
			continue;

		DBCounter = 0;
		pent2 = NULL;

		// Count the near DBs
	   	while ((pent2 = UTIL_FindEntityByClassname(pent2, "dragonball")) != NULL)
		{
			if (IsNull(pent2))
				continue;

			if (UTIL_Distance(pent1->v.origin, pent2->v.origin) < 75.0)
				DBCounter++;
		}

		if (DBCounter > MostDBSAtSpot)
		{
			pBestGatherSpot = pent1;
			MostDBSAtSpot = DBCounter;
		}
	}

	if (MostDBSAtSpot > (DBSAtCurrentSpot + 1))
	{
		pGatherDBForTeamEvil = pBestGatherSpot;
	}
}
*/
void DecideNextDBSpots()
{
	CheckForDBSpot = false;

	if ((IsTeamPlay() == false) || (CVAR_GET_FLOAT("mp_gamemode") != 2.0f))
		return; // No need to check if no db mode

	short GoodIndex = GetRandomDBSpotIndex(0);
	short EvilIndex = GetRandomDBSpotIndex(1);

	if ((GoodIndex > -1) && (EvilIndex > -1))
	{
		GatherDBPosForTeamGood = GatherSpotsForTeamGood[GoodIndex].origin;
		GatherDBPosForTeamEvil = GatherSpotsForTeamEvil[EvilIndex].origin;
		return; // No need to go further
	}

	if ((GoodIndex == -1) && (EvilIndex != -1))
		cout << "Error: missing DB gather spot for team good!\n";
	else if ((EvilIndex == -1) && (GoodIndex != -1))
		cout << "Error: missing DB gather spot for team evil!\n";
	
	edict_t *pDragonBall = NULL;
	TMultiChoice<edict_t *> DBChoiceForTeamGood;
	edict_t *pTeamGoodGatherSpot = NULL;

   	while ((pDragonBall = UTIL_FindEntityByClassname(pDragonBall, "item_dragonball")) != NULL)
	{
		if (IsNull(pDragonBall))
			continue;

		DBChoiceForTeamGood.Insert(pDragonBall, 50);
	}

	pTeamGoodGatherSpot = NULL;
	DBChoiceForTeamGood.GetSelection(pTeamGoodGatherSpot);

	if (pTeamGoodGatherSpot)
		GatherDBPosForTeamGood = pTeamGoodGatherSpot->v.origin;

	TMultiChoice<edict_t *> DBChoiceForTeamEvil;
	pDragonBall = NULL;

   	while ((pDragonBall = UTIL_FindEntityByClassname(pDragonBall, "item_dragonball")) != NULL)
	{
		if (IsNull(pDragonBall))
			continue;

		if (pDragonBall == pTeamGoodGatherSpot)
			continue;

		DBChoiceForTeamEvil.Insert(pDragonBall, 50);
	}

	pDragonBall = NULL;
	DBChoiceForTeamEvil.GetSelection(pDragonBall);

	if (pDragonBall)
		GatherDBPosForTeamEvil = pDragonBall->v.origin;
}

void InitCustomGatherDBSpots()
{
	int i;
	for(i=0;i<MAX_CUSTOM_DB_GATHER_SPOTS;i++)
	{
		GatherSpotsForTeamGood[i].origin = g_vecZero;
		GatherSpotsForTeamGood[i].Deleted = true;
		GatherSpotsForTeamEvil[i].origin = g_vecZero;
		GatherSpotsForTeamEvil[i].Deleted = true;
	}

	char szMapName[32] = "";
	char FileName[256] = "";
	FILE *pDBFile = NULL;
	
   	strcpy(szMapName, STRING(gpGlobals->mapname));
	strcat(szMapName, ".dbg");
	
	UTIL_BuildFileName(FileName, "dragonball", szMapName);

	cout << "Reading dragonball info file...";
	pDBFile = fopen(FileName, "rb");

	if (!pDBFile)
	{
		cout << " file not found, db spawnpoints will be used as gather spots.\n";
		return;
	}

	SDBFileHeader FileHeader;
	fread(&FileHeader, sizeof(FileHeader), 1, pDBFile);

	if ((FileHeader.flVersion != DB_SPOT_VERSION) || 
        (FStrEq(STRING(gpGlobals->mapname), FileHeader.szMapName) == false))
	{
		fclose(pDBFile);
		return;
	}

	for (i=0;i<MAX_CUSTOM_DB_GATHER_SPOTS;i++)
	{
		if (!pDBFile)
		{
			cout << "Error reading dragonball file!\n";
			return;
		}
		fread(&GatherSpotsForTeamGood[i], sizeof(GatherSpotsForTeamGood[0]), 1, pDBFile);
	}

	for (i=0;i<MAX_CUSTOM_DB_GATHER_SPOTS;i++)
	{
		if (!pDBFile)
		{
			cout << "Error reading dragonball file!\n";
			return;
		}

		fread(&GatherSpotsForTeamEvil[i], sizeof(GatherSpotsForTeamEvil[0]), 1, pDBFile);
	}

	fclose(pDBFile);
	cout << " Done\n";
}

void SaveCustomGatherDBSpots()
{
	if ((IsTeamPlay() == false) || (CVAR_GET_FLOAT("mp_gamemode") != 2.0f))
		return;

	char szMapName[32] = "";
	char FileName[256] = "";
	FILE *pDBFile = NULL;
	byte i;
	
   	strcpy(szMapName, STRING(gpGlobals->mapname));
	strcat(szMapName, ".dbg");
	
	UTIL_BuildFileName(FileName, "dragonball", szMapName);

	cout << "Saving dragonball info file...";
	pDBFile = fopen(FileName, "wb");

	if (!pDBFile)
	{
		cout << " couldn't create file!\n";
		return;
	}

	SDBFileHeader FileHeader;

	// Set file header
	FileHeader.flVersion = DB_SPOT_VERSION;
	strcpy(FileHeader.szMapName, STRING(gpGlobals->mapname));

	fwrite(&FileHeader, sizeof(FileHeader), 1, pDBFile);

	for (i=0;i<MAX_CUSTOM_DB_GATHER_SPOTS;i++)
	{
		fwrite(&GatherSpotsForTeamGood[i], sizeof(GatherSpotsForTeamGood[0]), 1, pDBFile);
	}

	for (i=0;i<MAX_CUSTOM_DB_GATHER_SPOTS;i++)
	{
		fwrite(&GatherSpotsForTeamEvil[i], sizeof(GatherSpotsForTeamEvil[0]), 1, pDBFile);
	}

	fclose(pDBFile);
	cout << " Done\n";
}

void PlaceNewDBGatherSpot(edict_t *pEdict)
{
	if ((IsTeamPlay() == false) || (CVAR_GET_FLOAT("mp_gamemode") != 2.0f))
		return;

	short Team = UTIL_GetTeam(pEdict);
	byte b = 0;
	TraceResult tr;

	if (Team == 0)
	{
		while (b < MAX_CUSTOM_DB_GATHER_SPOTS)
		{
			if (GatherSpotsForTeamGood[b].Deleted)
			{
				UTIL_TraceLine(pEdict->v.origin, pEdict->v.origin + Vector(0, 0, -9999),
							   ignore_monsters, pEdict, &tr);
				if (tr.flFraction < 1.0) // Should alway be true
					GatherSpotsForTeamGood[b].origin = tr.vecEndPos; // Put it on the ground....
				else
					GatherSpotsForTeamGood[b].origin = pEdict->v.origin;

				GatherSpotsForTeamGood[b].Deleted = false;
				return;
			}
			b++;
		}
		
		cout << "Cannot create new DragonBall spot for team good, max reached.\n";
		return;
	}
	else
	{
		while (b < MAX_CUSTOM_DB_GATHER_SPOTS)
		{
			if (GatherSpotsForTeamEvil[b].Deleted)
			{
				UTIL_TraceLine(pEdict->v.origin, pEdict->v.origin + Vector(0, 0, -9999),
							   ignore_monsters, pEdict, &tr);
				if (tr.flFraction < 1.0) // Should alway be true
					GatherSpotsForTeamEvil[b].origin = tr.vecEndPos; // Put it on the ground....
				else
					GatherSpotsForTeamEvil[b].origin = pEdict->v.origin;

				GatherSpotsForTeamEvil[b].Deleted = false;
				return;
			}
			b++;
		}
		
		cout << "Cannot create new DragonBall spot for team evil, max reached.\n";
		return;
	}
}

void DeleteDBGatherSpot(edict_t *pEdict)
{
	if ((IsTeamPlay() == false) || (CVAR_GET_FLOAT("mp_gamemode") != 2.0f))
		return;

	short Team = UTIL_GetTeam(pEdict);
	short index;

	if (Team == 0)
	{
		index = GetNearestDBSpotIndex(pEdict, 75.0f);
		if (index > -1)
		{
			GatherSpotsForTeamGood[index].Deleted = true;
		}
	}
	else
	{
		index = GetNearestDBSpotIndex(pEdict, 75.0f);
		if (index > -1)
		{
			GatherSpotsForTeamEvil[index].Deleted = true;
		}
	}
}

short GetNearestDBSpotIndex(edict_t *pEdict, float flRadius)
{
	Vector v_src = pEdict->v.origin;
	short Team = UTIL_GetTeam(pEdict);
	float NearestDist = 9999.9f;
	float Dist;
	short NearestIndex = -1;
	short s;
	
	if (Team == 0)
	{
		for (s=0;s<MAX_CUSTOM_DB_GATHER_SPOTS;s++)
		{
			if (GatherSpotsForTeamGood[s].Deleted)
				continue;

			Dist = UTIL_Distance(v_src, GatherSpotsForTeamGood[s].origin);

			if ((Dist < NearestDist) && (Dist <= flRadius))
			{
				if (FVisible(GatherSpotsForTeamGood[s].origin, pEdict))
				{
					NearestDist = Dist;
					NearestIndex = s;
				}
			}
		}
	}
	else
	{
		for (s=0;s<MAX_CUSTOM_DB_GATHER_SPOTS;s++)
		{
			if (GatherSpotsForTeamGood[s].Deleted)
				continue;

			Dist = UTIL_Distance(v_src, GatherSpotsForTeamEvil[s].origin);

			if ((Dist < NearestDist) && (Dist <= flRadius))
			{
				if (FVisible(GatherSpotsForTeamEvil[s].origin, pEdict))
				{
					NearestDist = Dist;
					NearestIndex = s;
				}
			}
		}
	}

	return NearestIndex;
}

short GetRandomDBSpotIndex(short Team)
{
	short s;
	TMultiChoice<short> ValidIndexChoices;

	for (s=0;s<MAX_CUSTOM_DB_GATHER_SPOTS;s++)
	{
		if (Team == 0)
		{
			if (GatherSpotsForTeamGood[s].Deleted)
				continue;
		}
		else
		{
			if (GatherSpotsForTeamEvil[s].Deleted)
				continue;
		}
		ValidIndexChoices.Insert(s, 50);
	}

	if(ValidIndexChoices.GetSelection(s) == false)
		return -1; // No valid index

	return s;
}

void DrawDBGatherSpots()
{
	static float LastDBSpotDrawTime = 0.0f;

	if (LastDBSpotDrawTime > gpGlobals->time)
		return;

	LastDBSpotDrawTime = gpGlobals->time + 0.25f;
	
	if ((IsTeamPlay() == false) || (CVAR_GET_FLOAT("mp_gamemode") != 2.0f))
		return;

	if (!UTIL_IsValidPlayer(listenserver_edict, false, false))
		return;

	byte i;
	Vector v_src, v_end;

	// Draw all good DB spots
	for (i=0;i<MAX_CUSTOM_DB_GATHER_SPOTS;i++)
	{
		if (GatherSpotsForTeamGood[i].Deleted)
			continue;

		v_src = GatherSpotsForTeamGood[i].origin;
		v_end = v_src + Vector(0, 0, 68);
		UTIL_DrawBeam(listenserver_edict, v_src, v_end, 30, 0, 0, 255, 0, 250, 5);
	}

	// Draw all evil DB spots
	for (i=0;i<MAX_CUSTOM_DB_GATHER_SPOTS;i++)
	{
		if (GatherSpotsForTeamEvil[i].Deleted)
			continue;

		v_src = GatherSpotsForTeamEvil[i].origin;
		v_end = v_src + Vector(0, 0, 68);
		UTIL_DrawBeam(listenserver_edict, v_src, v_end, 30, 0, 255, 0, 0, 250, 5);
	}
}
