#include "general_header.h"

// ==================================================================
// Code of CAttackManager - Begin
// ==================================================================

CAttackManager *pAttackManager = NULL;

CAttackManager::CAttackManager() : m_fPreviousThinkTime(0.0)
{
	pAttackList = new TLinkedList<SAttackNode*>;
}

CAttackManager::~CAttackManager()
{
	if (pAttackList)
	{
		while (pAttackList->Empty() == false)
			delete pAttackList->Pop();

		delete pAttackList;
	}
}

void CAttackManager::AddAttack(edict_t *pAttack)
{
	if (pAttackList->FindIf(this, &CAttackManager::AttackIsInList, pAttack))
		return;

	short type;

	if (IsBeamAttack(pAttack))
		type = AttackTypeBeam;
	else if (IsNonBeamAttack(pAttack))
		type = AttackTypeBall;
	else if (IsKiBall(pAttack))
		type = AttackTypeKiblast;
	else if (IsOtherAttack(pAttack))
		type = AttackTypeOther;
	else
		type = AttackTypeUnknown;

	SAttackNode *pNode = new SAttackNode(pAttack, type, gpGlobals->time);

	pAttackList->AddNode(pNode);
}

void CAttackManager::Think()
{
	pAttackList->ForEach(this, &CAttackManager::RemoveDeletedAttack);

	if (m_fPreviousThinkTime > gpGlobals->time)
		return;

	m_fPreviousThinkTime = gpGlobals->time + 0.2;

	TLinkedList<SAttackNode*>::SNode *pNode = pAttackList->GetFirst();
	TLinkedList<SAttackNode*>::SNode *pTempNode;

	if (!pNode)
		return;

	while (pNode)
	{
		try
		{
			if ((pNode->Entry->pAttack == NULL) || (IsNull(pNode->Entry->pAttack)))
			{

#ifdef _DEBUG
				FILE *fp = fopen("bot.txt","a");
				fprintf(fp,"Attack delete failure !, attack type: %d\nList:\n\n",
						pNode->Entry->type);
				fclose(fp);

				lcout << "Attack delete failure !\n";
#endif

				if (pNode->Entry)
				{
					delete pNode->Entry;
					pNode->Entry = NULL;
				}

				pTempNode = pNode;
				pNode = pNode->next;
				pAttackList->DeleteNode(pTempNode);
				
				continue;
			}
		}
		catch (...)
		{
			try
			{
				pNode = pNode->next;
				continue;
			}
			catch(...)
			{
				return; // sigh
			}
		}

		if (pNode->Entry->type == AttackTypeUnknown)
		{
			if (IsBeamAttack(pNode->Entry->pAttack))
				pNode->Entry->type = AttackTypeBeam;
			else if (IsNonBeamAttack(pNode->Entry->pAttack))
				pNode->Entry->type = AttackTypeBall;
			else if (IsKiBall(pNode->Entry->pAttack))
				pNode->Entry->type = AttackTypeKiblast;
			else if (IsOtherAttack(pNode->Entry->pAttack))
				pNode->Entry->type = AttackTypeOther;
			else if ((gpGlobals->time - pNode->Entry->CreateTime) > 2.0)
			{
				if (pNode->Entry)
				{
					pNode->Entry = NULL;
					delete pNode->Entry;
				}

				pTempNode = pNode;
				pNode = pNode->next;
				pAttackList->DeleteNode(pTempNode);

				continue;
			}
		}

		if ((pNode->Entry->type != AttackTypeUnknown) &&
			(IsAttack(pNode->Entry->pAttack) == false))
			pNode->Entry->type = AttackTypeUnknown;

		pNode = pNode->next;
	}
}

bool CAttackManager::IsHostileAttack(edict_t *pPlayer, edict_t *pAttack, bool CheckFOV)
{
	if ((!pPlayer) || (!pAttack) || (!pAttack->v.owner))
		return false;

	if (pAttack->v.owner == pPlayer)
	{
		return false;
	}

	if (pAttack->v.velocity.Length() == 0.0)
	{
		return false;
	}

	if ((IsTeamPlay()) && (CVAR_GET_FLOAT("mp_friendlyfire") == 0.0))
	{
		int attacker_team = UTIL_GetTeam(pAttack->v.owner);
		int player_team = UTIL_GetTeam(pPlayer);

		if (attacker_team == player_team)
		{
			return false;
		}
	}
	
	if (EntityHeadsToPlayer(pPlayer, pAttack) == false)
	{
		return false;
	}

	Vector v_end = pAttack->v.origin;

	if (((CheckFOV) && (!FInViewCone(&v_end, pPlayer))) ||
		(!FVisible(v_end, pPlayer)))
	{
		return false;
	}

	return true;
}

void CAttackManager::RemoveDeletedAttack(SAttackNode *pNode)
{
	if (!pNode)
		return;

	if (IsNull(pNode->pAttack))
	{
		pAttackList->DeleteEntry(pNode);
		delete pNode;
		return;
	}

	// clear left overs made by sbots....
	if ((pNode->Attacked) && ((pNode->AttackedTime + 0.75f) < gpGlobals->time))
	{
		edict_t *pPlayer = NULL;
			
		if (pNode->OldVelocity.Length() > 0.0)
			pNode->pAttack->v.velocity = pNode->OldVelocity;
		else
		{
			// search the first valid player who can let explode the attack by
			// 'touching' it
			for (int i=1;i<=gpGlobals->maxClients;i++)
			{
				pPlayer = INDEXENT(i);

				if (UTIL_IsValidPlayer(pPlayer, true, false) == false)
					continue;

				if ((IsTeamPlay()) && (IsNull(pNode->pAttack->v.owner) == false))
				{
					if (pPlayer->v.team == pNode->pAttack->v.owner->v.team)
					{
						continue;
					}
				}

				DispatchTouch(pNode->pAttack, pPlayer);
				break;
			}
		}
	}
}

edict_t *CAttackManager::GetNearestHostileAttack(edict_t *pPlayer, float Radius, bool CheckFOV)
{
	Vector v_src = pPlayer->v.origin;
	Vector v_end;

	edict_t *NearestEnt = NULL;
	float nearestdist = 9999.9;
	float dist = 0.0;

	TLinkedList<SAttackNode*>::SNode *pNode = pAttackList->GetFirst();
	if (!pNode)
		return NULL;

	for (;pNode;pNode=pNode->next)
	{
		if (pNode->Entry->type == AttackTypeUnknown)
		{
			continue;
		}

		v_end = pNode->Entry->pAttack->v.origin;
		dist = (v_end - v_src).Length();

		if ((dist > Radius) || (dist >= nearestdist))
		{
			continue;
		}
 
		if (UTIL_IsValidPlayer(pNode->Entry->pAttack->v.owner, true, false))
		{
			if (pNode->Entry->pAttack->v.owner == pPlayer)
			{
				continue;
			}

			if (pNode->Entry->pAttack->v.velocity.Length() == 0.0)
			{
				continue;
			}

			if ((IsTeamPlay()) && (CVAR_GET_FLOAT("mp_friendlyfire") == 0.0))
			{
	            int attacker_team = UTIL_GetTeam(pNode->Entry->pAttack->v.owner);
		        int player_team = UTIL_GetTeam(pPlayer);

				if (attacker_team == player_team)
				{
					continue;
				}
			}
		}

		if (EntityHeadsToPlayer(pPlayer, pNode->Entry->pAttack) == false)
		{
			continue;
		}

		if (((CheckFOV) && (!FInViewCone(&v_end, pPlayer))) ||
			(!FVisible(v_end, pPlayer)))
		{
			continue;
		}

		nearestdist = dist;
		NearestEnt = pNode->Entry->pAttack;
	}

	return NearestEnt;
}

edict_t *CAttackManager::GetNearestHostileAttack(edict_t *pPlayer, float MinDist, float MaxDist,
												 bool CheckFOV)
{
	Vector v_src = pPlayer->v.origin;
	Vector v_end;

	edict_t *NearestEnt = NULL;
	float nearestdist = 9999.9;
	float dist = 0.0;

	TLinkedList<SAttackNode*>::SNode *pNode = pAttackList->GetFirst();
	if (!pNode)
		return NULL;

	for (;pNode;pNode=pNode->next)
	{
		if (pNode->Entry->type == AttackTypeUnknown)
		{
			continue;
		}

		v_end = pNode->Entry->pAttack->v.origin;
		dist = (v_end - v_src).Length();

		if ((dist > MaxDist) || (dist >= nearestdist) || (dist < MinDist))
		{
			continue;
		}
 
		if (UTIL_IsValidPlayer(pNode->Entry->pAttack->v.owner, true, false))
		{
			if (pNode->Entry->pAttack->v.owner == pPlayer)
			{
				continue;
			}

			if (pNode->Entry->pAttack->v.velocity.Length() == 0.0)
			{
				continue;
			}

			if ((IsTeamPlay()) && (CVAR_GET_FLOAT("mp_friendlyfire") == 0.0))
			{
	            int attacker_team = UTIL_GetTeam(pNode->Entry->pAttack->v.owner);
		        int player_team = UTIL_GetTeam(pPlayer);

				if (attacker_team == player_team)
				{
					continue;
				}
			}
		}

		if (EntityHeadsToPlayer(pPlayer, pNode->Entry->pAttack) == false)
		{
			continue;
		}

		if (((CheckFOV) && (!FInViewCone(&v_end, pPlayer))) ||
			(!FVisible(v_end, pPlayer)))
		{
			continue;
		}

		nearestdist = dist;
		NearestEnt = pNode->Entry->pAttack;
	}

	return NearestEnt;
}

short CAttackManager::CountHostileAttacks(edict_t *pPlayer, float Radius, bool CheckFOV)
{
	Vector v_src = pPlayer->v.origin;
	Vector v_end;

	short AttackCount = 0;
	float dist = 0.0;

	TLinkedList<SAttackNode*>::SNode *pNode = pAttackList->GetFirst();
	if (!pNode)
		return 0;

	for (;pNode;pNode=pNode->next)
	{
		if (pNode->Entry->type == AttackTypeUnknown)
		{
			continue;
		}

		v_end = pNode->Entry->pAttack->v.origin;
		dist = (v_end - v_src).Length();

		if (dist > Radius)
		{
			continue;
		}
 
		if (UTIL_IsValidPlayer(pNode->Entry->pAttack->v.owner, true, false))
		{
			if (pNode->Entry->pAttack->v.owner == pPlayer)
			{
				continue;
			}

			if (pNode->Entry->pAttack->v.velocity.Length() == 0.0)
			{
				continue;
			}

			if ((IsTeamPlay()) && (CVAR_GET_FLOAT("mp_friendlyfire") == 0.0))
			{
	            int attacker_team = UTIL_GetTeam(pNode->Entry->pAttack->v.owner);
		        int player_team = UTIL_GetTeam(pPlayer);

				if (attacker_team == player_team)
				{
					continue;
				}
			}
		}

		if (EntityHeadsToPlayer(pPlayer, pNode->Entry->pAttack) == false)
		{
			continue;
		}

		if (((CheckFOV) && (!FInViewCone(&v_end, pPlayer))) ||
			(!FVisible(v_end, pPlayer)))
		{
			continue;
		}

		AttackCount++;
	}

	return AttackCount;
}

short CAttackManager::CountAttacksByName(edict_t *pPlayer, const char *Name)
{
	short AttackCount = 0;

	TLinkedList<SAttackNode*>::SNode *pNode = pAttackList->GetFirst();
	if (!pNode)
		return 0;

	for (;pNode;pNode=pNode->next)
	{
		if (pNode->Entry->pAttack->v.owner != pPlayer)
		{
			continue;
		}

		if (strcmp(STRING(pNode->Entry->pAttack->v.classname), Name) == 0)
			AttackCount++;
	}

	return AttackCount;
}

float CAttackManager::GetNearestAttackDistByName(edict_t *pPlayer, const char *Name)
{
	float NearestDist = 9999.9;

	TLinkedList<SAttackNode*>::SNode *pNode = pAttackList->GetFirst();
	if (!pNode)
		return -1.0;

	for (;pNode;pNode=pNode->next)
	{
		if (pNode->Entry->pAttack->v.owner != pPlayer)
		{
			continue;
		}

		if (strcmp(STRING(pNode->Entry->pAttack->v.classname), Name) == 0)
		{
			float AttackDist = UTIL_Distance(pPlayer->v.origin, pNode->Entry->pAttack->v.origin);
			if (AttackDist < NearestDist)
				NearestDist = AttackDist;
		}
	}

	return (NearestDist != 9999.9) ? NearestDist : -1.0;
}

float CAttackManager::GetFurthestAttackDistByName(edict_t *pPlayer, const char *Name)
{
	float FurthestDist = -1.0;

	TLinkedList<SAttackNode*>::SNode *pNode = pAttackList->GetFirst();
	if (!pNode)
		return -1.0;

	for (;pNode;pNode=pNode->next)
	{
		if (pNode->Entry->pAttack->v.owner != pPlayer)
		{
			continue;
		}

		if (strcmp(STRING(pNode->Entry->pAttack->v.classname), Name) == 0)
		{
			float AttackDist = UTIL_Distance(pPlayer->v.origin, pNode->Entry->pAttack->v.origin);
			if (AttackDist > FurthestDist)
				FurthestDist = AttackDist;
		}
	}

	return FurthestDist;
}

SSurroundingAttacksInfo CAttackManager::GetSurroundingAttackInfo(edict_t *pPlayer, float Radius,
																 bool CheckFOV)
{
	SSurroundingAttacksInfo SurroundingAttacksInfo;

	Vector v_src = pPlayer->v.origin;
	Vector v_end;

	float dist = 0.0;
	short Damage;
	byte OverAllDangerLevel = 0;
	byte DangerLevel = 0;
	byte HighestDangerLevel = 0;

	TLinkedList<SAttackNode*>::SNode *pNode = pAttackList->GetFirst();
	if (!pNode)
		return SurroundingAttacksInfo;

	for (;pNode;pNode=pNode->next)
	{
		if (pNode->Entry->type == AttackTypeUnknown)
		{
			continue;
		}

		v_end = pNode->Entry->pAttack->v.origin;
		dist = (v_end - v_src).Length();

		if ((dist > Radius) || (dist > 2500.0))
		{
			continue;
		}
 
		if (UTIL_IsValidPlayer(pNode->Entry->pAttack->v.owner, true, false))
		{
			if (pNode->Entry->pAttack->v.owner == pPlayer)
			{
				continue;
			}

			if (pNode->Entry->pAttack->v.velocity.Length() == 0.0)
			{
				continue;
			}

			if ((IsTeamPlay()) && (CVAR_GET_FLOAT("mp_friendlyfire") == 0.0))
			{
	            int attacker_team = UTIL_GetTeam(pNode->Entry->pAttack->v.owner);
		        int player_team = UTIL_GetTeam(pPlayer);

				if (attacker_team == player_team)
				{
					continue;
				}
			}
		}

		if (EntityHeadsToPlayer(pPlayer, pNode->Entry->pAttack) == false)
		{
			continue;
		}

		if (((CheckFOV) && (!FInViewCone(&v_end, pPlayer))) ||
			(!FVisible(v_end, pPlayer)))
		{
			continue;
		}

		// When we get here, the attack is hostile...

		SurroundingAttacksInfo.NumAttacks++; // Yet another one
		
		// Determine how hostile it is
		DangerLevel = 0;

		// How far away are you?
		if (dist > 2250.0)
			DangerLevel = 0; // Nothing to worry(yet)
		else if (dist > 1750.0)
			DangerLevel++;
		else if (dist > 1250.0)
			DangerLevel += 3;
		else if (dist > 950.0)
			DangerLevel += 6;
		else if (DangerLevel > 250.0)
			DangerLevel += 10;
		else
			DangerLevel += 13; // ok FLEE!

		// How strong are you? (looks stupid if we teleport 5 times for a kiblast)
		Damage = pNode->Entry->pAttack->v.dmg; // short-cut

		if (Damage >= 100)
			DangerLevel += 13; // WAAAAAAAAAAAAA!
		else if (Damage >= 75)
			DangerLevel += 10;
		else if (Damage >= 50)
			DangerLevel += 6;
		else if (Damage >= 25)
			DangerLevel += 3;
		else if (Damage >= 0)
			DangerLevel++;

		if (HighestDangerLevel < DangerLevel)
		{
			HighestDangerLevel = DangerLevel;
			SurroundingAttacksInfo.pMostDangerousAttack = pNode->Entry->pAttack;
		}

		if (DangerLevel >= 16)
			OverAllDangerLevel += 9;
		else if (DangerLevel >= 13)
			OverAllDangerLevel += 7;
		else if (DangerLevel >= 9)
			OverAllDangerLevel += 5;
		else if (DangerLevel >= 6)
			OverAllDangerLevel += 3;
		else if (DangerLevel >= 4)
			OverAllDangerLevel += 1;
	}

	// Determine the overall danger for the player
	if (OverAllDangerLevel >= 9)
		SurroundingAttacksInfo.DangerLevel = danger_very_heavy;
	else if (OverAllDangerLevel >= 7)
		SurroundingAttacksInfo.DangerLevel = danger_heavy;
	else if (OverAllDangerLevel >= 5)
		SurroundingAttacksInfo.DangerLevel = danger_hard;
	else if (OverAllDangerLevel >= 3)
		SurroundingAttacksInfo.DangerLevel = danger_medium;
	else if (OverAllDangerLevel >= 1)
		SurroundingAttacksInfo.DangerLevel = danger_light;

	return SurroundingAttacksInfo;
}

CAttackManager::TAttackListNode *CAttackManager::GetNextHostileAttack(TAttackListNode *pCurrentNode,
																	  edict_t *pPlayer, bool CheckFOV)
{
	TAttackListNode *pNode = pCurrentNode;

	if (pNode == NULL)
		pNode = pAttackList->GetFirst();
	else
		pNode = pNode->next;

	if (pNode == NULL)
		return pNode;

	while(pNode)
	{
		if (IsHostileAttack(pPlayer, pNode->Entry->pAttack, CheckFOV))
			return pNode;

		pNode = pNode->next;
	}

	return NULL;
}

edict_t *CAttackManager::SearchAttackByName(edict_t *pPlayer, const char *szAttackName)
{
	SAttackNode *pNode = pAttackList->FindIf(this, &CAttackManager::AttackIsFromPlayer,
											 pPlayer, szAttackName);
	return ((pNode) ? pNode->pAttack : NULL);
}

edict_t *CAttackManager::SearchLastAttack(edict_t *pPlayer)
{
	SAttackNode *pNode = pAttackList->FindIfLast(this, &CAttackManager::IsLastAttackFromPlayer,
												 pPlayer);
	return ((pNode) ? pNode->pAttack : NULL);
}

// Attack list operation functions
bool CAttackManager::IsLastAttackFromPlayer(SAttackNode *pNode, edict_t *pPlayer)
{
	return (pNode->pAttack->v.owner == pPlayer);
}

bool CAttackManager::AttackIsFromPlayer(SAttackNode *pNode, edict_t *pPlayer, const char *Name)
{
	if (pNode->pAttack->v.owner == pPlayer)
	{
		return (FStrEq(Name, STRING(pNode->pAttack->v.classname)) == TRUE);
	}
	return false;
}

// debug function(s)

void CAttackManager::PrintEntry(SAttackNode *pNode)
{
	if (!pNode)
	{
		clog << "Error! Wrong node\n";
		return;
	}

	clog << "AttackName: " << STRING(pNode->pAttack->v.classname) << "\n";
	clog << "Type: " << pNode->type << "\n";
	clog << "Attacked: " << pNode->Attacked << "\n";
	clog << "AttackedTime: " << (gpGlobals->time - pNode->AttackedTime) << "\n";
	clog << "Create time: " << (gpGlobals->time - pNode->CreateTime) << "\n";
}

void CAttackManager::PrintAttackList()
{
	clog << "\nAttack list on " << gpGlobals->time << "\n\n";
	pAttackList->ForEach(this, &CAttackManager::PrintEntry);
}

// ==================================================================
// Code of CAttackManager - End
// ==================================================================

// ==================================================================
// Misc functions - Begin
// ==================================================================

#ifdef usestats
void IdentifyAttack(edict_t *pPlayer, char *szAttackName)
{
	if ((!pPlayer) || (!pPlayer->v.dmg_inflictor))
		return;

	edict_t *pDamager = pPlayer->v.dmg_inflictor;

	if (!(pDamager->v.flags & FL_CLIENT))
	{
		if (pDamager->v.owner == pPlayer)
		{
			szAttackName[0] = 0;
			lcout << "Damaged by self owned attack\n";
			return;
		}

		const char *DmgInflictorName = STRING(pDamager->v.classname);

		if (FStrEq(DmgInflictorName, "worldspawn"))
		{
			szAttackName[0] = 0;
			return;
		}

		if ((FStrEq(DmgInflictorName,"blueattack")) || (FStrEq(DmgInflictorName,"greenattack")) ||
			(FStrEq(DmgInflictorName,"purpleattack")) || (FStrEq(DmgInflictorName,"yellowattack")))
			strcpy(szAttackName, "genattack");
		else if (FStrEq(DmgInflictorName, "kiblob"))
			strcpy(szAttackName, "kiblast");
		else
			strcpy(szAttackName, DmgInflictorName);
		return;
	}

	if (pDamager == pPlayer)
	{
		szAttackName[0] = 0;
		lcout << "Damaged by self\n";
		return;
	}

	ClientInfo_t *pClient = UTIL_GetClientPointer(pDamager);

	// For now just see what the damagers current weapon is...

	if (pClient->CurrentWeapon.iId == ESF_WEAPON_MELEE)
		strcpy(szAttackName, "Melee");
	else if (pClient->CurrentWeapon.iId == ESF_WEAPON_SWORD)
		strcpy(szAttackName, "Melee");
	else if (pClient->CurrentWeapon.iId == ESF_WEAPON_EYELASER)
		strcpy(szAttackName, "Eye Laser");
	else if (pClient->CurrentWeapon.iId == ESF_WEAPON_FINGERLASER)
		strcpy(szAttackName, "Finger Laser");
	else if (pClient->CurrentWeapon.iId == ESF_WEAPON_CANDY)
		strcpy(szAttackName, "Candy");
	else
		szAttackName[0] = 0;

	return;
}
#endif


// ==================================================================
// Misc functions - End
// ==================================================================
