#ifndef MISC_H
#define MISC_H

// ==================================================================
// Code of TLinkedList - Start
// ==================================================================

template <class C> class TLinkedList
{
public:
	struct SNode
	{
		C Entry;
		SNode *next;
		SNode *prev;

		SNode(void) : next(NULL), prev(NULL)
		{
			memset(&Entry, 0, sizeof(Entry));
		};
	};

	// member functions

	void AddNode(C entry)
	{
		if (!pNodeList)
		{
			pNodeList = new SNode;
			pNodeList->Entry = entry;
			pNodeList->next = NULL;
			pNodeList->prev = NULL;
			pLastNode = pNodeList;
		}
		else
		{
			pLastNode->next = new SNode;
			pLastNode->next->prev = pLastNode;
			pLastNode = pLastNode->next;
			pLastNode->Entry = entry;
			pLastNode->next = NULL;
		}
	}

	void DeleteEntry(C Entry)
	{
		SNode *pNode = pNodeList;
		if (!pNode)
			return;

		if (pNode->Entry == Entry) // first node
		{
			if (pNodeList == pLastNode)
				pLastNode = NULL;

			pNodeList = pNodeList->next;
			if (pNodeList)
				pNodeList->prev = NULL;
			pNode->next = NULL;
			delete pNode;
			pNode = NULL;
			return;
		}

		if (Entry == pLastNode->Entry) // last node
		{
			pNode = pLastNode;
			pLastNode = pLastNode->prev;
			pLastNode->next = NULL;

			pNode->next = NULL;
			pNode->prev = NULL;
			delete pNode;
			pNode = NULL;
			return;
		}

		// node is somewhere in the middle
		pNode = SearchNode(Entry);

		if (!pNode)
			return;

		SNode *pPrevNode = pNode->prev;

		if (!pPrevNode)
			return;

		// unlink pNode
		//pPrevNode->prev = pPrevNode;
		pNode->next->prev = pPrevNode;
		pPrevNode->next = pNode->next;

		pNode->next = NULL;
		pNode->prev = NULL;
		delete pNode;
		pNode = NULL;
	}

	void DeleteNode(SNode *pNode)
	{
		if (!pNode)
			return;

		if (pNodeList == pNode) // first node
		{
			if (pNodeList == pLastNode)
				pLastNode = pNodeList->next;

			pNodeList = pNodeList->next;
			if (pNodeList)
				pNodeList->prev = NULL;
			pNode->next = NULL;
			delete pNode;
			pNode = NULL;
			return;
		}

		if (pNode == pLastNode) // last node
		{
			pNode = pLastNode;
			pLastNode = pLastNode->prev;
			pLastNode->next = NULL;

			pNode->next = NULL;
			pNode->prev = NULL;
			delete pNode;
			pNode = NULL;
			return;
		}

		// node is somewhere in the middle

		SNode *pPrevNode = pNode->prev;

		if (!pPrevNode)
			return;

		// unlink pNode
		pNode->next->prev = pPrevNode;
		pPrevNode->next = pNode->next;

		pNode->next = NULL;
		pNode->prev = NULL;
		delete pNode;
		pNode = NULL;
	}

	void DeleteAllNodes(void)
	{
		SNode *pNode = pNodeList;
		SNode *pTemp;
		while (pNode != NULL)
		{
			pTemp = pNode;
			pNode = pNode->next;
			pTemp->next = NULL;
			pTemp->prev = NULL;
			delete pTemp;
		}
		pNodeList = pLastNode = NULL;
	}

	SNode *SearchNode(C Entry)
	{
		SNode *pNode = pNodeList;
		while(pNode)
		{
			if (pNode->Entry == Entry)
				return pNode;
			pNode = pNode->next;
		}
		return NULL;
	}

	C Pop(void)
	{
		if (!pNodeList)
			return static_cast<C>(NULL);

		C Entry = pNodeList->Entry;
		DeleteNode(pNodeList);
		return Entry;
	}

	SNode *GetFirst(void) { return pNodeList; };
	SNode *GetLast(void) { return pLastNode; };

	void LogList(void)
	{
		short max = 350;
		clog << "Logging list...\n";
		SNode *pNode = pNodeList;
		if (!pNode)
		{
			clog << "List is empty\n";
			return;
		}

		while((pNode) && (max > 0))
		{
			clog << "Entry: " << pNode->Entry << "\n";
			clog << "next: " << int(pNode->next) << "\n";
			clog << "prev " << int(pNode->prev) << "\n";
			pNode = pNode->next;
			max--;
		}
	}

	// Call on each entry the member function of a class until it returns true
	template <class T> bool Untill(T *ClassObj, bool (T::*MemberFunc)(C))
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next; // Go to the next node incase the memberfunc deletes it
			if ((ClassObj->*MemberFunc)(Entry))
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	// Call on each entry the FuncToCall until it returns false
	template <class S> bool While(bool (*FuncToCall)(C))
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;

			if (FuncToCall(Entry) == false)
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	// Call on each entry the FuncToCall until it returns false 1 arg version
	template <class T, class T2> bool While(bool (*FuncToCall)(C, T2), T2 arg1)
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;

			if (FuncToCall(Entry, arg1) == false)
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	// Call on each entry the FuncToCall until it returns false 2 arg version
	template <class T, class T2, class T3> bool While(bool (*FuncToCall)(C, T2, T3), T2 arg1,
													  T3 arg2)
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;

			if (FuncToCall(Entry, arg1, arg2) == false)
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	// Call on each entry the FuncToCall until it returns false 3 arg version
	template <class T, class T2, class T3, class T4>
	bool While(bool (*FuncToCall)(C, T2, T3, T4), T2 arg1, T3 arg2, T4 arg3)
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;

			if (FuncToCall(Entry, arg1, arg2, arg3) == false)
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	// Call on each entry the ClassFunction until it returns false
	template <class S> bool While(S ClassFuncToCall)
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;

			if (ClassFuncToCall(Entry) == false)
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	// Call on each entry the member function of a class until it returns false
	template <class T> bool While(T *ClassObj, bool (T::*MemberFunc)(C))
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next; // Go to the next node incase the memberfunc deletes it
			if ((ClassObj->*MemberFunc)(Entry) == false)
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	// Call on each entry the member function of a class until it returns false - 1 parameter ver
	template <class T, class T2> bool While(T *ClassObj, bool (T::*MemberFunc)(C, T2), T2 arg1)
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next; // Go to the next node incase the memberfunc deletes it
			if ((ClassObj->*MemberFunc)(Entry, arg1) == false)
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	// Call on each entry the member function of a class until it returns false - 2 parameter ver
	template <class T, class T2, class T3> 
	bool While(T *ClassObj, bool (T::*MemberFunc)(C, T2, T3), T2 arg1, T3 arg2)
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return false;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next; // Go to the next node incase the memberfunc deletes it
			if ((ClassObj->*MemberFunc)(Entry, arg1, arg2) == false)
				return true;
		}
		while (pNode);
		return false; // Didn't stopped
	}

	template<class S> void CallMemberFuncForEach(void (S::*MemberFunc)())
	{
		SNode *pNode = pNodeList;
		while (pNode)
		{
			(pNode->Entry->*MemberFunc)();
			pNode = pNode->next;
		}
	}

	// Class function version
	template <class S> void ForEach(S& pClass)
	{
		SNode *pNode = pNodeList;
		C Entry;
		if (!pNode)
			return;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;
			pClass(Entry);
		}
		while (pNode);
	}

	// Member function call version
	template <class T> void ForEach(T *ClassObj, void (T::*MemberFunc)(C))
	{
		SNode *pNode = pNodeList;
		C Entry;
		if (!pNode)
			return;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;
			(ClassObj->*MemberFunc)(Entry);
		}
		while (pNode);
	}

	// Member function call version, takes 1 parameter extra
	template <class T, class T2> void ForEach(T *ClassObj, void (T::*MemberFunc)(C, T2),
											  T2 arg1)
	{
		SNode *pNode = pNodeList;
		C Entry;
		if (!pNode)
			return;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;
			(ClassObj->*MemberFunc)(Entry, arg1);
		}
		while (pNode);
	}

	// Member function call version, takes 2 parameters extra
	template <class T, class T2, class T3>
	void ForEach(T *ClassObj, void (T::*MemberFunc)(C, T2, T3), T2 arg1, T3 arg2)
	{
		SNode *pNode = pNodeList;
		C Entry;
		if (!pNode)
			return;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;
			(ClassObj->*MemberFunc)(Entry, arg1, arg2);
		}
		while (pNode);
	}

	// Member function call version, takes 3 parameters extra
	template <class T, class T2, class T3, class T4>
	void ForEach(T *ClassObj, void (T::*MemberFunc)(C, T2, T3, T4), T2 arg1, T3 arg2, T4 arg3)
	{
		SNode *pNode = pNodeList;
		C Entry;
		if (!pNode)
			return;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;
			(ClassObj->*MemberFunc)(Entry, arg1, arg2, arg3);
		}
		while (pNode);
	}

	void EditEntry(C OrigEntry, C NewVal)
	{
		SNode *pNode = SearchNode(OrigEntry);

		if (pNode)
			pNode->Entry = NewVal;
	}

	bool IsInList(C Entry)
	{
		return (SearchNode(Entry) != NULL);
	}

	bool Empty(void)		{ return (pNodeList == NULL); };

	template<class T> C FindIf(T *pClass, bool(T::*MemberFunc)(C))
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return NULL;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;
			if ((pClass->*MemberFunc)(Entry))
				return Entry;
		}
		while (pNode);

		return NULL;
	}

	template<class T, class T2> C FindIf(T *pClass, bool(T::*MemberFunc)(C, T2), T2 arg1)
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return NULL;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;
			if ((pClass->*MemberFunc)(Entry, arg1))
				return Entry;
		}
		while (pNode);

		return NULL;
	}

	template<class T, class T2, class T3> C FindIf(T *pClass, bool(T::*MemberFunc)(C, T2, T3), 
												   T2 arg1, T3 arg2)
	{
		SNode *pNode = pNodeList;
		C Entry;

		if (!pNode)
			return NULL;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->next;
			if ((pClass->*MemberFunc)(Entry, arg1, arg2))
				return Entry;
		}
		while (pNode);

		return NULL;
	}

	C FindIfLast(bool(*FuncToCall)(C))
	{
		SNode *pNode = pLastNode;
		C Entry;

		if (!pNode)
			return NULL;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->prev;
			if ((FuncToCall)(Entry))
				return Entry;
		}
		while (pNode);

		return NULL;
	}

	template<class T> C FindIfLast(T *pClass, bool(T::*MemberFunc)(C))
	{
		SNode *pNode = pLastNode;
		C Entry;

		if (!pNode)
			return NULL;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->prev;
			if ((pClass->*MemberFunc)(Entry))
				return Entry;
		}
		while (pNode);

		return NULL;
	}

	template<class T, class T2> C FindIfLast(T *pClass, bool(T::*MemberFunc)(C, T2), T2 arg1)
	{
		SNode *pNode = pLastNode;
		C Entry;

		if (!pNode)
			return NULL;

		do
		{
			Entry = pNode->Entry;
			pNode = pNode->prev;
			if ((pClass->*MemberFunc)(Entry, arg1))
				return Entry;
		}
		while (pNode);

		return NULL;
	}

	// construction/destruction
	TLinkedList(void)
	{
		pNodeList = NULL; 
		pLastNode = NULL;
	};

	~TLinkedList(void)
	{
		DeleteAllNodes();
	};

private:
	SNode *pNodeList;
	SNode *pLastNode;
};

// ==================================================================
// Code of TLinkedList - End
// ==================================================================

// ==================================================================
// Code of TMutiChoice - Begin
// ==================================================================

template <class C> class TMultiChoice 
{
	struct SMultiChoice
	{
		int MinVal;
		int MaxVal;
		C Choice;
		SMultiChoice(void) : MinVal(0), MaxVal(0){};
	};

	int TotalVal;
	TLinkedList<SMultiChoice*> *pChoiceList;

public:
	TMultiChoice(void) : pChoiceList(new TLinkedList<SMultiChoice*>), TotalVal(0) // Constructor
	{
	};
	
	~TMultiChoice(void) // Destructor
	{
		while(pChoiceList->Empty() == false)
		{
			SMultiChoice *pEntry = pChoiceList->Pop();
			if (pEntry)
				delete pEntry;
			else
				break;
		}
		delete pChoiceList;
		pChoiceList = NULL;
	}

	void Insert(C Choice, short Percent)
	{
		if (Percent == 0)
			return;

		SMultiChoice *pChoiceEntry = new SMultiChoice;
		
		pChoiceEntry->MinVal = TotalVal;
		pChoiceEntry->MaxVal = TotalVal + Percent;
		pChoiceEntry->Choice = Choice;

		pChoiceList->AddNode(pChoiceEntry);

		TotalVal += Percent;
	};

	bool FindSelection(SMultiChoice *MS, int Choice)
	{
		if ((Choice >= MS->MinVal) && (Choice < MS->MaxVal))
		{
			return true;
		}
		return false;
	}

	void ClearChoices(void)
	{
		TotalVal = 0;
		while(pChoiceList->Empty() == false)
			delete pChoiceList->Pop();
	}

	bool GetSelection(C &Var)
	{
		int Choice = RANDOM_LONG(0, TotalVal);
		SMultiChoice *pEntry = pChoiceList->FindIf(this, &TMultiChoice::FindSelection, Choice);

		if (!pEntry)
			return false;

		Var = pEntry->Choice;

		return true;
	};
};

// ==================================================================
// Code of TMutiChoice - End
// ==================================================================


// ==================================================================
// Structure for info about surrounding attacks
// ==================================================================

struct SSurroundingAttacksInfo
{
	edict_t *pMostDangerousAttack;
	byte DangerLevel;	// How much danger is there for all attacks?
	byte NumAttacks;	// How much visible attacks are hostile?
	SSurroundingAttacksInfo(void) : pMostDangerousAttack(NULL), DangerLevel(0), NumAttacks(0) {};
};


// ==================================================================
// Code of CAttackManager - Start
// ==================================================================

class CAttackManager
{
	struct SAttackNode
	{
		edict_t *pAttack;
		short type;
		float CreateTime;

		// for sbots
		bool Attacked;
		float AttackedTime;
		Vector OldVelocity;

		SAttackNode(edict_t *pEntity, short t, float CT) : pAttack(pEntity), type(t),
														   CreateTime(CT), Attacked(false),
														   AttackedTime(0.0),
														   OldVelocity(g_vecZero)
		{};
		~SAttackNode(void)
		{
			pAttack = NULL;
		}
	};

	TLinkedList<SAttackNode*> *pAttackList;
	float m_fPreviousThinkTime;

public:
	typedef TLinkedList<SAttackNode*>::SNode TAttackListNode;

	CAttackManager(void); // Construction
	~CAttackManager(void); // Destruction

	void AddAttack(edict_t *pAttack);
	void ClearList(void)	{ pAttackList->DeleteAllNodes(); };
	void Think(void);

	bool IsHostileAttack(edict_t *pPlayer, edict_t *pAttack, bool CheckFOV);

	edict_t *GetNearestHostileAttack(edict_t *pPlayer, float Radius, bool CheckFOV);
	edict_t *GetNearestHostileAttack(edict_t *pPlayer, float MinDist, float MaxDist,
									 bool CheckFOV);
	short CountHostileAttacks(edict_t *pPlayer, float Radius, bool CheckFOV);
	short CountAttacksByName(edict_t *pPlayer, const char *Name);
	// returns -1 if no attack found
	float GetNearestAttackDistByName(edict_t *pPlayer, const char *Name);
	// returns -1 if no attack found
	float GetFurthestAttackDistByName(edict_t *pPlayer, const char *Name);
	SSurroundingAttacksInfo GetSurroundingAttackInfo(edict_t *pPlayer, float Radius,
													bool CheckFOV);

	edict_t *SearchLastAttack(edict_t *pPlayer);
	edict_t *SearchAttackByName(edict_t *pPlayer, const char *szAttackName);
	TAttackListNode *GetNextHostileAttack(TAttackListNode *pCurrentNode, edict_t *pPlayer,
										  bool CheckFOV);

	TAttackListNode *GetFirstAttack(void) { return pAttackList->GetFirst(); };
	// List operation functions
	void RemoveDeletedAttack(SAttackNode *pAttackNode);
	bool AttackIsInList(SAttackNode *pNode, edict_t *pAttack){ return (pNode->pAttack == pAttack); };
	bool IsLastAttackFromPlayer(SAttackNode *pNode, edict_t *pPlayer);
	bool AttackIsFromPlayer(SAttackNode *pNode, edict_t *pPlayer, const char *Name);

	// For debugging
	void PrintEntry(SAttackNode *pNode);
	void PrintAttackList(void);
};


// ==================================================================
// Code of CAttackManager - End
// ==================================================================

// ==================================================================
// Misc functions - Begin
// ==================================================================

void IdentifyAttack(edict_t *pPlayer, char *szAttackName);

// ==================================================================
// Misc functions - End
// ==================================================================

#endif // MISC_H