#include <exception>
#include "general_header.h"

extern edict_t *listenserver_edict;
extern ClientInfo_t ClientInfo[32];
extern BotPersonality_t BotPersonalities[9];
extern float delays[5][2];
extern bool follow_mode;
extern bool debug_mode;
extern bool BotsAreIdle;
extern bool BotDontSearchEnemy;
extern bool b_botdontshoot;
extern bool should_bot_chat;
extern WAYPOINT waypoints[MAX_WAYPOINTS];

//////////////////////////////////////////////////////
//					Delays							//
//////////////////////////////////////////////////////

float GeneralReactionDelays[5][2] =
{
	{0.015, 0.035},
	{0.035, 0.045},
	{0.075, 0.100},
	{0.150, 0.250},
	{0.300, 0.500}
};


//////////////////////////////////////////////////////
//				Class code BEGIN				    //
//////////////////////////////////////////////////////

// Base task class
CBaseTask::CBaseTask(bot_t *pB)
{
	pBot = pB;
	pEdict = pB->pEdict;
	TaskState = busy;

	NesseserelyChecks = BotSenseChecks = BotMiscChecks = BotStatusChecks = 0;

	SetGeneralDelay();
}

CBaseTask::~CBaseTask()
{
	pBot = NULL;
	pEdict = NULL;
}

void CBaseTask::MakeCombo(ComboTypes CT)
{
	TaskState = done;
	pBot->TaskAfterTask = TaskCombo;
	pBot->pBotCombo = new BotCombo(pBot, CT);
}

void CBaseTask::GoToAnPreviousLocation()
{
	for (int i=MaxOldPositions-1;i>=0;i--)
	{
		if ((pBot->pBotClient->OldPositions[i].CreateTime > (gpGlobals->time - MaxOldPositionLifeTime)) &&
			(FVisible(pBot->pBotClient->OldPositions[i].Place, pEdict)))
		{
			int vecDir = UTIL_VecDir(pEdict, pBot->pBotClient->OldPositions[i].Place);

			pBot->f_move_speed = pBot->f_strafe_speed = pBot->f_updown_speed = 0.0;

			if (vecDir & dir_left)
				pBot->f_strafe_speed = -pBot->f_max_speed;
			if (vecDir & dir_right)
				pBot->f_strafe_speed = pBot->f_max_speed;
			if (vecDir & dir_down)
				pBot->f_updown_speed = -pBot->f_max_speed;
			if (vecDir & dir_up)
				pBot->f_updown_speed = pBot->f_max_speed;
			if (vecDir & dir_back)
				pBot->f_move_speed = -pBot->f_max_speed;
			if (vecDir & dir_forward)
				pBot->f_strafe_speed = pBot->f_max_speed;

			return;
		}
	}
}

void CBaseTask::SetGeneralDelay()
{
	if (pBot->SBotType > 0)
		Delay = gpGlobals->time; // Super bots don't have a reaction delay :-)
	else
	{
		float MinDelay = GeneralReactionDelays[pBot->bot_skill][0];
		float MaxDelay = GeneralReactionDelays[pBot->bot_skill][1];
		Delay = gpGlobals->time + RANDOM_FLOAT(MinDelay, MaxDelay);
	}
}

void CBaseTask::ManageEnemies()
{
	edict_t *pTemp = pBot->pBotEnemy;
	pBot->pBotEnemy = GetBestEnemyFromList(pBot);

	// Did the bot lost his enemy ?
	if ((pBot->pBotEnemy == NULL) && (UTIL_IsValidPlayer(pTemp, false)))
	{
		// Add delay
		extern float EnemySearchDelays[5][2];
		float MinDelay = EnemySearchDelays[pBot->bot_skill][0];
		float MaxDelay = EnemySearchDelays[pBot->bot_skill][1];

		pBot->f_search_enemy_delay = gpGlobals->time + RANDOM_FLOAT(MinDelay, MaxDelay);
	}
}

bool CBaseTask::ShouldManageEnemys()
{
	return false;
}

// Power up task
CPowerUpTask::CPowerUpTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = 0;
}

CPowerUpTask::~CPowerUpTask()
{
	FakeClientCommand(pEdict,"-powerup",NULL,NULL);
	pBot->BotPowerUpDelay = gpGlobals->time + 0.6;
}

bool CPowerUpTask::KillTaskOnDanger(short DangerLevel)
{
	int MinKiLevel = 990; // You cannot reach 1000(=max) when someone flys...

	if (pBot->pBotEnemy)
	{
		switch (pBot->AggressionLevel)
		{
		case Aggressive:
			MinKiLevel = 550;
			break;
		case Normal:
			MinKiLevel = 700;
			break;
		case Defensive:
			MinKiLevel = 850;
			break;
		}
	}

	switch (DangerLevel)
	{
	case danger_medium:
		MinKiLevel = 100; // hurry up
		break;
	case danger_hard:
	case danger_heavy:
		MinKiLevel = 20;
		break;
	}

	if (pBot->bot_ki > MinKiLevel)
	{
		TaskState = done;
		return true;
	}

	return false;
}

void CPowerUpTask::TaskThink()
{
	int MinKiLevel = 990;

	if (pBot->pBotEnemy)
	{
		float flDist = UTIL_Distance(pEdict->v.origin, pBot->pBotEnemy->v.origin);
		if (flDist < 1250.0f)
		{
			switch (pBot->AggressionLevel)
			{
			case Aggressive:
				MinKiLevel = 700;
				break;
			case Normal:
				MinKiLevel = 750;
				break;
			case Defensive:
				MinKiLevel = 800;
				break;
			}
		}
	}

	if (pBot->bot_ki >= MinKiLevel)
	{
		TaskState = done; // we're done now :)
		return;
	}

	pBot->f_move_speed = pBot->f_strafe_speed = pBot->f_updown_speed = 0.0;

	if ((!pBot->BotPowerUps) && (RANDOM_LONG(0, 1) == 0))
	{
		FakeClientCommand(pEdict, "+powerup", NULL, NULL);
	}
}

ETaskCode CPowerUpTask::TaskIdentify()
{
	return TaskPowerup;
}

bool CPowerUpTask::ShouldManageEnemys()
{
	return true;
}

// Teleport task
CTeleportTask::CTeleportTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckHit | CheckSounds | CheckSurroundingPlayers | CheckAttacks);
	BotMiscChecks = CheckKameTorpLeftOvers;

	if (!pBot->TeleportDir)
		pBot->TeleportDir = GetRandomTeleDir(pBot);

	if (pBot->teleport_delay < gpGlobals->time)
		pBot->teleport_delay = Delay + 0.2;

	if (BotShouldFallDown(pBot))
	{
		if (pEdict->v.gravity < 0.1)
			FakeClientCommand(pEdict, "togglefly", NULL, NULL);

		pBot->ShouldFly = false;
		pBot->CheckFlyDelay = gpGlobals->time + RANDOM_FLOAT(0.8, 1.2);
	}
}

CTeleportTask::~CTeleportTask()
{
	pBot->TeleportDir = 0;
	pBot->teleport_delay = 0.0;
}

bool CTeleportTask::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel == danger_heavy)
		pBot->teleport_delay = 0.0;
	else if (DangerLevel == danger_hard)
	{
		if ((pBot->teleport_delay - gpGlobals->time) > 0.1)
			pBot->teleport_delay = gpGlobals->time + 0.1;
	}

	return false;
}

void CTeleportTask::TaskThink()
{
	pBot->f_move_speed = 0.0;
	pBot->f_strafe_speed = 0.0;
	pBot->f_updown_speed = 0.0;

	pBot->AimVec.x = 0;

//	    ClientPrint(listenserver_edict,HUD_PRINTNOTIFY,"teleport\n");
	if (pBot->TeleportDir & dir_forward)
		pBot->f_move_speed = pBot->f_max_speed;
	if (pBot->TeleportDir & dir_back)
		pBot->f_move_speed = -pBot->f_max_speed;
	if (pBot->TeleportDir & dir_left)
		pBot->f_strafe_speed = -pBot->f_max_speed;
	if (pBot->TeleportDir & dir_right)
		pBot->f_strafe_speed = pBot->f_max_speed;
	if (pBot->TeleportDir & dir_up)
		pBot->f_updown_speed = pBot->f_max_speed;
	if (pBot->TeleportDir & dir_down)
		pBot->f_updown_speed = -pBot->f_max_speed;

	if (pBot->teleport_delay <= gpGlobals->time)
	{
		FakeClientCommand(pEdict, "teleport", NULL, NULL);
		TaskState = done;
	}
}

ETaskCode CTeleportTask::TaskIdentify()
{
	return TaskTeleport;
}

bool CTeleportTask::ShouldManageEnemys()
{
	return false;
}

// Block task
CBlockTask::CBlockTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;

	BlockTime = Delay + 1.0;
}

CBlockTask::~CBlockTask()
{
	FakeClientCommand(pEdict,"-block",NULL,NULL);
}

bool CBlockTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CBlockTask::TaskThink()
{
//	    ClientPrint(listenserver_edict,HUD_PRINTNOTIFY,"block\n");

	ResetMoveSpeed(pBot);

	if (((BlockTime < gpGlobals->time) && (pBot->pBotClient->BlockStruggles == false) &&
		(pBot->pBotClient->GetAutoPunched == false)) || (pBot->bot_ki < 8))
	{
		FakeClientCommand(pEdict, "-block", NULL, NULL);
		TaskState = done;
		return;
	}

	if (pBot->pBotClient->BlockStruggles)
	{
		if (pBot->bot_skill < 4)
		{
			ToggleTurbo(pEdict,1);
			pBot->turbo_time = gpGlobals->time + 1.5;
		}

		// Get the attack the bot is block struggling with
		if ((IsNull(pBot->pTouchEnt) == false) && (IsAttack(pBot->pTouchEnt)))
		{
//			lcout << "BS'ing with: " << STRING(pBot->pTouchEnt->v.classname) << "\n";

			if (IsNull(pBot->pTouchEnt->v.owner) == false)
			{
				float AttackY = pBot->pTouchEnt->v.angles.y;
				float AttackX = pBot->pTouchEnt->v.angles.x;

				Vector VecToAimAt = WrapAngles(UTIL_VecToAngles(pBot->pTouchEnt->v.owner->v.origin - pEdict->v.origin));

				float YDiff = fabs(UTIL_WrapAngle(AttackY + 180) - VecToAimAt.y);
				float XDiff = fabs(UTIL_WrapAngle(AttackX + 180) - VecToAimAt.x);

				if ((YDiff < 20) && (XDiff < 20))
				{
					pBot->AimVec = VecToAimAt;
				}
			}
		}
	}

	FakeClientCommand(pEdict, "+block", NULL, NULL);
}

ETaskCode CBlockTask::TaskIdentify()
{
	return TaskBlock;
}

bool CBlockTask::ShouldManageEnemys()
{
	return (!pBot->pBotClient->BlockStruggles);
}

// Kick enemy with melee task
CKickEnemyTask::CKickEnemyTask(bot_t *pB) : CBaseTask(pB)
{

	BotStatusChecks = CheckFly;
	BotSenseChecks = (CheckSounds | CheckAttacks);
	BotMiscChecks = 0;

	LifeTime = Delay + 1.5;
}

CKickEnemyTask::~CKickEnemyTask()
{
}

bool CKickEnemyTask::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel >= danger_medium)
		return true;

	return false;
}

void CKickEnemyTask::TaskThink()
{
	if (pBot->pBotEnemy)
	{
		ClientInfo_t *pClient = UTIL_GetClientPointer(pBot->pBotEnemy);
		if (!pClient)
		{
			TaskState = done;
			return;
		}

		if ((pClient->IsInMeleeMode) || (pClient->Ascends) || (LifeTime < gpGlobals->time) ||
			(pClient->Swoops))
		{
			TaskState = done;
			return;
		}

		float dist = UTIL_Distance(pEdict->v.origin, pBot->pBotEnemy->v.origin);

		if (dist < 30.0f)
		{
			SetMoveDir(pBot, dir_back);
		}
		else if (dist < 350.0f)
		{
			SetMoveDir(pBot, dir_forward);

			if (dist > 150.0f)
			{
				// Check if the bot isn't facing the enemy and the enemy isn't facing the bot
				if ((EntIsInFieldOfView(pEdict, pBot->pBotEnemy) == false) &&
					(EntIsInFieldOfView(pBot->pBotEnemy, pEdict) == false))
				{
#ifdef _DEBUG
					lcout << "Stop moving for aim\n";
#endif
					ResetMoveSpeed(pBot); // Give the bot some time to aim at his enemy.
				}
			}
		}
		else if (dist < 750.0f)
		{
			if ((pBot->bot_ki > 450) && (RANDOM_LONG(1, 100) <= pBot->TeleportPercent))
				FakeClientCommand(pEdict, "teleport", NULL, NULL);
			else
			{
				TaskState = done;
				return;
			}
		}
		else
		{
			// Enemy is to far away
			TaskState = done;
			return;
		}

		if ((pBot->current_weapon.iId != ESF_WEAPON_MELEE) && 
			(pBot->current_weapon.iId != ESF_WEAPON_SWORD))
		{
			if (pEdict->v.playerclass == 7)
			{
				UTIL_SelectItem(pEdict, "weapon_sword");
			}
			else
			{
				UTIL_SelectItem(pEdict, "weapon_melee");
			}
		}

		if ((pBot->bot_skill <= 2) && (pBot->bot_ki > 200))
		{
			ToggleTurbo(pEdict, 1);
			pBot->turbo_time = gpGlobals->time + 1.5;
		}
		pEdict->v.button |= IN_ATTACK2;
	}
	else
	{
		TaskState = done;
	}
}

ETaskCode CKickEnemyTask::TaskIdentify()
{
	return TaskKickEnemy;
}

bool CKickEnemyTask::ShouldManageEnemys()
{
	return true;
}

// Counter melee task
CCounterMeleeTask::CCounterMeleeTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;

	PunchKickTime = Delay + RANDOM_FLOAT(1.1, 2.4);
	ComboChargeButton = IN_ATTACK;
	NewKickPunchButtonDelay = 0.0f;
}

CCounterMeleeTask::~CCounterMeleeTask()
{
	pBot->f_shoot_time = AddShootDelay(pBot);
	pBot->pBotClient->DefendButtonsList.DeleteAllNodes();
	pBot->pBotClient->MaxComboChargeTime = 0.0f;
}

bool CCounterMeleeTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CCounterMeleeTask::TaskThink()
{
	if (pBot->pBotClient->IsInMeleeMode == false)
	{
		TaskState = done;
		return;
	}

	ResetMoveSpeed(pBot); // Othwerwise bots won't perform combo's properly...

	// Set this here so bots will hit a fire button when they are blocking aswell. If we
	// don't do this the bot doesn't retaliate.
	if (PunchKickTime <= gpGlobals->time)
	{
		if (RANDOM_LONG(0,1))
			ComboChargeButton = IN_ATTACK2;
		else
			ComboChargeButton = IN_ATTACK;

		PunchKickTime = gpGlobals->time + RANDOM_FLOAT(0.6, 2.3);
	}

	pEdict->v.button |= ComboChargeButton;

	if ((pEdict->v.velocity.Length() > 1.0) || (pBot->pBotClient->IsInMeleeMode == false))
	{
		TaskState = done;
		return;
	}

	// Is the bot attacking?
	if (pBot->pBotClient->Attacker)
	{
		if (NewKickPunchButtonDelay < gpGlobals->time)
		{
			TMultiChoice<float> ComboChargeTimeChoice;

			short WeakButtonValue = 50;
			short MediumButtonValue = 50;
			short StrongButtonValue = 50;

			float WeakButtonChargeTime = RANDOM_FLOAT(0.15f, pBot->pBotClient->WeakButtonChargeTime);
			float MediumButtonChargeTime = RANDOM_FLOAT(pBot->pBotClient->WeakButtonChargeTime,
														pBot->pBotClient->MediumButtonChargeTime);
			float StrongButtonChargeTime = RANDOM_FLOAT(pBot->pBotClient->MediumButtonChargeTime,
														pBot->pBotClient->StrongButtonChargeTime);

			if (pBot->bot_skill > 2)
			{
				// if low skill higher chance for weak buttons
				WeakButtonValue = 75;
				// and lower chance for medium buttons
				MediumButtonValue = 25;
			}
			else if ((pBot->pBotClient->MaxComboChargeTime > 0.0f) && 
				     ((pBot->pBotClient->MaxComboChargeTime - gpGlobals->time) < StrongButtonChargeTime))
			{
				// If about ending a combo the strong button chance should be higher
				StrongButtonValue = 150;
			}

			if (pBot->AggressionLevel == Defensive)
				StrongButtonValue *= 1.2;
			else if (pBot->AggressionLevel == Aggressive)
				WeakButtonValue *= 1.2;

			ComboChargeTimeChoice.Insert(WeakButtonChargeTime, WeakButtonValue);
			ComboChargeTimeChoice.Insert(MediumButtonChargeTime, MediumButtonValue);
			ComboChargeTimeChoice.Insert(StrongButtonChargeTime, StrongButtonValue);

			float ChargeTime = 0.0f;
			ComboChargeTimeChoice.GetSelection(ChargeTime);

			NewKickPunchButtonDelay = ChargeTime + gpGlobals->time;

			// Set a random charge button....
			short RandNr;

			RandNr = RANDOM_LONG(0, 2);

			if (RandNr == 0)
				SetMoveDir(pBot, dir_forward);
			else if (RandNr == 1)
				SetMoveDir(pBot, dir_back);

			if (RandNr == 2) // Didn't set a button yet?
				RandNr = RANDOM_LONG(0, 1);
			else
				RandNr = RANDOM_LONG(0, 2);
					  
			if (RandNr == 0)			  
				SetMoveDir(pBot, dir_left);
			else if (RandNr == 1)
				SetMoveDir(pBot, dir_right);
		}
	}
	else if (pBot->pBotClient->Defender) // the bot is/should be blocking
	{
		if (pBot->pBotClient->DefendButtonsList.Empty() == false)
		{
			EComboChargeButton ButtonToReactOn = pBot->pBotClient->DefendButtonsList.Pop();

			bool TakeARandomButton = (ButtonToReactOn == COMBO_BUTTON_RANDOM);

			if (!TakeARandomButton)
			{
				//byte Percent = (pBot->bot_skill + 1) * 15;
				if (RANDOM_LONG(1, 100) <= RandomComboButtonPercent[pBot->bot_skill])
					TakeARandomButton = true;
			}

			if (TakeARandomButton)
			{
				// Set a random combo button
				switch (RANDOM_LONG(1, 4))
				{
				case 1:
					ButtonToReactOn = COMBO_BUTTON_PUNCH_UP;
					break;
				case 2:
					ButtonToReactOn = COMBO_BUTTON_PUNCH_DOWN;
					break;
				case 3:
					ButtonToReactOn = COMBO_BUTTON_PUNCH_LEFT;
					break;
				case 4:
					ButtonToReactOn = COMBO_BUTTON_PUNCH_RIGHT;
					break;
				}
			}

			SetButtonsFromAComboChargeButton(pBot, ButtonToReactOn);
		}
	}/*
	else if ((!pBot->pBotClient->IsStunnedByMelee) && (!pBot->pBotClient->DoesComboStruggle))
	{
		// Just push forward, otherwise the bots never will retaliate :S
		//SetMoveDir(pBot, dir_forward);
	}*/
}

ETaskCode CCounterMeleeTask::TaskIdentify()
{
	return TaskCounterMelee;
}

bool CCounterMeleeTask::ShouldManageEnemys()
{
	return false;
}

// Bot combat task
CCombatTask::CCombatTask(bot_t *pB) : CBaseTask(pB)
{
#ifdef LOGBOTINFO
		clog << "Combat construction begin\n";
#endif

	BotStatusChecks = (CheckKi | CheckFly | CheckAscend | CheckHealth);
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = (CheckKameTorpLeftOvers | CheckForChatMsg | CheckDragonBall);

#ifdef LOGBOTINFO
		clog << "Combat construction end\n";
#endif
}

CCombatTask::~CCombatTask()
{
}

bool CCombatTask::KillTaskOnDanger(short DangerLevel)
{
	return (DangerLevel >= danger_medium);
}

void CCombatTask::TaskThink()
{
	CombatThink();
}

ETaskCode CCombatTask::TaskIdentify()
{
	return TaskCombat;
}

bool CCombatTask::ShouldManageEnemys()
{
	return true;
}

// Normal movement task
CNormalMovementTask::CNormalMovementTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = (CheckKi | CheckFly | CheckAscend | CheckHealth);
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = (CheckForEnemys | CheckForChatMsg | CheckDragonBall);
	if (follow_mode)
		BotMiscChecks |= CheckFollowTeamMate;
}

CNormalMovementTask::~CNormalMovementTask()
{
}

bool CNormalMovementTask::KillTaskOnDanger(short DangerLevel)
{
	return true;
}

void CNormalMovementTask::TaskThink()
{
    // Hunt an enemy ?
    if (CheckForEnemyTraces(pBot))
		; // Don't do anything here
	else
		BotNormalNavagation(pBot);
}

ETaskCode CNormalMovementTask::TaskIdentify()
{
	return TaskNormalMovement;
}

bool CNormalMovementTask::ShouldManageEnemys()
{
	return false;
}

// Swoop task
CSwoopTask::CSwoopTask(bot_t *pB) : CBaseTask(pB)
{
	Delay = gpGlobals->time; // No delay otherwise bot might swoop not at all sometimes :|

	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;

	CheckedSecondary = false;
	HoldSecondary = false;

	SelectSwoopType();
}

CSwoopTask::~CSwoopTask()
{
	pBot->SwoopButton = 0;
	pBot->BotHoldsSecondary = false;
	pBot->SwoopOffset = g_vecZero;
	pBot->SwoopType = SWOOP_NORMAL;
}

bool CSwoopTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CSwoopTask::TaskThink()
{
	if ((pBot->pBotClient->Swoops == false) || (!pBot->pBotEnemy))
	{
		TaskState = done;
		ResetMoveSpeed(pBot);
		return;
	}
	
	if ((pBot->MeleePercent == 0) || (b_botdontshoot) || (BotDontSearchEnemy))
	{
		TaskState = done;
		ResetMoveSpeed(pBot);
		if ((pBot->current_weapon.iId == ESF_WEAPON_MELEE) ||
			(pBot->current_weapon.iId == ESF_WEAPON_SWORD))
		{
			// Select to another weapon incase the bot hits his enemy so that the bot won't melee
			// his enemy 'by accident'(suggested by eekhoorn)
			UTIL_SelectItem(pEdict, "weapon_genericbeam");
		}
		return;
	}

	ClientInfo_t *pEnemyClient = UTIL_GetClientPointer(pBot->pBotEnemy);
	if (pEnemyClient)
	{
		// Can the enemy not be hit?
		if ((pEnemyClient->Powerstruggles) || (pEnemyClient->MeleeStruggles) ||
			(pEnemyClient->IsInMeleeMode) || (pEnemyClient->IsAutoPunching) ||
			(pEnemyClient->GetAutoPunched) || (pEnemyClient->Ascends) ||
			(pEnemyClient->IsStuckAgainstWall) || (pEnemyClient->LaysOnGround) ||
			(pEnemyClient->FlysAfterHit) || (pEnemyClient->ChargesWallJump))
		{
			ResetMoveSpeed(pBot);
			TaskState = done;
			return;
		}
	}

	float EnemyDist = UTIL_Distance(pEdict->v.origin,pBot->pBotEnemy->v.origin);

	if ((pBot->YawSpeed > 100.0f) && (EnemyDist > 700.0f))
	{
		TaskState = done; // Don't swoop when the bot has to turn much & enemy isn't close
		ResetMoveSpeed(pBot);
		return;
	}

	pEdict->v.button = IN_FORWARD;
	pBot->f_move_speed = pBot->f_max_speed;

	if ((pBot->current_weapon.iId != ESF_WEAPON_SWORD) && (pEdict->v.playerclass == 7))
		UTIL_SelectItem(pEdict, "weapon_sword");
	else if ((pBot->current_weapon.iId != ESF_WEAPON_MELEE) && (pEdict->v.playerclass != 7))
		UTIL_SelectItem(pEdict, "weapon_melee");

	if (CheckedSecondary == false)
		CheckSecondaryFire();

	if (HoldSecondary)
		pEdict->v.button |= IN_ATTACK2;

	if (pBot->SwoopType != SWOOP_NORMAL)
	{
		float MinDist, MaxDist;

		if (pBot->SwoopType == SWOOP_RANDOM_OFFSET)
		{
			MinDist = 200.0f;
			MaxDist = 1700.0f;
		}
		else
		{
			MinDist = 750.0f;
			MaxDist = 2200.0f;
		}

		if ((EnemyDist >= MinDist) && (EnemyDist <= MaxDist))
		{
			pBot->AimVec = WrapAngles(UTIL_VecToAngles(pBot->pBotEnemy->v.origin - pEdict->v.origin));

			if (pBot->SwoopType == SWOOP_RANDOM_OFFSET)
			{
#ifdef _DEBUG
				rcout << "Rand swoop offset\n";
#endif
				if (SwoopOffsetXFactor > 0)
				{
					if (pBot->SwoopOffset.x >= SwoopOffsetXTarget)
					{
						SwoopOffsetXTarget = RANDOM_LONG(-18, -16);
						SwoopOffsetXFactor = RANDOM_LONG(-2, -3);
					}
				}
				else
				{
					if (pBot->SwoopOffset.x <= SwoopOffsetXTarget)
					{
						SwoopOffsetXTarget = RANDOM_LONG(16, 18);
						SwoopOffsetXFactor = RANDOM_LONG(2, 3);
					}
				}

				if (SwoopOffsetYFactor > 0)
				{
					if (pBot->SwoopOffset.y >= SwoopOffsetYTarget)
					{
						SwoopOffsetYTarget = RANDOM_LONG(-18, -16);
						SwoopOffsetYFactor = RANDOM_LONG(-2, -3);
					}
				}
				else
				{
					if (pBot->SwoopOffset.y <= SwoopOffsetYTarget)
					{
						SwoopOffsetYTarget = RANDOM_LONG(16, 18);
						SwoopOffsetYFactor = RANDOM_LONG(2, 3);
					}
				}

				pBot->SwoopOffset.x += SwoopOffsetXFactor;
				pBot->SwoopOffset.y += SwoopOffsetYFactor;

				//pBot->SwoopOffsetUpdateDelay = gpGlobals->time + RANDOM_FLOAT(0.03, 0.06); 

				pBot->AimVec = WrapAngles(pBot->AimVec + pBot->SwoopOffset);
			}
			else
			{
#ifdef _DEBUG
				rcout << "Static swoop offset\n";
#endif
/*
				Vector2D vec2LOS;
				float flDot;

				UTIL_MakeVectors ( pEdict->v.angles );

				vec2LOS = (pBot->pBotEnemy->v.origin - pEdict->v.origin ).Make2D();
				vec2LOS = vec2LOS.Normalize();

				flDot = DotProduct (vec2LOS , gpGlobals->v_forward.Make2D() );

				if (flDot < 0.9)*/
					pBot->AimVec = WrapAngles(pBot->AimVec + pBot->SwoopOffset);
			}
		}
	}
	if (pBot->bot_skill == 0)
		AvoidAttackWhenSwoop(pBot);
}

ETaskCode CSwoopTask::TaskIdentify()
{
	return TaskSwoop;
}

bool CSwoopTask::ShouldManageEnemys()
{
	return true;
}

void CSwoopTask::CheckSecondaryFire()
{
	CheckedSecondary = true;

	if (CVAR_GET_FLOAT("mp_simplemelee") == 1.0f)
	{
		HoldSecondary = true; // Always hold secondary fire
		return;
	}

	short SecondaryPercent = BotPersonalities[pBot->Personality].meleekick_percent;

	// Is the enemy low on health and the bot smart?
	if ((pBot->pBotEnemy->v.health < 5) && (pBot->bot_skill <= 2))
		SecondaryPercent *= 1.5f;

	if (pBot->bot_ki < 300)
		SecondaryPercent *= 1.2f; // Prepunching takes more ki...

	if (RANDOM_LONG(1, 100) <= SecondaryPercent)
		HoldSecondary = true;
}

void CSwoopTask::SelectSwoopType()
{
	if (!pBot->pBotEnemy)
		return;

	float dist = UTIL_Distance(pEdict->v.origin, pBot->pBotEnemy->v.origin);

	pBot->SwoopType = SWOOP_NORMAL;

	if ((RANDOM_LONG(1, 100) <= 70) && (pBot->bot_skill == 0) && (dist > 800.0))
	{
		if (RANDOM_LONG(0, 1))
		{
			pBot->SwoopType = SWOOP_RANDOM_OFFSET;
			pBot->SwoopOffsetUpdateDelay = 0.0f;
			pBot->SwoopOffset = g_vecZero;

			if (RANDOM_LONG(0, 1))
			{
				SwoopOffsetXTarget = RANDOM_LONG(16, 18);
				SwoopOffsetXFactor = RANDOM_LONG(2, 3);
			}
			else
			{
				SwoopOffsetXTarget = RANDOM_LONG(-18, -16);
				SwoopOffsetXFactor = RANDOM_LONG(-3, -2);
			}

			if (RANDOM_LONG(0, 1))
			{
				SwoopOffsetYTarget = RANDOM_LONG(16, 18);
				SwoopOffsetYFactor = RANDOM_LONG(2, 3);
			}
			else
			{
				SwoopOffsetYTarget = RANDOM_LONG(-18, -16);
				SwoopOffsetYFactor = RANDOM_LONG(-3, -2);
			}
		}
		else
		{
			TMultiChoice<Vector> SwoopOffsetChoice;
			Vector OffsetVec = g_vecZero;

			if (CheckForCollide(pEdict, dir_left, 200.0f) == false)
				SwoopOffsetChoice.Insert(Vector(0, 20, 0), 50);
			if (CheckForCollide(pEdict, dir_right, 200.0f) == false)
				SwoopOffsetChoice.Insert(Vector(0, -20, 0), 50);
			if (CheckForCollide(pEdict, dir_up, 200.0f) == false)
				SwoopOffsetChoice.Insert(Vector(20, 0, 0), 50);
			if (CheckForCollide(pEdict, dir_down, 200.0f) == false)
				SwoopOffsetChoice.Insert(Vector(-20, 0, 0), 50);

			if (SwoopOffsetChoice.GetSelection(OffsetVec))
			{
				pBot->SwoopOffset = OffsetVec;
				pBot->SwoopType = SWOOP_OFFSET;
			}
		}
	}
}


// Power struggle task
CPowerStruggleTask::CPowerStruggleTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = CheckForChatMsg;
}

CPowerStruggleTask::~CPowerStruggleTask()
{
}

bool CPowerStruggleTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CPowerStruggleTask::TaskThink()
{
	if (pBot->bot_powerstrug < 127)
	{
//		ClientPrint(listenserver_edict,HUD_PRINTNOTIFY,"ps\n");
				
		if ((pBot->bot_skill < 3) && (pBot->bot_ki > 2))
		{
			pEdict->v.button |= IN_ATTACK;

			if (pBot->pBotClient->FlysAfterHit)
				pEdict->v.button |= IN_JUMP;
			else
			{
				ToggleTurbo(pEdict,1);
				pBot->turbo_time = gpGlobals->time + 1.5;
			}
		}
		else
			if (RANDOM_LONG(1, 100) < 50)
				pEdict->v.button |= IN_ATTACK;
	}
	else
	{
		TaskState = done;
	}
}

ETaskCode CPowerStruggleTask::TaskIdentify()
{
	return TaskPowerstruggle;
}

bool CPowerStruggleTask::ShouldManageEnemys()
{
	return false;
}

// Finish melee task
CFinishMeleeTask::CFinishMeleeTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;

//	Delay = gpGlobals->time + RANDOM_FLOAT(1.0, 1.4);

	short MeleeComboPercent = BotPersonalities[pBot->Personality].meleecombo_percent;

	if (pBot->bot_ki < 400) // Is the bot low on ki?
	{
		float KiFactor = (pBot->bot_ki * 0.001f) * 2.0f;
		MeleeComboPercent *= (2.0f - KiFactor);
	}

	if (RANDOM_LONG(1, 100) <= MeleeComboPercent)
		FinishMeleeType = 1;
	else
		FinishMeleeType = 2;

	if (FinishMeleeType == 1)
		FinishMeleeTime = Delay + RANDOM_FLOAT(4.3, 6.1);
	else
		FinishMeleeTime = Delay + RANDOM_FLOAT(3.8, 4.6);

	PunchKickTime = Delay + RANDOM_FLOAT(1.5, 2.4);
	ComboChargeButton = IN_ATTACK;
	NewKickPunchButtonDelay = 0.0;

	byte MaxStunpunches = static_cast<byte>(CVAR_GET_FLOAT("am_prepunch_maxstun"));
	byte MinPrepunches = static_cast<byte>(CVAR_GET_FLOAT("am_prepunch_mincombo"));
	byte MaxPrepunches = static_cast<byte>(CVAR_GET_FLOAT("am_prepunch_max")) - 1;

	byte MinPunch = MaxStunpunches - (RANDOM_LONG(0, ((pBot->bot_skill+1)/2)));
	byte MaxPunch = MaxPrepunches - (RANDOM_LONG(0, pBot->bot_skill)) - byte(pBot->AggressionLevel);

	if (MinPunch <= MinPrepunches)
		MinPunch = MinPrepunches + 1;

	if (MaxPunch <= MinPunch)
		MaxPunch = MinPrepunches + 1;

	PrepunchAmountGoal = RANDOM_LONG(MinPunch, MaxPunch);

	pPlayerToThrowAt = NULL;
	ObstacleToThrowAt = g_vecZero;
	CheckedObstacle = false;
}

CFinishMeleeTask::~CFinishMeleeTask()
{
	pBot->f_shoot_time = AddShootDelay(pBot);
}

bool CFinishMeleeTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CFinishMeleeTask::TaskThink()
{
	if ((pBot->pBotClient->IsInMeleeMode == false) &&
		(pBot->pBotClient->IsAutoPunching == false))
	{
		TaskState = done;
		return;
	}

	ResetMoveSpeed(pBot);

	if (pBot->pBotClient->IsAutoPunching)
	{
		if ((pBot->pBotClient->PrepunchHits < PrepunchAmountGoal) && (pBot->bot_ki > 150))
			return; // Not enough prepunches yet, continue prepunching
	}

	// Check if the bot is an attacker aswell; since if your enemy blocked the prepunches
	// you can't throw him anymore but you will charge a combo started with kick.
	if ((FinishMeleeType == 1) || (pBot->pBotClient->Attacker))
	{
		if ((pBot->pBotClient->MaxComboChargeTime >= gpGlobals->time) ||
			(FinishMeleeTime >= gpGlobals->time))
		{
			if (PunchKickTime <= gpGlobals->time)
			{
				if (RANDOM_LONG(0,1))
					ComboChargeButton = IN_ATTACK2;
				else
					ComboChargeButton = IN_ATTACK;

				PunchKickTime = gpGlobals->time + RANDOM_FLOAT(1.3, 2.5);
			}

			pEdict->v.button = ComboChargeButton;
/*
			if (SwitchComboButtonDelay <= gpGlobals->time)
			{
				pBot->f_move_speed = pBot->f_strafe_speed = pBot->f_updown_speed = 0.0;

				short RandNr;

				short offset = (pBot->bot_skill * 15);
				RandNr = RANDOM_LONG(0, 50 + offset);

				if (RandNr == 0)
					SetMoveDir(pBot, dir_forward);
				else if (RandNr == 1)
					SetMoveDir(pBot, dir_back);

				RandNr = RANDOM_LONG(0, 50 + offset);
					  
				if (RandNr == 0)			  
					SetMoveDir(pBot, dir_left);
				else if (RandNr == 1)
					SetMoveDir(pBot, dir_right);
			}
*/

			if (NewKickPunchButtonDelay < gpGlobals->time)
			{
				TMultiChoice<float> ComboChargeTimeChoice;

				short WeakButtonValue = 50;
				short MediumButtonValue = 50;
				short StrongButtonValue = 50;

				float WeakButtonChargeTime = RANDOM_FLOAT(0.15f, pBot->pBotClient->WeakButtonChargeTime);
				float MediumButtonChargeTime = RANDOM_FLOAT(pBot->pBotClient->WeakButtonChargeTime,
															pBot->pBotClient->MediumButtonChargeTime);
				float StrongButtonChargeTime = RANDOM_FLOAT(pBot->pBotClient->MediumButtonChargeTime,
															pBot->pBotClient->StrongButtonChargeTime);

				if (pBot->bot_skill > 2)
				{
					// if low skill higher chance for weak buttons
					WeakButtonValue = 75;
					// and lower chance for medium buttons
					MediumButtonValue = 25;
				}
				else if ((pBot->pBotClient->MaxComboChargeTime > 0.0f) && 
					     ((pBot->pBotClient->MaxComboChargeTime - gpGlobals->time) < StrongButtonChargeTime))
				{
					// If about ending a combo the strong button chance should be higher
					StrongButtonValue = 150;
				}

				if (pBot->AggressionLevel == Defensive)
					StrongButtonValue *= 1.2;
				else if (pBot->AggressionLevel == Aggressive)
					WeakButtonValue *= 1.2;

				ComboChargeTimeChoice.Insert(WeakButtonChargeTime, WeakButtonValue);
				ComboChargeTimeChoice.Insert(MediumButtonChargeTime, MediumButtonValue);
				ComboChargeTimeChoice.Insert(StrongButtonChargeTime, StrongButtonValue);

				float ChargeTime = 0.0f;
				ComboChargeTimeChoice.GetSelection(ChargeTime);

				NewKickPunchButtonDelay = ChargeTime + gpGlobals->time;

				// Set a random charge button....
				short RandNr;

				RandNr = RANDOM_LONG(0, 2);

				if (RandNr == 0)
					SetMoveDir(pBot, dir_forward);
				else if (RandNr == 1)
					SetMoveDir(pBot, dir_back);

				if (RandNr == 2) // Didn't set a button yet?
					RandNr = RANDOM_LONG(0, 1); // Set this so the bot atleast will set 1 button
				else
					RandNr = RANDOM_LONG(0, 2);
					  
				if (RandNr == 0)			  
					SetMoveDir(pBot, dir_left);
				else if (RandNr == 1)
					SetMoveDir(pBot, dir_right);
			}
		}
		else
		{
			TaskState = done;
			return;
		}
	}
	else
	{
		// Time to start looking for a throw place/victim?
		if (pBot->ChargeBeamVal >= 30)
		{
			// Is the choosen victim not valid anymore?
			if ((UTIL_IsValidEnemy(pBot, pPlayerToThrowAt) == false) ||
				(IsValidThrowVictim(pPlayerToThrowAt) == false))
				pPlayerToThrowAt = NULL; // Reset

			// Didnt found a player yet?
			if (!pPlayerToThrowAt)
				CheckPlayerToThrowAt();

			// Din't checked for an obstacle and have no player to throw at?
			if ((CheckedObstacle == false) && (!pPlayerToThrowAt))
			{
				CheckObstacleToThrowAt();
			}

			if (pPlayerToThrowAt)
			{
				pBot->AimVec = WrapAngles(UTIL_VecToAngles(BotBodyTarget(pPlayerToThrowAt, pBot) - pEdict->v.origin));
			}
			else // else aim to the obstacle to throw at
				pBot->AimVec = WrapAngles(UTIL_VecToAngles(ObstacleToThrowAt - pEdict->v.origin));
		}

		if (pBot->ChargeBeamVal == 100)
		{
			TaskState = done;
			return;
		}

		if (((FinishMeleeTime - gpGlobals->time) < 1.7) && (pBot->ChargeBeamVal == 0))
		{
			TaskState = done;
			return;
		}

		pEdict->v.button = IN_ATTACK2;
	}
}

ETaskCode CFinishMeleeTask::TaskIdentify()
{
	return TaskFinishMelee;
}

bool CFinishMeleeTask::ShouldManageEnemys()
{
	return false;
}

void CFinishMeleeTask::CheckPlayerToThrowAt()
{
	float distance = 0.0f;
	float nearestdistance = 9999.0f;
	byte i;

	Vector v_src;
	Vector v_dest, v_dest2;
	Vector vecEnd;
	Vector IdealThrowVec = g_vecZero;

	edict_t *pPlayer = NULL;
	edict_t *pThrowPlayer = NULL;
	
	referential_t referential;
	TraceResult tr, tr2;

	bool OnScouter = false, InViewCone = false;

	if (!pBot->MadePlayerList)
		MakePlayerList(pBot);

	if (pBot->PlayerCount)
	{
		for (i = 0;i<pBot->PlayerCount;i++)
		{
			pPlayer = pBot->PlayerList[i];

			if (IsNull(pPlayer))
				continue;

			distance = UTIL_Distance(pEdict->v.origin,pPlayer->v.origin);

			if ((distance > 1500.0) || (distance >= nearestdistance))
				continue;

		   	// is team play enabled?
			if (IsTeamPlay())
			{
   				int player_team = UTIL_GetTeam(pPlayer);
				int bot_team = UTIL_GetTeam(pEdict);

				// don't target your teammates...
				if (bot_team == player_team)
				{
					continue;
				}	
			}

			vecEnd = pPlayer->v.origin + pPlayer->v.view_ofs;

			OnScouter = ((distance <= 2300) || (pBot->SBotType > 0));
			InViewCone = (FInViewCone(&vecEnd, pEdict));
		
			if ((!InViewCone) && (!OnScouter))
				continue;

			if (IsValidThrowVictim(pPlayer) == false)
				continue;

			nearestdistance = distance;
			pThrowPlayer = pPlayer;
		}
		if (pThrowPlayer)
		{
			pPlayerToThrowAt = pThrowPlayer;
			return;
		}
	}
}

void CFinishMeleeTask::CheckObstacleToThrowAt()
{
	float x=0.0,y=0.0;
	float distance = 0.0, distance2 = 0.0;
	float nearestdistance = 9999.0;
	float dist = 0.0;
	byte i;

	Vector v_src;
	Vector v_dest, v_dest2;
	Vector vecEnd;
	Vector IdealThrowVec = g_vecZero;

	referential_t referential;
	TraceResult tr, tr2;

	for (i=0;i<5;i++)
	{ 
		switch (i)
		{
		case 0:
			x=0.0;
			break;
		case 1:
			x=-40.0;
			break;
		case 2:
			x=40.0;
			break;
		case 3:
			x=-89.0;
			break;
		case 4:
			x=89.0;
			break;
		default:
			break;
		}

		for (int j=0;j<5;j++)
		{
			switch (j)
			{
			case 0:
				y=-90.0;
				break;
			case 1:
				y=-45.0;
				break;
			case 2:
				y=0.0;
				break;
			case 3:
				y=45.0;
				break;
			case 4:
				y=90.0;
				break;
			default:
				break;
			}

			v_src = WrapAngles(pEdict->v.v_angle); // restore bot's current v_angle
			v_src.y = UTIL_WrapAngle (v_src.y + y);
			v_src.x = x;

			referential = ReferentialOfAngles (v_src);

			v_dest = pEdict->v.origin + referential.v_forward * 3000;

			v_dest2 = pEdict->v.origin - referential.v_forward * 3000;

	        UTIL_TraceHull (pEdict->v.origin, v_dest,
		                    ignore_monsters, human_hull, pEdict, &tr);

	        UTIL_TraceHull (pEdict->v.origin, v_dest2,
		                    ignore_monsters, human_hull, pEdict, &tr2);

			distance = (tr.vecEndPos - pEdict->v.origin).Length();
			distance2 = (tr2.vecEndPos - pEdict->v.origin).Length();

			if ((distance < distance2) && (tr.flFraction > 0.0f) && (distance > 125.0f))
			{
				dist = distance;
			}
			else if ((tr2.flFraction > 0.0f) && (distance2 > 125.0f))
			{
				dist = distance2;
				v_dest = v_dest2;
			}
			else
				continue;

			if (dist < nearestdistance)
			{
				nearestdistance = dist;
				IdealThrowVec = v_dest;
			}
		}
	}

	if (IdealThrowVec != g_vecZero) // Found a spot to throw to?
	{
		ObstacleToThrowAt = IdealThrowVec;
	}

	CheckedObstacle = true;
}

bool CFinishMeleeTask::IsValidThrowVictim(edict_t *pPlayer)
{
	if (pPlayer->v.velocity.Length() > 600.0f)
		return false;

	if (UTIL_Distance(pEdict->v.origin, pPlayer->v.origin) > 1500.0f)
		return false;

	ClientInfo_t *pClient = UTIL_GetClientPointer(pPlayer);
	if (!pClient)
		return false;

	if (pClient->Ascends)
		return false;

	if (pClient->IsInMeleeMode)
		return false;

	if (pClient->BlockStruggles)
		return false;

	if (pClient->GetAutoPunched)
		return false;

	if (pClient->BeamJumps)
		return false;

	if (pClient->FlysAfterHit)
		return false;

	if (pClient->IsAutoPunching)
		return false;

	if (pClient->MeleeStruggles)
		return false;

	if (pClient->Powerstruggles)
		return false;

	return true;
}

// Ascend task (called while a bot is ascending)
CAscendTask::CAscendTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckHit);
	BotMiscChecks = CheckForChatMsg;
}

CAscendTask::~CAscendTask()
{
}

bool CAscendTask::KillTaskOnDanger(short DangerLevel)
{
	if (pBot->DangerType == DangerEnemy)
		return false;

	if (DangerLevel < danger_hard)
		return false;

	float MaxTimeLeftToStop = 2.0;

	if (pBot->AggressionLevel == Defensive)
		MaxTimeLeftToStop *= 1.25;
	else if (pBot->AggressionLevel == Aggressive)
		MaxTimeLeftToStop *= 0.75;

	if ((DangerLevel >= danger_hard) && ((pBot->AscendTime - gpGlobals->time) > MaxTimeLeftToStop))
	{
		FakeClientCommand(pEdict,"descend",NULL,NULL);
		pBot->LastAscendCheck = gpGlobals->time + 7.75f; // Don't check for ascending for a while
		return true;
	}

	return false;
}

void CAscendTask::TaskThink()
{
	if (pBot->AscendTime < gpGlobals->time)
	{
		TaskState = done;
		return;
	}
	pBot->f_move_speed = pBot->f_strafe_speed = pBot->f_updown_speed = 0.0;
}

ETaskCode CAscendTask::TaskIdentify()
{
	return TaskAscend;
}

bool CAscendTask::ShouldManageEnemys()
{
	return false;
}

// Blinded task
CBlindedTask::CBlindedTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;
}

CBlindedTask::~CBlindedTask()
{
}

bool CBlindedTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CBlindedTask::TaskThink()
{
    if (pBot->blinded_time < gpGlobals->time)
	{
		TaskState = done;
		return;
	}

    pBot->f_move_speed = -pBot->f_max_speed;

    switch (RANDOM_LONG(0, 2))
	{
    case 0:
        pBot->f_strafe_speed = -pBot->f_max_speed;
        break;
    case 1:
        pBot->f_strafe_speed = pBot->f_max_speed;
        break;
	}

    if ((RANDOM_LONG(1, 1000) < 40) && (RANDOM_LONG(1, 100) < pBot->TeleportPercent) &&
        (pBot->bot_skill < 3))
        FakeClientCommand(pEdict,"teleport",NULL,NULL);
}

ETaskCode CBlindedTask::TaskIdentify()
{
	return TaskBlinded;
}

bool CBlindedTask::ShouldManageEnemys()
{
	return false;
}

// Combo task
CComboTask::CComboTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = CheckKameTorpLeftOvers;
}

CComboTask::~CComboTask()
{
	if (pBot->pBotCombo)
		DeleteCombos(pBot); // just in case
}

bool CComboTask::KillTaskOnDanger(short DangerLevel)
{
	// if bot wants to react, then he does it in the combo code 
	return false;
}

void CComboTask::TaskThink()
{
	if (pBot->pBotCombo)
	{
		if (pBot->pBotCombo->state == done)
		{
			pBot->TaskAfterTask = pBot->pBotCombo->FinishingTask;

			DeleteCombos(pBot);
			TaskState = done;
			return;
		}

		if (pBot->pBotCombo->Activated == false)
			pBot->pBotCombo->Activate();
		pBot->pBotCombo->DoCombo();
	}
	else
	{
		TaskState = done;
	}
}

ETaskCode CComboTask::TaskIdentify()
{
	return TaskCombo;
}

bool CComboTask::ShouldManageEnemys()
{
	return true;
}

// Charge attack task
CChargeAttackTask::CChargeAttackTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = 0;

	if (pBot->f_primary_charging > 0.0)
		pBot->f_primary_charging += Delay;
	if (pBot->f_secondary_charging > 0.0)
		pBot->f_secondary_charging += Delay;

	EnemyMovingFastTime = 0.0f;
	EnemyMovingSlowTime = Delay;
}

CChargeAttackTask::~CChargeAttackTask()
{
	ResetChargeAttacks(pBot);
}

bool CChargeAttackTask::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel >= danger_medium)
	{
		if (RANDOM_LONG(1, 100) <= pBot->TeleportPercent)
		{
			bool GoTeleport = false;

			if (pBot->ChargeBeamVal > 0)
			{
				if (pEdict->v.playerclass == 2)
				{
					char model[32] = "";
					UTIL_GetModel(pEdict, model);
					if (FStrEq(model, "ssjgoku"))
						GoTeleport = true;
				}
			}
			else
				GoTeleport = true;

			if (GoTeleport)
			{
				FakeClientCommand(pEdict, "teleport", NULL, NULL);
				return false;
			}
		}
		
		if ((DangerLevel == danger_medium) && (BotShouldFallDown(pBot)))
		{
			if (pEdict->v.gravity < 0.1)
				FakeClientCommand(pEdict, "togglefly", NULL, NULL);

			pBot->ShouldFly = false;
			pBot->CheckFlyDelay = gpGlobals->time + RANDOM_FLOAT(0.8, 1.2);
			return false;
		}
		
		if (pBot->ChargeBeamVal >= 33)
		{
			if ((AttackIsBeamJumpable(pBot->charging_weapon_id, (pBot->f_secondary_charging > 0.0))) &&
				(BotCanBeamJump(pBot, false)))
			{
				pBot->TaskAfterTask = TaskBeamJump;
				TaskState = done;
				return false;
			}

			if (pBot->DangerType != DangerAttack)
				return true;

			if (pBot->pBotEnemy)
			{
//				edict_t *pEnemyAttack = GetAttackFromPlayer(pBot->pBotEnemy);

				edict_t *pEnemyAttack = pAttackManager->SearchLastAttack(pBot->pBotEnemy);

				if (pEnemyAttack)
				{
					if (EntityHeadsToPlayer(pEdict, pEnemyAttack))
					{
						// stop charging...

						// set next time to shoot
						pBot->f_shoot_time = AddShootDelay(pBot);

						if (RANDOM_LONG(1, 100) <= 
							BotPersonalities[pBot->Personality].beamexplode_percent)
							pBot->LetBeamExplode = true;
						else
							pBot->LetBeamExplode = false;

						pBot->TaskAfterTask = TaskControlBeam;
					}

					TaskState = done; 
					return false; // ....but don't avoid the danger
				}
			}
			
			// Charging attack is a beam?
			if (AttackIsBeam(pBot->charging_weapon_id, (pBot->f_secondary_charging > 0.0)))
			{
				// Switch to melee or sword, otherwise the attack will be releashed
				if (pEdict->v.playerclass == 7)
					UTIL_SelectItem(pEdict, "weapon_sword");
				else
					UTIL_SelectItem(pEdict, "weapon_melee");
			}
		}
		return true;
	}
	return false;
}

void CChargeAttackTask::TaskThink()
{
	if ((pBot->pBotEnemy) || (pBot->HasAnEnemyTime > (gpGlobals->time - 2.0)))
	{
		bool is_primary_charging = (pBot->f_primary_charging > 0.0);
		bool is_secondary_charging = (pBot->f_secondary_charging > 0.0);

		if ((is_primary_charging) && (is_secondary_charging)) // weird bug
		{
			pBot->f_secondary_charging = -1;
			is_secondary_charging = false;
		}

		if ((is_primary_charging) || (is_secondary_charging))
		{
			int iId = pBot->charging_weapon_id;

			if ((iId == ESF_WEAPON_MASENKO) || (iId == ESF_WEAPON_POWERBEAM) ||
				(iId == ESF_WEAPON_SPIRIT_BOMB) || (iId == ESF_WEAPON_FINAL_FLASH) ||
				(iId == ESF_WEAPON_BIGBANG) || (iId == ESF_WEAPON_FINISHING_BUSTER) ||
				(iId == ESF_WEAPON_SCATTERBEAM))	
			{
				pBot->f_move_speed = 0.0;  // don't move back or forward
				pBot->f_strafe_speed = 0.0; // don't strafe
				pBot->f_updown_speed = 0.0; // don't move up or down
			}
			else if (pBot->pBotEnemy == NULL)
				GoToAnPreviousLocation();

			float EnemyDist = 9999.9;

			if (pBot->pBotEnemy)
			{
				EnemyDist = UTIL_Distance(pEdict->v.origin,pBot->pBotEnemy->v.origin);
		
				if (pBot->pBotEnemy->v.velocity.Length() > 550.0f)
					EnemyMovingFastTime = gpGlobals->time;
				else
					EnemyMovingSlowTime = gpGlobals->time;
			}

			bool ReleaseFire = false;
			ClientInfo_t *pClient = NULL;

			if ((pBot->f_primary_charging <= gpGlobals->time) && (is_primary_charging))
				ReleaseFire = true;

			if ((pBot->f_secondary_charging <= gpGlobals->time) && (is_secondary_charging))
				ReleaseFire = true;

			if (pBot->ChargeBeamVal == 100)
				ReleaseFire = true;

			if (pBot->bot_ki < 5)
				ReleaseFire = true;
			else
			{
				// Is the enemy moving fast?(but not for a very long while(bots don't want to wait to long...))
				if ((EnemyMovingFastTime > (gpGlobals->time - 0.5f)) &&
					(EnemyMovingSlowTime > (gpGlobals->time - 4.0f)))
				{
					// Also the bot doesn't head 'very much' to the bot?
					if (UTIL_GetHeadingPercentage(pEdict, pBot->pBotEnemy) < 75)
						ReleaseFire = false; // Then wait with fire
				}
				else
				{
					pClient = UTIL_GetClientPointer(pBot->pBotEnemy);
					if (pClient)
					{
						if ((pClient->IsAutoPunching) || (pClient->IsInMeleeMode) || 
							(pClient->MeleeStruggles) || (pClient->Powerstruggles) ||
							(pClient->GetAutoPunched))
							ReleaseFire = false; // Wait untill the enemy can be hit again.

						// Is the enemy transforming and is the bot NOT a transkiller?
						if ((pClient->Ascends) &&
							(!BotPersonalities[pBot->Personality].transkiller))
							ReleaseFire = false; // Don't attack the enemy
					}
				}
			}

			// is it time to fire the charged weapon?
			if (ReleaseFire)
			{
				// set next time to shoot
				pBot->f_shoot_time = AddShootDelay(pBot);

				// Check first if the enemy is ascending and if the bot NOT a transkiller?
				if (!pClient) // Didn't checked already ? (Small optimization)
					pClient = UTIL_GetClientPointer(pBot->pBotEnemy);

				if (pClient)
				{
					if ((pClient->Ascends) && (!BotPersonalities[pBot->Personality].transkiller))
					{
						// Don't fire but just switch to another weapon(melee) so the attack will
						// 'disappear'
						if (pEdict->v.playerclass == 7) // Class == trunks?
							UTIL_SelectItem(pEdict, "weapon_sword"); // Trunks don't have melee
						else
							UTIL_SelectItem(pEdict, "weapon_sword");

						TaskState = done;
						return;
					}
				}

				if (pBot->charging_weapon_id == ESF_WEAPON_FRIEZADISC)
				{
					pBot->TaskAfterTask = TaskControlFriezaDisc;
				}
				else if (pBot->ChargeBeamVal > 0)
				{
					// Is the bot a beam lamer?
					if (RANDOM_LONG(1, 100) <= 
						BotPersonalities[pBot->Personality].beamexplode_percent)
						pBot->LetBeamExplode = true;
					else
						pBot->LetBeamExplode = false;

					if ((!pBot->pBotEnemy) && (BotCanBeamJump(pBot, false)) &&
						(AttackIsBeamJumpable(pBot->charging_weapon_id, is_secondary_charging)))
						pBot->TaskAfterTask = TaskBeamJump;
					else
						pBot->TaskAfterTask = TaskControlBeam;
				}

				TaskState = done;
			}
			else
			{
				if (is_primary_charging)
					pEdict->v.button |= IN_ATTACK;   // charge the weapon

				if (is_secondary_charging)
					pEdict->v.button |= IN_ATTACK2;  // charge the weapon

				pBot->f_shoot_time = gpGlobals->time;  // keep charging

				// If the bot has not much ki left and turbo is on...
				if ((pBot->bot_ki < 125) && (IsTurboOn(pEdict)))
					ToggleTurbo(pEdict, false); // ...toggle turbo off
				// Else if the bot is charging a beam, has a enough ki, is skilled enough
				// and just begun charging, turn turbo on.
				else if ((iId != ESF_WEAPON_DESTRUCTO_DISC) &&
						(iId != ESF_WEAPON_FRIEZADISC) && (iId != ESF_WEAPON_TELEKINESIS) &&
						(pBot->bot_skill < 2) && (pBot->bot_ki > 250))
				{
					if (pBot->ChargeBeamVal <= 10)
       					ToggleTurbo(pEdict, true);
				}

				pBot->turbo_time = gpGlobals->time + 1.5; // Keep turbo on(if on...)
			}
		}
		else
		{
			TaskState = done;
		}
	}
	else
	{
		TaskState = done;
	}
}

ETaskCode CChargeAttackTask::TaskIdentify()
{
	return TaskChargeAttack;
}

bool CChargeAttackTask::ShouldManageEnemys()
{
	return true;
}

// Shoot non-chargeble attack(like lasers) task
CShootNonChargebleTask::CShootNonChargebleTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = 0;

	if (pBot->f_primary_charging > 0.0)
		pBot->f_primary_charging += Delay;
	if (pBot->f_secondary_charging > 0.0)
		pBot->f_secondary_charging += Delay;
}

CShootNonChargebleTask::~CShootNonChargebleTask()
{
	ResetChargeAttacks(pBot);
}

bool CShootNonChargebleTask::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel >= danger_medium)
	{
		if (RANDOM_LONG(1, 100) <= pBot->TeleportPercent)
		{
			FakeClientCommand(pEdict, "teleport", NULL, NULL);
			return false;
		}
		
		if ((DangerLevel == danger_medium) && (BotShouldFallDown(pBot)))
		{
			if (pEdict->v.gravity < 0.1)
				FakeClientCommand(pEdict, "togglefly", NULL, NULL);

			pBot->ShouldFly = false;
			pBot->CheckFlyDelay = gpGlobals->time + RANDOM_FLOAT(0.8, 1.2);
			return false;
		}
		
		return true;
	}
	return false;
}

void CShootNonChargebleTask::TaskThink()
{
	if ((pBot->pBotEnemy) || (pBot->HasAnEnemyTime > (gpGlobals->time - 2.0)))
	{
		bool is_primary_charging = (pBot->f_primary_charging > 0.0);
		bool is_secondary_charging = (pBot->f_secondary_charging > 0.0);

		if ((is_primary_charging) && (is_secondary_charging)) // weird bug
		{
			pBot->f_secondary_charging = -1;
			is_secondary_charging = false;
		}

		if ((is_primary_charging) || (is_secondary_charging))
		{
			if (pBot->pBotEnemy == NULL)
				GoToAnPreviousLocation();

			float EnemyDist = 9999.9;

			if (pBot->pBotEnemy)
			{
				EnemyDist = UTIL_Distance(pEdict->v.origin,pBot->pBotEnemy->v.origin);
			}

			bool PressFire = true, EndTask = false;

			if ((pBot->f_primary_charging <= gpGlobals->time) && (is_primary_charging))
				EndTask = true;

			if ((pBot->f_secondary_charging <= gpGlobals->time) && (is_secondary_charging))
				EndTask = true;

			if (pBot->bot_ki < 5)
				EndTask = true;
			else
			{
				ClientInfo_t *pClient = UTIL_GetClientPointer(pBot->pBotEnemy);
				if (pClient)
				{
					if ((pClient->IsAutoPunching) || (pClient->IsInMeleeMode) || 
						(pClient->MeleeStruggles) || (pClient->Powerstruggles) ||
						(pClient->GetAutoPunched))
						PressFire = false; // Wait untill the enemy can be hit again.

					// Is the enemy transforming and is the bot NOT a transkiller?
					if ((pClient->Ascends) && (!BotPersonalities[pBot->Personality].transkiller))
						PressFire = false; // Don't attack the enemy
				}
			}

			// is it time to stop firing?
			if (EndTask)
			{ 
				// set next time to shoot
				pBot->f_shoot_time = AddShootDelay(pBot);
				TaskState = done;
			}
			else
			{
				if ((is_primary_charging) && (PressFire))
					pEdict->v.button |= IN_ATTACK;   // Fire the weapon

				if ((is_secondary_charging) && (PressFire))
					pEdict->v.button |= IN_ATTACK2;  // Fire the weapon
			}
		}
		else
		{
			TaskState = done;
		}
	}
	else
	{
		TaskState = done;
	}
}

ETaskCode CShootNonChargebleTask::TaskIdentify()
{
	return TaskShootNonChargeble;
}

bool CShootNonChargebleTask::ShouldManageEnemys()
{
	return true;
}

// Shoot scattershots task
CShootScatterShotsTask::CShootScatterShotsTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = 0;

	ShootTime = Delay + RANDOM_FLOAT(2.75, 3.45);
	StartTime = Delay;

	if ((pBot->bot_skill < 2) && (RANDOM_LONG(1, 100) < 20))
		UseScatterCheat = true;
	else
		UseScatterCheat = false;
}

CShootScatterShotsTask::~CShootScatterShotsTask()
{
	pBot->f_shoot_time = AddShootDelay(pBot);
	pEdict->v.button |= IN_ATTACK2;
}

bool CShootScatterShotsTask::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel >= danger_medium)
	{
		if (RANDOM_LONG(1, 100) <= pBot->TeleportPercent)
		{
			FakeClientCommand(pEdict, "teleport", NULL, NULL);
			return false;
		}
		else if ((DangerLevel == danger_medium) && (BotShouldFallDown(pBot)))
		{
			if (pEdict->v.gravity < 0.1)
				FakeClientCommand(pEdict, "togglefly", NULL, NULL);

			pBot->ShouldFly = false;
			pBot->CheckFlyDelay = gpGlobals->time + RANDOM_FLOAT(0.8, 1.2);
			return false;
		}
		return true;
	}
	return false;
}

void CShootScatterShotsTask::TaskThink()
{
	// Time to stop or not much ki left?
	if ((ShootTime < gpGlobals->time) || (pBot->bot_ki < 5))
	{
		TaskState = done;
		return;
	}

	if ((pBot->pBotEnemy) || (pBot->HasAnEnemyTime > (gpGlobals->time - 2.0)))
	{
//		lcout << "Firing scattershot\n";

		if (pBot->pBotEnemy == NULL)
		{
			GoToAnPreviousLocation();
			return;
		}

		bool PressFire = true;

		ClientInfo_t *pClient = UTIL_GetClientPointer(pBot->pBotEnemy);
		if (pClient)
		{
			if ((pClient->IsAutoPunching) || (pClient->IsInMeleeMode) || 
				(pClient->MeleeStruggles) || (pClient->Powerstruggles) ||
				(pClient->GetAutoPunched))
				PressFire = false; // Wait untill the enemy can be hit again.

			// Is the enemy transforming and is the bot NOT a transkiller?
			if ((pClient->Ascends) && (!BotPersonalities[pBot->Personality].transkiller))
				PressFire = false; // Don't attack the enemy
		}

		if (!PressFire)
			return; // Don't press attack

		pEdict->v.button |= IN_ATTACK;   // Fire another blast

		if (UseScatterCheat)
			pEdict->v.button |= IN_ATTACK2; // And home it immediately :D
		else if ((StartTime + 0.55f) <= gpGlobals->time)
		{
			float FurthestScatterShotDist = pAttackManager->GetFurthestAttackDistByName(pEdict,
																				  "scattershot");
			if (FurthestScatterShotDist != -1.0) // -1 means 'no attack found'
			{
				float EnemyDist = UTIL_Distance(pEdict->v.origin, pBot->pBotEnemy->v.origin);

				// Time to let all shots come together?
				if (EnemyDist < (FurthestScatterShotDist - (225.0 + (25 * (pBot->bot_skill + 1)))))
				{
//					lcout << "Binding scattershots\n";
					pEdict->v.button |= IN_ATTACK2;
					TaskState = done;
					return;
				}
			}
		}

		if (pBot->bot_skill < 2)
		{
			ToggleTurbo(pEdict, true);
			pBot->turbo_time = gpGlobals->time + 1.5;
		}
	}
	else
	{
		TaskState = done;
	}
}

ETaskCode CShootScatterShotsTask::TaskIdentify()
{
	return TaskShootScatterShots;
}

bool CShootScatterShotsTask::ShouldManageEnemys()
{
	return true;
}

// Control frieza disc task
CControlFriezaDiscTask::CControlFriezaDiscTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = CheckFly;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = 0;

	RealDiscTime = RANDOM_FLOAT(10.0, 22.0);
	ControlDiscTime = Delay + RealDiscTime;
	EnemyMovingFastTime = 0.0f;
	EnemyMovingSlowTime = Delay;
}

CControlFriezaDiscTask::~CControlFriezaDiscTask()
{
	UTIL_SelectItem(pEdict, "weapon_genericbeam"); // Make the disc go away...
}

bool CControlFriezaDiscTask::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel >= danger_medium)
		return true;

	return false;
}

void CControlFriezaDiscTask::TaskThink()
{
	if (ControlDiscTime < gpGlobals->time)
	{
		TaskState = done;
		return;
	}

	if (pBot->pBotEnemy == NULL)
	{
		TaskState = done;
		return;
	}

//	edict_t *pAttack = SearchFriezaDisc(pEdict);

	edict_t *pAttack = pAttackManager->SearchAttackByName(pEdict, "Frieza Disc");

	if (IsNull(pAttack))
	{
		if ((ControlDiscTime - gpGlobals->time) < (RealDiscTime - 1.0))
		{
			TaskState = done;
		}
		return;
	}

	if (pBot->pBotEnemy->v.velocity.Length() > 550.0f)
		EnemyMovingFastTime = gpGlobals->time;
	else
		EnemyMovingSlowTime = gpGlobals->time;

	float DiscDist = UTIL_Distance(pEdict->v.origin, pAttack->v.origin);
	float EnemyDist = UTIL_Distance(pEdict->v.origin, pBot->pBotEnemy->v.origin);
	float MinCallBackDist = DiscDist - ((float(pBot->bot_skill + 1) * 250.0) / 1.5);

	if ((MinCallBackDist > EnemyDist) && (pBot->f_max_speed == 0.0))
		pEdict->v.button |= IN_ATTACK2; // call the disc back

	// Is the disc back 'at home'?
	if (pAttack->v.velocity.Length() == 0.0)
	{
		bool ThrowDisc = true;

		// Is the enemy moving fast?(but not for a very long while(bots don't want to wait to long...))
		if ((EnemyMovingFastTime > (gpGlobals->time - 0.5f)) &&
			(EnemyMovingSlowTime > (gpGlobals->time - 4.0f)))
		{
			// Also the bot doesn't head 'very much' to the bot?
			if (UTIL_GetHeadingPercentage(pEdict, pBot->pBotEnemy) < 75)
				ThrowDisc = false; // Wait a bit with firing the disc
		}

		if (ThrowDisc)
		{
			ClientInfo_t *pClient = UTIL_GetClientPointer(pBot->pBotEnemy);

			if (pClient)
			{
				if ((pClient->IsAutoPunching) || (pClient->IsInMeleeMode) || 
					(pClient->MeleeStruggles) || (pClient->Powerstruggles) ||
					(pClient->GetAutoPunched))
					ThrowDisc = false; // Wait untill the enemy can be hurt again.

				// Is the enemy transforming and is the bot NOT a transkiller?
				if ((pClient->Ascends) && (!BotPersonalities[pBot->Personality].transkiller))
					ThrowDisc = false; // Don't attack the enemy
			}
		}

		if (ThrowDisc)
			pEdict->v.button |= IN_ATTACK; // throw the disc
	}
}

ETaskCode CControlFriezaDiscTask::TaskIdentify()
{
	return TaskControlFriezaDisc;
}

bool CControlFriezaDiscTask::ShouldManageEnemys()
{
	return true;
}

// Control beam task
CControlBeamTask::CControlBeamTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = 0;

	ControlBeamTime = Delay + 10.0;
	BotShouldAimAtAttack = false;
}

CControlBeamTask::~CControlBeamTask()
{
}

bool CControlBeamTask::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel < danger_medium)
		return false;

	if (BotCanBeamJump(pBot, true))
	{
		pBot->TaskAfterTask = TaskBeamJump;
		TaskState = done;
		return false;
	}

	if (pBot->DangerType != DangerAttack)
	{
		pEdict->v.button |= IN_ATTACK2;
		return true;
	}

//	edict_t *pAttack = GetAttackFromPlayer(pEdict);

	edict_t *pAttack = pAttackManager->SearchLastAttack(pEdict);

	if ((pAttack) && (pBot->pBotEnemy))
	{
//		edict_t *pEnemyAttack = GetAttackFromPlayer(pBot->pBotEnemy);
		edict_t *pEnemyAttack = pAttackManager->SearchLastAttack(pBot->pBotEnemy);

		Vector v_src = pEdict->v.origin; 
		Vector v_attack = pAttack->v.origin;

		float BotToAttackDist = UTIL_Distance(v_src, v_attack);
		float MinExplodeAttackDist = (pAttack->v.dmg * 10.0 - (100.0 * float(DangerLevel)));

		if (pEnemyAttack)
		{
			if (AttacksWillCollide(pAttack, pEnemyAttack))
			{
//				float BotToEnemyAttackDist = UTIL_Distance(v_src, pEnemyAttack->v.origin);

//				if ((BotToEnemyAttackDist+50.0) >= BotToAttackDist)
				pBot->AimVec = WrapAngles(UTIL_VecToAngles(pEnemyAttack->v.origin - pEdict->v.origin));
				BotShouldAimAtAttack = true;
				return false;
			}
		}

//		if ((AttacksWillCollide) && (pEnemyAttack))
//			rcout << "Attack " << STRING(pEnemyAttack->v.classname) << " will collide\n";

		if (BotToAttackDist > MinExplodeAttackDist) 
			//((AttacksWillCollide == false)))// || (DangerLevel == danger_heavy)))
		{
			pEdict->v.button |= IN_ATTACK2;
//			rcout << "No attack will collide\n";
			return true;
		}
	}
	else
		return true;

	return false;
}

void CControlBeamTask::TaskThink()
{
	pBot->turbo_time = gpGlobals->time + 1.5; // Keep turbo on(if on...)

	if ((pBot->f_max_speed == 0) || ((ControlBeamTime - gpGlobals->time) > 9.5))
	{
//		ClientPrint(listenserver_edict,HUD_PRINTNOTIFY,"cbeam\n");
		if (ControlBeamTime >= gpGlobals->time)
		{
			if (UTIL_IsValidEnemy(pBot, pBot->pBotEnemy))
			{
				Vector v_aim;

//				edict_t *pAttack = GetAttackFromPlayer(pEdict);
				edict_t *pAttack = pAttackManager->SearchLastAttack(pEdict);

				if (pAttack)
				{
					edict_t *pEnemyAttack = pAttackManager->SearchLastAttack(pBot->pBotEnemy);
					Vector v_src = pEdict->v.origin; 
					Vector v_attack = pAttack->v.origin;

					if ((pEnemyAttack) && ((IsBeamAttack(pEnemyAttack)) || 
						(IsNonBeamAttack(pEnemyAttack))))
					{
						// Aim at the enemies attack instead?
						if (AttacksWillCollide(pAttack, pEnemyAttack))
						{
							pBot->AimVec = WrapAngles(UTIL_VecToAngles(pEnemyAttack->v.origin - pEdict->v.origin));
							BotShouldAimAtAttack = true;
							return;
						}
					}

					float BotToAttackDist = (v_attack - v_src).Length();
					float EnemyToAttackDist = (v_attack - pBot->pBotEnemy->v.origin).Length();

					if ((pBot->LetBeamExplode) && 
						(EnemyToAttackDist < ExplodeBeamDistances[pBot->bot_skill][RANDOM_LONG(0, 1)]))
					{			
						pEdict->v.button |= IN_ATTACK2;
						TaskState = done;
						return;
					}

					float EnemyDist = UTIL_Distance(pEdict->v.origin, v_attack);
					// If the attack is much further away then the enemy...
					if ((BotToAttackDist - 500.0f) > EnemyDist)
					{
						// ... then explode the attack
						pEdict->v.button |= IN_ATTACK2;
						TaskState = done;
						return;
					}
						
					if ((BotToAttackDist - 100.0) < EnemyDist)
						v_aim =	BotBodyTarget(pBot->pBotEnemy,pBot) - pEdict->v.origin;
					else
						v_aim = pBot->pBotEnemy->v.origin - pEdict->v.origin;
				}
				else if ((ControlBeamTime - gpGlobals->time) < 9.0)
				{
					pEdict->v.button |= IN_ATTACK2;
					TaskState = done;
					return;
				}
				else
					v_aim =	BotBodyTarget(pBot->pBotEnemy,pBot) - pEdict->v.origin;

				pBot->AimVec = WrapAngles(UTIL_VecToAngles(v_aim));
			}
			else
			{
//				edict_t *pAttack = GetAttackFromPlayer(pEdict);
				edict_t *pAttack = pAttackManager->SearchLastAttack(pEdict);
				if (pAttack)
				{
					float MinExplodeAttackDist = pAttack->v.dmg * 10.0;

					float AttackDist = UTIL_Distance(pEdict->v.origin, pAttack->v.origin);

					if (AttackDist < MinExplodeAttackDist)
					{
						// DON'T explode the attack when its (still) near the bot...

						UTIL_MakeVectors(pAttack->v.angles);
						Vector v_src = pAttack->v.origin;
						Vector v_end = v_src + gpGlobals->v_forward * 800.0;
						TraceResult tr;

						UTIL_TraceHull(v_src, v_end, ignore_monsters, human_hull, pAttack, &tr);
						float TraceDist = tr.flFraction * 800.0;

						// Is the attack about to hit a wall ?
						if (TraceDist < 800.0)
						{
							// Beam jump ?
							if (BotCanBeamJump(pBot, true))
							{
								pBot->TaskAfterTask = TaskBeamJump;
								TaskState = done;
								return;
							}

							// Try to move it away
							if (RANDOM_LONG(0, 1) == 0)
							{
								// first check left then right

								if (CheckForCollide(pAttack, dir_left, 800.0) == false)
								{
									pBot->AimVec.y = 
										UTIL_WrapAngle(pBot->AimVec.y + RANDOM_FLOAT(40.0, 50.0));
									return;
								}
								else if (CheckForCollide(pAttack, dir_right, 800.0) == false)
								{
									pBot->AimVec.y = 
										UTIL_WrapAngle(pBot->AimVec.y - RANDOM_FLOAT(40.0, 50.0));
									return;
								}
							}
							else
							{
								// first check right then left

								if (CheckForCollide(pAttack, dir_right, 800.0) == false)
								{
									pBot->AimVec.y = 
										UTIL_WrapAngle(pBot->AimVec.y - RANDOM_FLOAT(40.0, 50.0));
									return;
								}
								else if (CheckForCollide(pAttack, dir_left, 800.0) == false)
								{
									pBot->AimVec.y = 
										UTIL_WrapAngle(pBot->AimVec.y + RANDOM_FLOAT(40.0, 50.0));
									return;
								}
							}

							// Still no solution found ? Then check up or down
							if (RANDOM_LONG(0, 1) == 0)
							{
								// first check down then up
								if (CheckForCollide(pAttack, dir_down, 800.0) == false)
								{
									pBot->AimVec.x = 
										UTIL_WrapAngle(pBot->AimVec.x + RANDOM_FLOAT(40.0, 50.0));
									return;
								}
								else if (CheckForCollide(pAttack, dir_up, 800.0) == false)
								{
									pBot->AimVec.x = 
										UTIL_WrapAngle(pBot->AimVec.x - RANDOM_FLOAT(40.0, 50.0));
									return;
								}
							}
							else
							{
								// first check up then down
								if (CheckForCollide(pAttack, dir_up, 800.0) == false)
								{
									pBot->AimVec.x = 
										UTIL_WrapAngle(pBot->AimVec.x - RANDOM_FLOAT(40.0, 50.0));
									return;
								}
								else if (CheckForCollide(pAttack, dir_down, 800.0) == false)
								{
									pBot->AimVec.x = 
										UTIL_WrapAngle(pBot->AimVec.x + RANDOM_FLOAT(40.0, 50.0));
									return;
								}
							}
						}
						return; 
					}
				}	
				pEdict->v.button |= IN_ATTACK2;
				TaskState = done;
				return;
			}
		}
		else
		{
			pEdict->v.button |= IN_ATTACK2;
			TaskState = done;
			return;
		}
	}
	else
	{
		TaskState = done;
		return;
	}
}

ETaskCode CControlBeamTask::TaskIdentify()
{
	return TaskControlBeam;
}

bool CControlBeamTask::ShouldManageEnemys()
{
	// Should the bot aim at the enemy's attack?
	// If so don't manage enemies(so bots won't aim at their enemy)
/*	if ((BotShouldAimAtAttack) && (UTIL_IsValidEnemy(pBot, pBot->pBotEnemy)))
	{
		// Reset, if the bot should aim at the attack again it will be set true again
		BotShouldAimAtAttack = false;
		return false;
	}
*/
	return false;
}

// Follow a team mate task
CFollowTeamMateTask::CFollowTeamMateTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = (CheckFly | CheckKi | CheckAscend | CheckHealth);
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = CheckForEnemys;
}

CFollowTeamMateTask::~CFollowTeamMateTask()
{
	pBot->pBotUser = NULL;
}

bool CFollowTeamMateTask::KillTaskOnDanger(short DangerLevel)
{
	return true;
}

void CFollowTeamMateTask::TaskThink()
{
	if ((pBot->pBotEnemy == NULL) && (pBot->pHuntingEdict == NULL) && (IsTeamPlay()) && 
	    (follow_mode))
	{
		pBot->pBotUser = BotFindUser(pBot);
		if (pBot->pBotUser == NULL)
			TaskState = done;
	}
	else
		TaskState = done;
}

ETaskCode CFollowTeamMateTask::TaskIdentify()
{
	return TaskFollowTeamMate;
}

bool CFollowTeamMateTask::ShouldManageEnemys()
{
	return false;
}

// Look around a bit task
CLookAroundTask::CLookAroundTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = (CheckFly | CheckKi | CheckAscend);
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = (CheckForEnemys | CheckForChatMsg | CheckDragonBall);

	if (follow_mode)
		BotMiscChecks |= CheckFollowTeamMate;

	Stage = 0;
	LookAroundDelay = 0.0;
}

CLookAroundTask::~CLookAroundTask()
{
	pBot->f_pause_time = gpGlobals->time + RANDOM_FLOAT(6.0, 12.0);
}

bool CLookAroundTask::KillTaskOnDanger(short DangerLevel)
{
	return true;
}

void CLookAroundTask::TaskThink()
{
	if (debug_mode)
		ClientPrint(listenserver_edict,HUD_PRINTNOTIFY,"Pause !\n");

	if (LookAroundDelay < gpGlobals->time)
	{
		if (debug_mode)
			ClientPrint(listenserver_edict,HUD_PRINTNOTIFY,"Pause\n");

		Stage++;
		if (Stage > 4)
		{
			float curdist = 0.0;
			float dist = 0.0;

			Vector DesiredSpot;

			for (int i=0;i<4;i++)
			{
				dist = UTIL_Distance(pEdict->v.origin, LookAroundSpots[i]);
				if (dist > curdist)
				{
					curdist = dist;
					DesiredSpot = LookAroundSpots[i];
				}
			}

			pBot->AimVec = WrapAngles(UTIL_VecToAngles(DesiredSpot - pEdict->v.origin));
			TaskState = done;
			return;
		}
		pBot->AimVec.y = UTIL_WrapAngle(pBot->AimVec.y+RANDOM_LONG(80, 100));

		UTIL_MakeVectors(pEdict->v.v_angle);

		Vector vecEnd = pEdict->v.origin + (gpGlobals->v_forward*10000);
        TraceResult tr;

		UTIL_TraceHull (pEdict->v.origin, vecEnd, ignore_monsters, human_hull, pEdict, &tr);
		
		LookAroundSpots[Stage-1] = tr.vecEndPos;

		LookAroundDelay = gpGlobals->time + RANDOM_FLOAT(0.4, 0.8);
		LookAroundDelay += (tr.flFraction*2);
	}
	pBot->f_move_speed = pBot->f_strafe_speed = pBot->f_updown_speed = 0.0;

	if (RANDOM_LONG(0,5) > 4)
		pBot->AimVec.x = UTIL_WrapAngle(pBot->AimVec.x+RANDOM_LONG(-10, 10));

	if (RANDOM_LONG(0,5) > 4)
		pBot->AimVec.y = UTIL_WrapAngle(pBot->AimVec.y+RANDOM_LONG(-10, 10));

}

ETaskCode CLookAroundTask::TaskIdentify()
{
	return TaskLookAround;
}

bool CLookAroundTask::ShouldManageEnemys()
{
	return false;
}

// Recover health task (for buu bots)
CRecoverHPTask::CRecoverHPTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = CheckForChatMsg;
}

CRecoverHPTask::~CRecoverHPTask()
{
}

bool CRecoverHPTask::KillTaskOnDanger(short DangerLevel)
{
	int MinHealth = 140;

	switch(DangerLevel)
	{
	case danger_light:
		MinHealth = 120;
		break;
	case danger_medium:
		MinHealth = 30;
		break;
	case danger_hard:
	case danger_heavy:
		MinHealth = 0;
		break;
	}

	MinHealth -= (100-((pBot->bot_ki/1000)*100));

	if (pEdict->v.health >= MinHealth)
		return true;

	return false;
}

void CRecoverHPTask::TaskThink()
{
	if ((pBot->bot_ki < 200) && (pEdict->v.health > 20))
	{
		TaskState = done;
		return;
	}

	if (pBot->bot_ki < 25)
	{
		TaskState = done;
		return;
	}

    // select this 'weapon' if it isn't already selected
    if (pBot->current_weapon.iId != ESF_WEAPON_REGENERATION)
       UTIL_SelectItem(pEdict, "weapon_regeneration");

	pEdict->v.button |= IN_ATTACK;
	pBot->f_move_speed = 0.0;
}

ETaskCode CRecoverHPTask::TaskIdentify()
{
	return TaskRecoverHP;
}

bool CRecoverHPTask::ShouldManageEnemys()
{
	return false;
}

// Move a direction for a while - task
CMoveDirTask::CMoveDirTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = CheckFly;
	BotSenseChecks = 0;
	BotMiscChecks = CheckKameTorpLeftOvers;
}

CMoveDirTask::~CMoveDirTask()
{
	pBot->MoveDir = 0;
}

bool CMoveDirTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CMoveDirTask::TaskThink()
{
	pBot->f_move_speed = 0.0;
	pBot->f_strafe_speed = 0.0;
	pBot->f_updown_speed = 0.0;

	if (pBot->MoveDir & dir_forward)
		pBot->f_move_speed = pBot->f_max_speed;
	if (pBot->MoveDir & dir_back)
		pBot->f_move_speed = -pBot->f_max_speed;
	if (pBot->MoveDir & dir_left)
		pBot->f_strafe_speed = -pBot->f_max_speed;
	if (pBot->MoveDir & dir_right)
		pBot->f_strafe_speed = pBot->f_max_speed;
	if (pBot->MoveDir & dir_up)
		pBot->f_updown_speed = pBot->f_max_speed/2;
	if (pBot->MoveDir & dir_down)
		pBot->f_updown_speed = -pBot->f_max_speed/2;

	if (pBot->MoveTime < gpGlobals->time)
		TaskState = done;
}

ETaskCode CMoveDirTask::TaskIdentify()
{
	return TaskMoveDir;
}

bool CMoveDirTask::ShouldManageEnemys()
{
	return false;
}

// Wall jump task
CWallJumpTask::CWallJumpTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckAttacks | CheckSurroundingPlayers | CheckSounds | CheckHit |
					  CheckForChatMsg);
	BotMiscChecks = CheckKameTorpLeftOvers;

	if ((pBot->pBotClient->WallJumps == false) && (pBot->pBotClient->ChargesWallJump == false))
		Delay = gpGlobals->time + 1.0;

	LifeTime = Delay + 3.0f;
	StartChargingTime = 0.0f;
}

CWallJumpTask::~CWallJumpTask()
{
}

bool CWallJumpTask::KillTaskOnDanger(short DangerLevel)
{
	if ((DangerLevel == danger_heavy) && (pBot->DangerType != DangerEnemy) &&
		(pBot->DangerType != DangerCurrentEnemy))
	{
		if ((pBot->pBotClient->WallJumps) || (pBot->pBotClient->ChargesWallJump))
		{
			if ((pBot->pBotClient->WallJumps) && (RANDOM_LONG(1, 100) <= pBot->TeleportPercent))
			{
				FakeClientCommand(pEdict, "teleport", NULL, NULL);
				return false;
			}
			//		lcout << "wjump danger\n";
			return true;
		}
	}
	return false;
}

void CWallJumpTask::TaskThink()
{
	if (CheckWallJump(pBot) == false)
	{
		TaskState = done;
//		lcout << "False wjump\n";
		return;
	}

	ResetMoveSpeed(pBot);
	if (pBot->pBotClient->ChargesWallJump)
	{
		if (StartChargingTime == 0.0f)
			StartChargingTime = gpGlobals->time;
	}

/*
	if ((pBot->pBotClient->WallJumps) || (pBot->f_max_speed > 0.0f) ||
		(pEdict->v.velocity.Length() > 0))
	{
//		lcout << "Holding forward\n";
		SetMoveDir(pBot, dir_forward);
	}
	else if (pBot->pBotClient->ChargesWallJump)
	{
		if (StartChargingTime == 0.0f)
			StartChargingTime = gpGlobals->time;
		else if (StartChargingTime > (gpGlobals->time - 0.5f))
			SetMoveDir(pBot, dir_back);
	}
	else
	{
//		lcout << "Holding back\n";
		SetMoveDir(pBot, dir_back);
	}
*/
	pEdict->v.button |= IN_JUMP;
}

ETaskCode CWallJumpTask::TaskIdentify()
{
	return TaskWallJump;
}

bool CWallJumpTask::ShouldManageEnemys()
{
	return false;
}

// Charge/Shoot kame torpedo task
CKameTorpTask::CKameTorpTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckAttacks | CheckSurroundingPlayers | CheckSounds | CheckHit);
	BotMiscChecks = 0;

	float BaseLifeTime = RANDOM_FLOAT(35.0f, 43.0f);
	if (pBot->AggressionLevel == Defensive)
		BaseLifeTime *= 1.2f;
	else if (pBot->AggressionLevel == Aggressive)
		BaseLifeTime *= 0.8f;

	LifeTime = Delay + BaseLifeTime;
	ShootLifeTime = 0.0;
	DesiredTorpCount = 6;
	ShootingTorps = false;
	KameTorpState = STATE_CHARGING;
	MoveAwayTime = 0.0;
	MoveAwayDir = 0;
	LookDelay = 0.0;
	MoveAwayDelay = 0.0;
	PickedAnAimSpot = false;
}

CKameTorpTask::~CKameTorpTask()
{
	pEdict->v.button |= IN_ATTACK2;
	pBot->ChargeKameTorp = false;
	pBot->CheckKameTorpTime = gpGlobals->time + RANDOM_FLOAT(22.0, 42.0);
}

bool CKameTorpTask::KillTaskOnDanger(short DangerLevel)
{
	if ((pBot->DangerType == DangerCurrentEnemy) && (DangerLevel == danger_hard) &&
		(pBot->pBotEnemy))
	{
		pEdict->v.button |= IN_ATTACK2;
		return false;
	}

	if (DangerLevel == danger_light)
		DesiredTorpCount = 3;
	else if (DangerLevel == danger_medium)
		DesiredTorpCount = 1;
	else
	{
		pEdict->v.button |= IN_ATTACK2;
		return true;
	}

	return false;
}

ETaskCode CKameTorpTask::TaskIdentify()
{
	return TaskKameTorp;
}

bool CKameTorpTask::ShouldManageEnemys()
{
	return true;
}

// Melee struggle task(obsolete)
CMeleeStruggleTask::CMeleeStruggleTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;

	ChargeTime = Delay + RANDOM_FLOAT(0.6, 1.1);
	RestTime = 0.0;
	TargetChargeVal = 33 + ((pBot->bot_skill + 1) * RANDOM_LONG(4, 7));
}

CMeleeStruggleTask::~CMeleeStruggleTask()
{
}

bool CMeleeStruggleTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CMeleeStruggleTask::TaskThink()
{
	if (pBot->pBotClient->MeleeStruggles == false)
	{
		TaskState = done;
		return;
	}

	if (RestTime < gpGlobals->time)
	{
		if (pBot->ChargeBeamVal < TargetChargeVal)
			pEdict->v.button |= IN_ATTACK;
		else
		{
			float MinDelay = GeneralReactionDelays[pBot->bot_skill][0];
			float MaxDelay = GeneralReactionDelays[pBot->bot_skill][1];

			RestTime = gpGlobals->time + RANDOM_FLOAT(MinDelay, MaxDelay);

			TargetChargeVal = 33 + (pBot->bot_skill + 1 * RANDOM_LONG(3, 10));
		}
	}
/*
	if (ChargeTime > gpGlobals->time)
		pEdict->v.button |= IN_ATTACK;

	if (ChargeTime < gpGlobals->time)
	{
		if (RestTime < gpGlobals->time)
		{
			if (RestTime > 0.0)
			{
				float MinDelay = 0.7 + (float(pBot->bot_skill) * 0.03);
				float MaxDelay = 1.2 + (float(pBot->bot_skill) * 0.03);
				ChargeTime = gpGlobals->time + RANDOM_FLOAT(MinDelay, MaxDelay);
				RestTime = 0.0;
			}
			else
			{
				float MinDelay = 0.06 + (float(pBot->bot_skill) * 0.03);
				float MaxDelay = 0.11 + (float(pBot->bot_skill) * 0.03);

				RestTime = gpGlobals->time + RANDOM_FLOAT(MinDelay, MaxDelay);
			}
		}
	}*/
}

ETaskCode CMeleeStruggleTask::TaskIdentify()
{
	return TaskMeleeStruggle;
}

bool CMeleeStruggleTask::ShouldManageEnemys()
{
	return false;
}

// Idle task, made when a bot can't do anything
CIdleTask::CIdleTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = CheckForChatMsg;
}

CIdleTask::~CIdleTask()
{
}

bool CIdleTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CIdleTask::TaskThink()
{
	bool BotIsIdle = ((BotsAreIdle) || (pBot->pBotClient->LaysOnGround) || 
					  (pBot->pBotClient->GetAutoPunched));

	if (BotIsIdle == false)
		TaskState = done;
	else if (BotsAreIdle)
		ResetMoveSpeed(pBot);
}

ETaskCode CIdleTask::TaskIdentify()
{
	return TaskIdle;
}

bool CIdleTask::ShouldManageEnemys()
{
	return false;
}

// Fly after hit task, get made when a bot just got hit by melee and needs to press attack to 
// recover
CFlyAfterHitTask::CFlyAfterHitTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;
}

CFlyAfterHitTask::~CFlyAfterHitTask()
{
}

bool CFlyAfterHitTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CFlyAfterHitTask::TaskThink()
{
    if (pBot->pBotClient->FlysAfterHit == false)
		TaskState = done;
	else
		pEdict->v.button |= (IN_ATTACK | IN_JUMP);
}

ETaskCode CFlyAfterHitTask::TaskIdentify()
{
	return TaskFlyAfterHit;
}

bool CFlyAfterHitTask::ShouldManageEnemys()
{
	return false;
}

// Beam jump task
CBeamJumpTask::CBeamJumpTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = (CheckAttacks | CheckSurroundingPlayers | CheckSounds | CheckHit);
	BotMiscChecks = CheckForChatMsg;

	LifeTime = Delay + RANDOM_FLOAT(2.2, 3.2);
	TraceDelay = 0.0;
	CreateTime = Delay;
}

CBeamJumpTask::~CBeamJumpTask()
{
}

bool CBeamJumpTask::KillTaskOnDanger(short DangerLevel)
{
	return ((DangerLevel >= danger_hard) && (pBot->pBotClient->BeamJumps == false));
}

void CBeamJumpTask::TaskThink()
{
	pEdict->v.button |= IN_JUMP;
	if (LifeTime <= gpGlobals->time)
	{
		TaskState = done;
		return;
	}

	if ((pBot->pBotClient->ChargesAttack == false) && (pBot->pBotClient->BeamJumps == false) &&
		(pBot->pBotClient->ControlsAttack == false))
	{
		TaskState = done;
		return;
	}

	if (!pBot->pBotClient->BeamJumps) // Bot doesn't beamjump yet?
	{
#ifdef _DEBUG
		UTIL_DrawBeam(listenserver_edict, listenserver_edict->v.origin, pBot->BeamJumpSpot, 10, 1,
					  0, 0, 255, 250, 5, 25);
#endif
		// Aim at the spot where the beam should collide to
		pBot->AimVec = WrapAngles(UTIL_VecToAngles(pBot->BeamJumpSpot - pEdict->v.origin));

		if ((pBot->ChargeBeamVal) && ((CreateTime + 0.2f) >= gpGlobals->time))
		{
#ifdef _DEBUG
			lcout << "BeamJump delay\n";
#endif
			pEdict->v.button |= IN_ATTACK;
		}
	}
	else if (BotSenseChecks)
		BotSenseChecks = 0; // Don't want to react on danger while beamjumping

//	edict_t *pAttack = GetAttackFromPlayer(pEdict);
	edict_t *pAttack = pAttackManager->SearchLastAttack(pEdict);

	if (IsNull(pAttack))
	{
		if (LifeTime < (gpGlobals->time+0.4))
		{
			TaskState = done;
			return;
		}
	}

	if ((pBot->pBotClient->BeamJumps) && (pBot->pBotEnemy))
	{
		float EnemyDist = UTIL_Distance(pEdict->v.origin, pBot->pBotEnemy->v.origin);
		float EnemyToAttackDist = UTIL_Distance(pBot->pBotEnemy->v.origin, pAttack->v.origin);
		float BotToAttackDist = UTIL_Distance(pEdict->v.origin, pAttack->v.origin);
		float ExplodeDistForBot = pAttack->v.dmg * 11;
		float ExplodeDistForEnemy = pAttack->v.dmg * 9;

		// If the the bot isn't close to his enemy AND not close to his own attack AND
		// the enemy is close to the bot's attack...
		if ((EnemyDist > 950.0f) && (EnemyToAttackDist < ExplodeDistForEnemy) &&
			(BotToAttackDist > ExplodeDistForBot))
		{
			TaskState = done; // Explode the beam >:)
			return;
		}
	}

	if (TraceDelay <= gpGlobals->time)
	{
		if (pBot->pBotClient->BeamJumps)
		{
			TraceDelay = gpGlobals->time + 0.1;
			if (CheckForCollide(pEdict, dir_back, 50.0))
			{
				TaskState = done;
			}
		}
		/*
		else
		{
			TraceDelay = gpGlobals->time + 0.25;
			Vector v_nearest = GetNearestCollision(pEdict, 500.0);
			if (v_nearest != g_vecZero)
				pBot->AimVec = WrapAngles(UTIL_VecToAngles(v_nearest - pEdict->v.origin));
			else
				TaskState = done;
		}*/
	}
}

ETaskCode CBeamJumpTask::TaskIdentify()
{
	return TaskBeamJump;
}

bool CBeamJumpTask::ShouldManageEnemys()
{
	return false;
}

// Normal jump task, to avoid danger
CNormalJumpTask::CNormalJumpTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = CheckKameTorpLeftOvers;

	LifeTime = Delay + RANDOM_FLOAT(0.8, 1.2);
	TeleCheckDelay = Delay + 0.2;
}

CNormalJumpTask::~CNormalJumpTask()
{
}

bool CNormalJumpTask::KillTaskOnDanger(short DangerLevel)
{
	return false;
}

void CNormalJumpTask::TaskThink()
{
	pEdict->v.button |= IN_JUMP;

	if (LifeTime <= gpGlobals->time)
	{
		TaskState = done;
		return;
	}

	if (pEdict->v.velocity.z < 0) // Going down?
	{
		TaskState = done;
		return;
	}

	if (pEdict->v.gravity < 0.1)
	{
		if (pEdict->v.flags & FL_ONGROUND)
			FakeClientCommand(pEdict, "togglefly", NULL, NULL);
		else
		{
			TaskState = done;
			return;
		}
	}

	SetMoveDir(pBot, pBot->JumpDir);

	if (TeleCheckDelay < gpGlobals->time)
	{
		TeleCheckDelay = gpGlobals->time + RANDOM_FLOAT(0.25, 0.4);

		if (RANDOM_LONG(1, 100) <= pBot->TeleportPercent)
			FakeClientCommand(pEdict, "teleport", NULL, NULL);
	}
}

ETaskCode CNormalJumpTask::TaskIdentify()
{
	return TaskNormalJump;
}

bool CNormalJumpTask::ShouldManageEnemys()
{
	return false;
}

// Shoot candy ray task
CShootCandyTask::CShootCandyTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = CheckFly;
	BotSenseChecks = (CheckHit | CheckAttacks);
	BotMiscChecks = 0;

	LifeTime = Delay + RANDOM_FLOAT(2.5, 5.0);
	TeleCheckTime = 0.0;
	MoveBackTime = 0.0;
	CookieCheckDelay = 0.0;
	MoveToCookieTime = 0.0;
	EnemyIsInRangeTime = Delay;
}

CShootCandyTask::~CShootCandyTask()
{
}

bool CShootCandyTask::KillTaskOnDanger(short DangerLevel)
{
	return (DangerLevel >= danger_medium);
}

void CShootCandyTask::TaskThink()
{
	if (pBot->bot_ki < 50)
	{
		TaskState = done;
		return;
	}

	if (pBot->pBotEnemy == NULL)
	{
		if (MoveToCookieTime == 0.0)
			MoveToCookieTime = gpGlobals->time + 1.0f;
		else if (MoveToCookieTime < gpGlobals->time)
		{
			TaskState = done;
			return;
		}

		if (CookieCheckDelay < gpGlobals->time)
		{
			CookieCheckDelay = gpGlobals->time + 0.15;

			edict_t *pCookie = NULL;
			bool FoundOne = false;

			while ((pCookie = UTIL_FindEntityByClassname(pCookie, "candy")) != NULL)
			{
				if (IsNull(pCookie))
					continue;

				if (pCookie->v.owner != pBot->pLastEnemy)
				{
//					lcout << "Wrong owner\n";
					continue;
				}

				if (UTIL_Distance(pEdict->v.origin, pCookie->v.origin) > 500.0)
					continue;

				if (EntIsInFieldOfView(pEdict, pCookie) == false)
					continue;

				FoundOne = true;
				break;
			}

			if ((FoundOne) || ((MoveToCookieTime - gpGlobals->time) > 0.75))
			{
				//lcout << "Found A Cookie\n";
				//pBot->AimVec = WrapAngles(UTIL_VecToAngles(pCookie->v.origin - pEdict->v.origin));
			}
			else
			{
				//lcout << "No Cookie\n";
				TaskState = done;
			}
		}
		return;
	}
				
	edict_t *pEnemy = pBot->pBotEnemy;

	ClientInfo_t *pClient = UTIL_GetClientPointer(pEnemy);
	if (pClient)
	{
		if ((pClient->Swoops) || (pClient->MeleeStruggles) || (pClient->Powerstruggles) ||
			(pClient->IsAutoPunching) || (pClient->GetAutoPunched))
		{
	//		lcout << "Invalid Enemy\n";
			TaskState = done;
			return;
		}
	}
	float EnemyDist = UTIL_Distance(pEdict->v.origin, pEnemy->v.origin);

	if (EnemyDist > 950.0)
	{
//		lcout << "Dist to high\n";
		TaskState = done;
		return;
	}
	else if (EnemyDist > 500.0)
	{
		if (TeleCheckTime <= gpGlobals->time)
		{
			TeleCheckTime = gpGlobals->time + 0.2;

			if (RANDOM_LONG(1, 100) <= pBot->TeleportPercent)
				FakeClientCommand(pEdict, "teleport", NULL, NULL);
		}
		else if (pBot->bot_ki > 250)
			ToggleTurbo(pEdict, true); // Just toggle turbo on
	}
	else if (EnemyDist < 75.0)
	{
		EnemyIsInRangeTime = gpGlobals->time;
		MoveBackTime = gpGlobals->time + 0.35;
//		lcout << "Go back\n";
	}
	else
		EnemyIsInRangeTime = gpGlobals->time;

	// Is the enemy to far for a while?
	if (EnemyIsInRangeTime < (gpGlobals->time - 3.5f))
	{
		TaskState = done;
		return;
	}

	if (MoveBackTime > gpGlobals->time)
		pBot->f_move_speed = -pBot->f_max_speed;

	if (pBot->current_weapon.iId != ESF_WEAPON_CANDY)
		FakeClientCommand(pEdict, "weapon_candy", NULL, NULL);

	pEdict->v.button |= IN_ATTACK;
}

ETaskCode CShootCandyTask::TaskIdentify()
{
	return TaskShootCandy;
}

bool CShootCandyTask::ShouldManageEnemys()
{
	return true;
}

// Dragonball task
CDBTask::CDBTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = (CheckFly);
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = 0;

	flDBThrowDelay = 0.0f;
}

CDBTask::~CDBTask()
{
	pBot->pGoalDB = NULL;
	pBot->waypoint_goal = -1;
	pBot->curr_waypoint_index = -1;
}

void CDBTask::HeadToNextDBWP()
{
	int i;
    Vector v_src, v_dest;
    TraceResult tr;
    float waypoint_distance, min_distance;
    int team; 
    bool touching;

	team = UTIL_GetTeam(pEdict);

    // check if the bot has been trying to get to this waypoint for a while...
    if ((pBot->f_waypoint_time + 5.0) < gpGlobals->time)
    {
		pBot->curr_waypoint_index = -1;  // forget about this waypoint
    }

    // check if we need to find a waypoint...
    if (pBot->curr_waypoint_index == -1)
    {
		// find the nearest reachable waypoint
		i = WaypointFindReachable(pEdict, 1250.0f, team);

		if (i == -1)
		{
			pBot->curr_waypoint_index = -1;
			TaskState = done;
			return;
      }

      pBot->curr_waypoint_index = i;
      pBot->waypoint_origin = waypoints[i].origin;

      pBot->f_waypoint_time = gpGlobals->time;
    }
    else
    {
		// check if we can still see our target waypoint...

        v_src = pEdict->v.origin + pEdict->v.view_ofs;
        v_dest = waypoints[pBot->curr_waypoint_index].origin;

        // trace a line from bot's eyes to destination...
        UTIL_TraceLine( v_src, v_dest, ignore_monsters,
                        pEdict->v.pContainingEntity, &tr );

        // check if line of sight to object is blocked (i.e. not visible)
        if (tr.flFraction < 1.0)
        {
            i = WaypointFindReachable(pEdict, 1250.0f, team);

            if (i == -1)
            {
				pBot->curr_waypoint_index = -1;
				TaskState = done;
                return;
            }

            pBot->curr_waypoint_index = i;
            pBot->waypoint_origin = waypoints[i].origin;

            pBot->f_waypoint_time = gpGlobals->time;
		}
	}

    // find the distance to the target waypoint
    waypoint_distance = (pEdict->v.origin - pBot->waypoint_origin).Length();

    // set the minimum distance from waypoint to be considered "touching" it
    min_distance = 50.0;

    if (waypoints[pBot->curr_waypoint_index].flags & W_FL_JUMP)
		min_distance = 25.0;

    // if this is a ladder waypoint, bot must be fairly close to get on ladder
    if (waypoints[pBot->curr_waypoint_index].flags & W_FL_LADDER)
		min_distance = 20.0;

    touching = FALSE;

    // did the bot run past the waypoint? (prevent the loop-the-loop problem)
    if ((pBot->prev_waypoint_distance > 1.0) &&
        (waypoint_distance > pBot->prev_waypoint_distance))
		touching = TRUE;

    // are we close enough to a target waypoint...
    if (waypoint_distance < min_distance)
		touching = TRUE;

    // save current distance as previous
    pBot->prev_waypoint_distance = waypoint_distance;

    if (touching)
    {
        pBot->prev_waypoint_distance = 0.0;

        // check if the bot has reached the goal waypoint...
        if (pBot->curr_waypoint_index == pBot->waypoint_goal)
		{
			pBot->waypoint_goal = -1;  // forget this goal waypoint
			TaskState = done;
			return;
		}
	     
        // get the next waypoint to reach goal...
        i = WaypointRouteFromTo(pBot->curr_waypoint_index, pBot->waypoint_goal, team);

        if (i != WAYPOINT_UNREACHABLE)  // can we get to the goal from here?
		{
            pBot->curr_waypoint_index = i;
            pBot->waypoint_origin = waypoints[i].origin;

            pBot->f_waypoint_time = gpGlobals->time;
		}
		else
		{
			TaskState = done;
			return;
		}
	}

    // keep turning towards the waypoint...

    Vector v_direction = pBot->waypoint_origin - pEdict->v.origin;
    Vector v_angles = UTIL_VecToAngles(v_direction);
    pBot->AimVec = WrapAngles(v_angles);
}

bool CDBTask::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel >= danger_hard)
	{
//		if (pBot->pBotClient->HasDB)
//			pEdict->v.button |= IN_ATTACK; // Throw it away
		return true;
	}
	else if (DangerLevel == danger_medium)
	{
		if (RANDOM_LONG(1, 100) <= pBot->TeleportPercent)
			FakeClientCommand(pEdict, "teleport", NULL, NULL);
	}
	return false;
}

void CDBTask::TaskThink()
{
	if (pBot->bot_ki < 250)
	{
		MakeTask(pBot, TaskPowerup);
	}

	if (DBSAreGathered())
	{
		if (!GoToGatheredSpot())
		{
			TaskState = done;
			return;
		}
	}
	else if (pBot->pBotClient->HasDB)
	{
		if (!GoToDBSpot())
		{
			TaskState = done;
			return;
		}
	}
	else
	{
		GoToDB();
	}

	ToggleTurbo(pEdict, true);
}

ETaskCode CDBTask::TaskIdentify()
{
	return TaskDB;
}

bool CDBTask::ShouldManageEnemys()
{
	return true;
}

// Jump of the ground task(after being thrown)
CJumpOffGroundTask::CJumpOffGroundTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;

	TMultiChoice<int> JumpDirChoices;
	JumpDirChoices.Insert(dir_forward, 50);
	JumpDirChoices.Insert(dir_back, 50);
	JumpDirChoices.Insert(dir_left, 50);
	JumpDirChoices.Insert(dir_right, 50);
	JumpDirChoices.Insert((dir_forward | dir_left), 50);
	JumpDirChoices.Insert((dir_forward | dir_right), 50);
	JumpDirChoices.Insert((dir_back | dir_left), 50);
	JumpDirChoices.Insert((dir_back | dir_right), 50);

	MoveDir = 0; // Incase no choice found....
	JumpDirChoices.GetSelection(MoveDir);
}

void CJumpOffGroundTask::TaskThink()
{
	pEdict->v.button |= IN_JUMP;

	ResetMoveSpeed(pBot);
	SetMoveDir(pBot, MoveDir);

	if (pBot->pBotClient->HasThrowLanded == false)
	{
		TaskState = done;
		return;
	}
}

// Handle swung task, when being swung this task is made
CHandleSwungTask::CHandleSwungTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = 0;
	BotSenseChecks = 0;
	BotMiscChecks = 0;
}

void CHandleSwungTask::TaskThink()
{
	pEdict->v.button |= IN_JUMP;

	ResetMoveSpeed(pBot);

	if (pBot->pBotClient->GetSwung == false)
	{
		TaskState = done;
		return;
	}
}


// Shoot Health laser task(for super bots)
CShootLaser1Task::CShootLaser1Task(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = CheckFly;
	BotSenseChecks = (CheckSounds | CheckAttacks | CheckSurroundingPlayers | CheckHit);
	BotMiscChecks = 0;

	ShootTime = Delay + RANDOM_FLOAT(3.0, 6.0);
}

CShootLaser1Task::~CShootLaser1Task()
{
	if (pBot->pBotEnemy)
		pBot->pBotEnemy->v.flags &= ~FL_FROZEN;
}

bool CShootLaser1Task::KillTaskOnDanger(short DangerLevel)
{
	if (DangerLevel >= danger_hard)
		return true;

	return false;
}

void CShootLaser1Task::TaskThink()
{
	edict_t *pEnemy = pBot->pBotEnemy;

	if (pEnemy == NULL)
	{
		TaskState = done;
		return;
	}

	if (ShootTime < gpGlobals->time)
	{
		TaskState = done;
		return;
	}

	if (pEdict->v.health >= pBot->max_health)
	{
		TaskState = done;
		return;
	}

	if (pEdict->v.fuser4 < 150.0)
	{
		TaskState = done;
		return;
	}

	ClientInfo_t *pClient = UTIL_GetClientPointer(pEnemy);
	if (!pClient)
	{
		TaskState = done;
		return;
	}

	if ((pClient->IsInMeleeMode) || (pClient->Powerstruggles) || (pClient->MeleeStruggles) ||
		(pClient->IsAutoPunching) || (pClient->GetAutoPunched))
	{
		pEnemy->v.flags &= ~FL_FROZEN;
		TaskState = done;
		return;
	}

	pBot->f_move_speed = pBot->f_strafe_speed = pBot->f_updown_speed = 0.0;

	pBot->pBotEnemy->v.flags |= FL_FROZEN;

	pClient->FrozenTime = gpGlobals->time + 0.5;

	//if (pBot->pBotEnemy->v.maxspeed > 0.0)
		//	(*g_engfuncs.pfnSetClientMaxspeed)(pBot->pBotEnemy, 0.0);

	ShootHealthBeam(pBot, pBot->pBotEnemy);

	if (pBot->LastBeamDrawTime <= gpGlobals->time)
		pBot->LastBeamDrawTime = gpGlobals->time + 0.1;
}

ETaskCode CShootLaser1Task::TaskIdentify()
{
	return TaskShootLaser1;
}

bool CShootLaser1Task::ShouldManageEnemys()
{
	// This task will manage enemies by his own...
	edict_t *pTemp = pBot->pBotEnemy;

	pBot->pBotEnemy = GetBestEnemyFromList(pBot);

	if ((pBot->pBotEnemy != pTemp) && (UTIL_IsValidPlayer(pTemp, false, false)))
		pTemp->v.flags &= ~FL_FROZEN; // Otherwise player keeps frozen

	return false;
}

// Keep danger away - task (for sbots)
CKeepDangerAwayTask::CKeepDangerAwayTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = CheckFly;
	BotSenseChecks = CheckHit;
	BotMiscChecks = 0;

	LifeTime = Delay + RANDOM_FLOAT(3.0, 7.0);
}

CKeepDangerAwayTask::~CKeepDangerAwayTask()
{
	pEdict->v.renderfx = kRenderFxNone;
  	pEdict->v.rendermode = kRenderTransTexture;
	pEdict->v.renderamt = 255;
}

bool CKeepDangerAwayTask::KillTaskOnDanger(short DangerLevel)
{
	return true;
}

void CKeepDangerAwayTask::TaskThink()
{
	if (pEdict->v.fuser4 < 150.0)
	{
		TaskState = done;
		return;
	}

	if ((LifeTime < gpGlobals->time) || (pBot->f_max_speed == 0.0))
	{
		TaskState = done;
		return;
	}

	bool PlaySound = false;

	pBot->f_move_speed = pBot->f_strafe_speed = pBot->f_updown_speed = 0.0;

	//	ClientPrint(listenserver_edict,HUD_PRINTCENTER,"DANGER\n");

	if (IsTurboOn(pEdict))
	{
		ToggleTurbo(pEdict,false); // turn turbo off when its on, otherwise we get crappy things...
		return;
	}

	pEdict->v.rendermode = kRenderNormal;
    pEdict->v.renderfx = kRenderFxGlowShell;
  
    pEdict->v.renderamt = 25;  // glow shell distance from entity
    pEdict->v.rendercolor.x = 255;
    pEdict->v.rendercolor.y = 255;
    pEdict->v.rendercolor.z = 255;

	bool BotMayDrawBeams = false;

	if (pBot->LastBeamDrawTime <= gpGlobals->time)
	{
		pBot->LastBeamDrawTime = gpGlobals->time + 0.1;
		BotMayDrawBeams = true;
	}

	if (KeepAttacksAwayFromBot(pBot, BotMayDrawBeams))
	{
		LifeTime += 1.5;
		PlaySound = true;
	}

	if (!pBot->MadePlayerList)
		MakePlayerList(pBot);

	if (pBot->PlayerCount)
	{
		edict_t *pPlayer = NULL;
		ClientInfo_t *pClient;
		float dist = 0.0;

		for (int i=0;i<pBot->PlayerCount;i++)
		{
			pPlayer = pBot->PlayerList[i];

			if (IsNull(pPlayer))
				continue;

			dist = UTIL_Distance(pEdict->v.origin, pPlayer->v.origin);

			if (dist < 350.0)
			{
				TaskState = done;
				return;
			}

		   	// is team play enabled?
			if (IsTeamPlay())
			{
       			int player_team = UTIL_GetTeam(pPlayer);
				int bot_team = UTIL_GetTeam(pEdict);

				// don't target your teammates...
				if (bot_team == player_team)
				{
				    continue;
				}	
			}

			pClient = UTIL_GetClientPointer(pPlayer);
			if (!pClient)
				continue;

			if (pClient->IsInMeleeMode)
				continue;

			if (pClient->MeleeStruggles)
				continue;

			if (pClient->IsAutoPunching)
				continue;

			if (pClient->Powerstruggles)
				continue;

			if (pClient->GetAutoPunched)
				continue;

			if ((dist < 800.0) && (pBot->bot_ki > 0))
			{
				if (!PlaySound)
					PlaySound = true;

				UTIL_MakeVectors(pPlayer->v.angles);

				if (BotMayDrawBeams)
					UTIL_DrawBeamToAll(pEdict->v.origin + pEdict->v.view_ofs, pPlayer->v.origin,
									   20, 8, 125, 255, 255, 250, 5);

				long int NewPlayerPl = pClient->PL;
				long int NewBotPl = pBot->PL;

				if (IsTurboOn(pPlayer))
					NewPlayerPl /= 1.25;

				if (IsTurboOn(pEdict))
					NewBotPl /= 1.25;

				if (pPlayer->v.velocity.Length() > 10.0)
				{
					float VelocityFactor = 0.95f;
					long PlDiff = NewBotPl - NewPlayerPl;

					if (PlDiff > 0)
						VelocityFactor = float(PlDiff) / 10000000.0f;

					if (VelocityFactor < 0.05)
						VelocityFactor = 0.05; // Player should be atleast 5% slowed down
					else if (VelocityFactor > 0.95f)
						VelocityFactor = 0.95f; // But he should be able to move aswell...

					pPlayer->v.velocity = pPlayer->v.velocity - pPlayer->v.velocity.Normalize() * (pPlayer->v.velocity.Length() * VelocityFactor);
				}

				NewPlayerPl -= ABSORB_PL_GAIN;
				NewBotPl += ABSORB_PL_GAIN;

				if (pClient->PL <= ABSORB_PL_GAIN)
				{
					// kill the enemy when he doesn't have much pl left
					UTIL_DamagePlayer(pEdict, pPlayer, "Super bot laser", pPlayer->v.health+1);
				}
				else if (NewBotPl >= 999999999)
				{
				}
				else
				{
					UTIL_ChangePl(pPlayer, NewPlayerPl);
					UTIL_ChangePl(pEdict, NewBotPl);
					UTIL_ChangeRealPl(pEdict, NewBotPl);
				}
					
				pEdict->v.fuser4 -= 3.0; // Decrease ki of the bot
				pPlayer->v.fuser4 -= 3.0; // Decrease ki of the victim

				if (pEdict->v.fuser4 < 0.0)
					pEdict->v.fuser4 = 0.0;

				if (pPlayer->v.fuser4 < 0.0)
					pPlayer->v.fuser4 = 0.0;

				if (LifeTime < (gpGlobals->time + 0.5))
					LifeTime = gpGlobals->time + 0.7;
			}
		}
	}

	if (PlaySound)
	{
		if (pBot->LastSoundPlayedTime <= gpGlobals->time)
		{
			EMIT_SOUND_DYN2(pEdict, CHAN_WEAPON, "weapons/basicbeam_chargeoriginal.wav", 1.0,
						    ATTN_NORM, 0, 100);

			// Small delay, otherwise the engine will get it to busy...
			pBot->LastSoundPlayedTime = gpGlobals->time + 0.3;
		}
	}
	else
	{
		if (LifeTime > (gpGlobals->time + 0.8))
			LifeTime = gpGlobals->time + 0.7;
	}
}

ETaskCode CKeepDangerAwayTask::TaskIdentify()
{
	return TaskKeepDangerAway;
}

bool CKeepDangerAwayTask::ShouldManageEnemys()
{
	return false;
}

// Absorb health from multiple enemys task (for sbots)
CAbsorbHealthTask::CAbsorbHealthTask(bot_t *pB) : CBaseTask(pB)
{
	BotStatusChecks = CheckFly;
	BotSenseChecks = CheckHit;
	BotMiscChecks = 0;

	LifeTime = Delay + RANDOM_FLOAT(3.0, 7.0);
}

CAbsorbHealthTask::~CAbsorbHealthTask()
{
	pEdict->v.renderfx = kRenderFxNone;
  	pEdict->v.rendermode = kRenderTransTexture;
	pEdict->v.renderamt = 255;
}

bool CAbsorbHealthTask::KillTaskOnDanger(short DangerLevel)
{
	return true;
}

void CAbsorbHealthTask::TaskThink()
{
	if (pEdict->v.health >= pBot->max_health)
	{
		TaskState = done;
		return;
	}

	if (pEdict->v.fuser4 < 100.0)
	{
		TaskState = done;
		return;
	}

	ToggleTurbo(pEdict,false); // turn turbo off when its on, otherwise we get crappy things...

    pEdict->v.rendermode = kRenderNormal;
    pEdict->v.renderfx = kRenderFxGlowShell;
  
    pEdict->v.renderamt = 25;  // glow shell distance from entity
    pEdict->v.rendercolor.x = 125;
    pEdict->v.rendercolor.y = 255;
    pEdict->v.rendercolor.z = 255;

	ResetMoveSpeed(pBot);

	if (!ShootAbsorbHealthBeams(pBot))
	{
		// couldn't absorb someone :(
		TaskState = done;
		return;
	}
}

ETaskCode CAbsorbHealthTask::TaskIdentify()
{
	return TaskAbsorbHealth;
}

bool CAbsorbHealthTask::ShouldManageEnemys()
{
	return false;
}


//////////////////////////////////////////////////////
//				Class code END					    //
//////////////////////////////////////////////////////



//////////////////////////////////////////////////////
//				Function code BEGIN				    //
//////////////////////////////////////////////////////

void MakeTask(bot_t *pBot, ETaskCode TaskCode)
{
#ifdef LOGBOTINFO
	clog << "New task: " << TaskCode << "\n";
#endif

	if (pBot->CurrentBotTask)
	{
		if (pBot->CurrentBotTask->TaskIdentify() == TaskCode)
			return; // don't make the task when its already made...

		pBot->PrevTask = pBot->CurrentBotTask->TaskIdentify();
		KillCurrentTask(pBot);
	}
#ifdef LOGBOTINFO
	clog << "Making new task: " << TaskCode << "\n";
#endif

	switch (TaskCode)
	{
	case TaskPowerup:
		pBot->CurrentBotTask = new CPowerUpTask(pBot);
		break;
	case TaskTeleport:
		pBot->CurrentBotTask = new CTeleportTask(pBot);
		pBot->CurrentBotSubTask = new CAimAtEnemySubTask(pBot);
		break;
	case TaskBlock:
		pBot->CurrentBotTask = new CBlockTask(pBot);
		pBot->CurrentBotSubTask = new CAimAtEnemySubTask(pBot);
		break;
	case TaskNormalMovement:
		pBot->CurrentBotTask = new CNormalMovementTask(pBot);
		break;
	case TaskCombat:
		pBot->CurrentBotTask = new CCombatTask(pBot);
		pBot->CurrentBotSubTask = new CCombatMovementSubTask(pBot);
		break;
	case TaskChargeAttack:
		pBot->CurrentBotTask = new CChargeAttackTask(pBot);
		pBot->CurrentBotSubTask = new CCombatMovementSubTask(pBot);
		break;
	case TaskSwoop:
		pBot->CurrentBotTask = new CSwoopTask(pBot);
		break;
	case TaskFollowTeamMate:
		pBot->CurrentBotTask = new CFollowTeamMateTask(pBot);
		break;
	case TaskControlFriezaDisc:
		pBot->CurrentBotTask = new CControlFriezaDiscTask(pBot);
		pBot->CurrentBotSubTask = new CCombatMovementSubTask(pBot);
		break;
	case TaskControlBeam:
		pBot->CurrentBotTask = new CControlBeamTask(pBot);
		break;
	case TaskAscend:
		pBot->CurrentBotTask = new CAscendTask(pBot);
		break;
	case TaskPowerstruggle:
		pBot->CurrentBotTask = new CPowerStruggleTask(pBot);
		break;
	case TaskCombo:
		pBot->CurrentBotTask = new CComboTask(pBot);
		break;
	case TaskLookAround:
		pBot->CurrentBotTask = new CLookAroundTask(pBot);
		break;
	case TaskRecoverHP:
		pBot->CurrentBotTask = new CRecoverHPTask(pBot);
		break;
	case TaskMoveDir:
		pBot->CurrentBotTask = new CMoveDirTask(pBot);
		pBot->CurrentBotSubTask = new CAimAtEnemySubTask(pBot);
		break;
	case TaskBlinded:
		pBot->CurrentBotTask = new CBlindedTask(pBot);
		break;
	case TaskFinishMelee:
		pBot->CurrentBotTask = new CFinishMeleeTask(pBot);
		break;
	case TaskKickEnemy:
		pBot->CurrentBotTask = new CKickEnemyTask(pBot);
		break;
	case TaskCounterMelee:
		pBot->CurrentBotTask = new CCounterMeleeTask(pBot);
		break;
	case TaskWallJump:
		pBot->CurrentBotTask = new CWallJumpTask(pBot);
		break;
	case TaskKameTorp:
		pBot->CurrentBotTask = new CKameTorpTask(pBot);
		break;
	case TaskMeleeStruggle:
		pBot->CurrentBotTask = new CMeleeStruggleTask(pBot);
		break;
	case TaskIdle:
		pBot->CurrentBotTask = new CIdleTask(pBot);
		break;
	case TaskFlyAfterHit:
		pBot->CurrentBotTask = new CFlyAfterHitTask(pBot);
		break;
	case TaskBeamJump:
		pBot->CurrentBotTask = new CBeamJumpTask(pBot);
		break;
	case TaskNormalJump:
		pBot->CurrentBotTask = new CNormalJumpTask(pBot);
		break;
	case TaskShootCandy:
		pBot->CurrentBotTask = new CShootCandyTask(pBot);
		break;
	case TaskShootScatterShots:
		pBot->CurrentBotTask = new CShootScatterShotsTask(pBot);
		pBot->CurrentBotSubTask = new CCombatMovementSubTask(pBot);
		break;
	case TaskShootNonChargeble:
		pBot->CurrentBotTask = new CShootNonChargebleTask(pBot);
		pBot->CurrentBotSubTask = new CCombatMovementSubTask(pBot);
		break;
	case TaskDB:
		pBot->CurrentBotTask = new CDBTask(pBot);
		break;
	case TaskGoToVec:
		pBot->CurrentBotTask = new CGoToVecTask(pBot);
		break;
	case TaskJumpOffGround:
		pBot->CurrentBotTask = new CJumpOffGroundTask(pBot);
		break;
	case TaskHandleSwung:
		pBot->CurrentBotTask = new CHandleSwungTask(pBot);
		break;

	// exception: tasks get made elsewhere
	case TaskSwoopAway:
		MakeAvoidSwoopTask(pBot);
		break;

	// tasks for super bots
	case TaskShootLaser1:
		pBot->CurrentBotTask = new CShootLaser1Task(pBot);
		break;
	case TaskKeepDangerAway:
		pBot->CurrentBotTask = new CKeepDangerAwayTask(pBot);
		break;
	case TaskAbsorbHealth:
		pBot->CurrentBotTask = new CAbsorbHealthTask(pBot);
		break;

	}

#ifdef _DEBUG
	if (!pBot->CurrentBotTask)
	{
		lcout << "Task " << TaskCode << " not found !!\n";
		clog << "Task " << TaskCode << " not found !!\n";
	}
#endif

#ifdef LOGBOTINFO
	clog << "Done making new task: " << TaskCode << "\n";
#endif
}

void KillCurrentTask(bot_t *pBot)
{
#ifdef LOGBOTINFO
	clog << "Begin kill task\n";
#endif

	if (pBot->CurrentBotTask)
	{
		delete pBot->CurrentBotTask;
		pBot->CurrentBotTask = NULL;
	}

	if (pBot->CurrentBotSubTask)
	{
		delete pBot->CurrentBotSubTask;
		pBot->CurrentBotSubTask = NULL;
	}

#ifdef LOGBOTINFO
	clog << "End kill task\n";
#endif
}

void CheckNesseserelyThings(bot_t *pBot)
{
	edict_t *pEdict = pBot->pEdict;
	bool CheckWJump = false;

	if (pBot->pBotClient->Blocks)
	{
		bool StopBlock = false;

		if (pBot->CurrentBotTask)
		{
			if (pBot->CurrentBotTask->TaskIdentify() != TaskBlock)
				StopBlock = true;
		}
		else
			StopBlock = true;

		if (StopBlock)
			FakeClientCommand(pEdict, "-block", NULL, NULL);
	}

	if (pBot->pBotClient->BlockStruggles)
	{
		MakeTask(pBot, TaskBlock);
		return;
	}

	if (pBot->CurrentBotTask)
	{
		if (pBot->CurrentBotTask->TaskIdentify() != TaskWallJump)
			CheckWJump = true;
	}
	else
		CheckWJump = true;

	if (CheckWJump)
	{
		//lcout << "Check WJump\n";
		if (CheckWallJump(pBot))
		{
			MakeTask(pBot, TaskWallJump);
			return;
		}
	}

	// This should be before the idle checking...otherwise the bot thinks he's idle
	if (pBot->AscendTime >= gpGlobals->time)
	{
		MakeTask(pBot, TaskAscend);
		return;
	}

	// Should the bot behave idle ?
	if ((BotsAreIdle) || (pBot->pBotClient->LaysOnGround))
	{
		MakeTask(pBot, TaskIdle);
		return;
	}

	if (pBot->pBotClient->GetAutoPunched)
	{
		MakeTask(pBot, TaskBlock);
		return;
	}
/*
	// This MUST be before the ps checking, otherwise the bot thinks he's in a ps
	if (pBot->pBotClient->MeleeStruggles)
	{
		MakeTask(pBot, TaskMeleeStruggle);
		return;
	}
*/
	if (pBot->pBotClient->HasThrowLanded)
	{
		MakeTask(pBot, TaskJumpOffGround);
		return;
	}

	if (pBot->pBotClient->GetSwung)
	{
		MakeTask(pBot, TaskHandleSwung);
		return;
	}

	if (pBot->pBotClient->IsAutoPunching)
	{
		MakeTask(pBot, TaskFinishMelee);
		return;
	}
	else if ((pBot->pBotClient->IsInMeleeMode) && (pEdict->v.velocity.Length() < 1.0))
	{
		if (pBot->CurrentBotTask)
		{
			// only counter melee when we don't charge a combo or throw...
			if (pBot->CurrentBotTask->TaskIdentify() != TaskFinishMelee)
				MakeTask(pBot, TaskCounterMelee);
		}
		else
			MakeTask(pBot, TaskCounterMelee);

		return;
	}

	if (pBot->bot_powerstrug < 127)
	{
		if (pBot->pBotCombo)
			DeleteCombos(pBot);

		MakeTask(pBot, TaskPowerstruggle);
		return;
	}

    if (pBot->blinded_time > gpGlobals->time)
	{
		if (pBot->pBotCombo)
			DeleteCombos(pBot);
    
		MakeTask(pBot, TaskBlinded);
		return;
	}

	if (pBot->pBotClient->Swoops)
	{
		if (pBot->CurrentBotTask)
		{
			if (pBot->CurrentBotTask->TaskIdentify() != TaskCombo)
			{
				MakeTask(pBot, TaskSwoop);
				return;
			}
		}
		else
		{
			MakeTask(pBot, TaskSwoop);
			return;
		}
	}

	if (pBot->BotPowerUps)
	{
		// stop with powerup when we don't want to powerup...
		if (pBot->CurrentBotTask)
		{
			if ((pBot->CurrentBotTask->TaskIdentify() != TaskPowerup) &&
				(pBot->CurrentBotTask->TaskIdentify() != TaskCombo))
				FakeClientCommand(pEdict,"-powerup",NULL,NULL);
		}
		else
			FakeClientCommand(pEdict,"-powerup",NULL,NULL);
	}

	if (pBot->pBotClient->ControlsAttack)
	{
		if (pBot->CurrentBotTask)
		{
			if ((pBot->CurrentBotTask->TaskIdentify() != TaskControlBeam) &&
				(pBot->CurrentBotTask->TaskIdentify() != TaskBeamJump))
				MakeTask(pBot, TaskControlBeam);
		}
		else
			MakeTask(pBot, TaskControlBeam);

		return;
	}

    if (pBot->pBotClient->FlysAfterHit)
    {
		MakeTask(pBot, TaskFlyAfterHit);
	    return;
    }
}

void CheckBotInternalStatus(bot_t *pBot)
{
	edict_t *pEdict = pBot->pEdict;

	int StatusChecks = 0;
	bool CheckEverything = false;

	if (pBot->CurrentBotTask)
	{
		StatusChecks = pBot->CurrentBotTask->GetBotStatusChecks();
	}
	else
		CheckEverything = true;

	if ((StatusChecks & CheckKi) || (CheckEverything))
	{
		if (!pBot->pBotClient->ChargesAttack)
		{
			if ((pBot->bot_ki < 1000) && (pBot->BotPowerUpDelay <= gpGlobals->time) &&
				(!pBot->BotPowerUps))
			{
				int MinKiLevel = 750;

				if (pBot->pBotEnemy)
				{
					float flDist = UTIL_Distance(pEdict->v.origin, pBot->pBotEnemy->v.origin);
					if (flDist < 1250.0f)
					{
						switch (pBot->AggressionLevel)
						{
						case Aggressive:
							MinKiLevel = 700;
							break;
						case Normal:
							MinKiLevel = 750;
							break;
						case Defensive:
							MinKiLevel = 800;
							break;
						}
					}
				}

				if (pBot->bot_skill > 0)
					MinKiLevel -= (pBot->bot_skill*80);

#ifdef USE_NEW_BOT_SENSE
				if (pBot->DangerLevel >= DANGER_MEDIUM_MIN_VAL)
				{
					if (pBot->DangerLevel <= DANGER_MEDIUM_MAX_VAL)
						MinKiLevel = 100;
					else
						MinKiLevel = 20;
				}
#else
				if (pBot->DangerLevel)
				{
					switch (pBot->DangerLevel)
					{
					case danger_light:
						MinKiLevel -= 100;
						break;
					case danger_medium:
						MinKiLevel = 100; // hurry up
						break;
					case danger_hard:
					case danger_heavy:
						MinKiLevel = 20;
						break;
					}
				}
#endif
				if (pBot->bot_ki < MinKiLevel)
				{
					if (BotShouldFallDown(pBot))
						MakeCombo(pBot, fallpowerup);
					else
					{
						MakeTask(pBot, TaskPowerup);
						FakeClientCommand(pEdict, "+powerup", NULL, NULL);
					}
					return;
				}
			}
		}
	}

	if ((StatusChecks & CheckFly) || (CheckEverything))
	{
		if (pBot->CheckFlyDelay < gpGlobals->time)
		{
			if (RANDOM_LONG(1, 100) <= pBot->FlyPercent)
			{
				pBot->ShouldFly = true;
			}
			else
			{
				pBot->ShouldFly = false;
			}

			pBot->CheckFlyDelay = gpGlobals->time + RANDOM_FLOAT(15.0, 30.0);
		}

		if (((pEdict->v.gravity > 0.1) && (pBot->f_max_speed > 0) && (pBot->ShouldFly)) ||
			((pEdict->v.gravity < 0.1) && (pBot->ShouldFly == false)))
			FakeClientCommand(pEdict,"togglefly",NULL,NULL);		
	}

	if ((StatusChecks & CheckAscend) || (CheckEverything))
	{
		if (pBot->LastAscendCheck <= gpGlobals->time)
		{
			pBot->LastAscendCheck = gpGlobals->time + 0.2; // don't check every frame
			short AscendPercent = 0;

#ifdef USE_NEW_BOT_SENSE
			if ((pBot->DangerLevel == 0) ||
			    ((pBot->DangerType == DangerEnemy) || (pBot->DangerType == DangerCurrentEnemy)))
			{	
				AscendPercent = 100;
			}
			else
			{
				if (pBot->DangerLevel <= DANGER_LIGHT_MIN_VAL)
					AscendPercent = 95;
				else if ((pBot->DangerLevel >= DANGER_MEDIUM_MIN_VAL) &&
				         (pBot->DangerLevel <= DANGER_MEDIUM_MAX_VAL))
					AscendPercent = 30;
				else
					AscendPercent = 0;
			}
#else
			if (pBot->DangerLevel == 0)
			{	
				AscendPercent = 100;
			}
			else
			{
				switch (pBot->DangerLevel)
				{
				case danger_light:
					AscendPercent = 95;
					break;
				case danger_medium:
					AscendPercent = 30;
					break;
				case danger_hard:
				case danger_heavy:
					AscendPercent = 0;
					break;
				}
			}
#endif
		    if ((RANDOM_LONG(1, 100) <= AscendPercent) && (BotCanAscend(pBot)))
			{
				if (BotShouldFallDown(pBot))
				{
					MakeCombo(pBot, fallascend);
					return;
				}
				else
           			FakeClientCommand(pEdict, "ascend", NULL, NULL);
			}
		}
	}

	if ((StatusChecks & CheckHealth) || (CheckEverything))
	{
		if ((pEdict->v.playerclass == 1) && (pEdict->v.health <= 75))
		{
			short HealthGainPercent = 100;

#ifdef USE_NEW_BOT_SENSE
			if (pBot->DangerLevel > 0)
				HealthGainPercent -= (pBot->DangerLevel / 1.5f);
#else
			HealthGainPercent -= (pBot->DangerLevel * 30);
#endif

			HealthGainPercent -= pEdict->v.health;

			if (RANDOM_LONG(1, 100) <= HealthGainPercent)
			{
				MakeTask(pBot, TaskRecoverHP);
				return;
			}
		}
	}

	return;
}

void CheckBotMiscThings(bot_t *pBot)
{
#ifdef LOGBOTINFO
	clog << "Begin MiscChecks\n";
#endif

	int MiscChecks = 0;
	bool CheckEverything = false;
	edict_t *pEdict = pBot->pEdict;

	if (pBot->CurrentBotTask)
	{
		MiscChecks = pBot->CurrentBotTask->GetBotMiscChecks();
	}
	else
		CheckEverything = true;

	if (((MiscChecks & CheckForEnemys) || (CheckEverything)) && (BotDontSearchEnemy == false))
	{
		pBot->pBotEnemy = GetBestEnemyFromList(pBot);

		if (pBot->pBotEnemy)
		{
			if (pBot->CurrentBotTask)
			{
				if (pBot->CurrentBotTask->TaskIdentify() != TaskCombat)
				{
					MakeTask(pBot, TaskCombat);
					return;
				}
			}
			else
			{
				MakeTask(pBot, TaskCombat);
				return;
			}
		}
	}
	if ((MiscChecks & CheckFollowTeamMate) || (CheckEverything))
	{
		pBot->pBotUser = BotFindUser(pBot);
		if (pBot->pBotUser)
			MakeTask(pBot, TaskFollowTeamMate);
	}
	if ((MiscChecks & CheckKameTorpLeftOvers) || (CheckEverything))
	{
		if (pBot->LastKameTorpLeftOverCheck < gpGlobals->time)
		{
			pBot->LastKameTorpLeftOverCheck = gpGlobals->time + 0.5;
			// if the bot is a goku and has kame torpedo's selected as current weapon and
			// has a enemy
			if ((pEdict->v.playerclass == 2) && (pBot->pBotEnemy != NULL) &&
				(pBot->current_weapon.iId == ESF_WEAPON_KAMETORPEDO))
			{
				edict_t *pKameTorp = pAttackManager->SearchAttackByName(pEdict, "Kame Torpedo");
				if (pKameTorp)
				{
					pEdict->v.button |= IN_ATTACK2;
				}
/*
				while((pAttackEntry) && (pAttackEntry->pAttack))
				{
					// Is this a kame torpedo and is it owned by the bot?
					if ((FStrEq(STRING(pAttackEntry->pAttack->v.classname), "Kame Torpedo")) &&
						(pAttackEntry->pAttack->v.owner == pEdict))
					{
						//lcout << "Releasing a torpedo\n";
						pEdict->v.button |= IN_ATTACK2; // release it
						break; // since you can only release one torpedo each time
					}
				}*/
			}
		}
	}
	if ((MiscChecks & CheckForChatMsg) || (CheckEverything))
	{
		// Can bots chat AND is the bot not an autopilot AND
		// does the bot not allready picked a chat msg?
		if ((should_bot_chat) &&
			(pEdict != listenserver_edict) && (pBot->f_bot_say < gpGlobals->time))
		{
			short RandNr = RANDOM_LONG(1, 100);

			// Time to look for an reply?
			if (pBot->flReplyDelay < gpGlobals->time)
			{
				if (RandNr <= pBot->chat_percent)
					pChatClass->LetBotCheckForReply(pBot);

				pBot->flReplyDelay = gpGlobals->time + RANDOM_FLOAT(5.5f, 15.0f);
			}

			// Time to say something random?
			if (pBot->f_bot_chat_time < gpGlobals->time)
			{
				if (RandNr <= pBot->chat_percent)
					pChatClass->LetBotSaySomething(pBot);

				pBot->f_bot_chat_time = gpGlobals->time + RANDOM_FLOAT(15.0f, 40.0f);
			}
		}
	}
#ifdef usedbcode
	if ((MiscChecks & CheckDragonBall) || (CheckEverything))
	{
		if (CheckDBS(pBot))
		{
			MakeTask(pBot, TaskDB);
			return;
		}
	}
#endif

#ifdef LOGBOTINFO
	clog << "End Misc checks\n";
#endif
}

void BotPerformTasks(bot_t *pBot)
{
#ifdef LOGBOTINFO
		clog << "Begin task perform\n";
#endif

	if (pBot->CurrentBotTask == NULL)
		return;

	if (pBot->CurrentBotTask->GetTaskState() == done)
	{
		KillCurrentTask(pBot);

		// hack hack... Since its not possible to create a task in a task, we do it here
		if (pBot->TaskAfterTask != TaskNone)
		{
			MakeTask(pBot, pBot->TaskAfterTask);
			pBot->TaskAfterTask = TaskNone;
		}
		return;
	}

	if (pBot->CurrentBotTask->GetTaskDelay() <= gpGlobals->time)
	{
		if (pBot->CurrentBotTask->ShouldManageEnemys())
		{
			edict_t *pTemp = pBot->pBotEnemy;
			pBot->pBotEnemy = GetBestEnemyFromList(pBot);

			// Did the bot lost his enemy ?
			if ((pBot->pBotEnemy == NULL) && (UTIL_IsValidPlayer(pTemp, false)))
			{
				// Add delay
				extern float EnemySearchDelays[5][2];
				float MinDelay = EnemySearchDelays[pBot->bot_skill][0];
				float MaxDelay = EnemySearchDelays[pBot->bot_skill][1];

				pBot->f_search_enemy_delay = gpGlobals->time + RANDOM_FLOAT(MinDelay, MaxDelay);
			}
		}

		if (pBot->CurrentBotSubTask)
			pBot->CurrentBotSubTask->TaskThink();

#ifdef LOGBOTINFO
		clog << "Begin task think\n";
		clog << "Task is: " << pBot->CurrentBotTask->TaskIdentify() << "\n";
#endif

		pBot->CurrentBotTask->TaskThink();

#ifdef LOGBOTINFO
		clog << "End task think\n";
#endif

	}
	else
	{
		ResetMoveSpeed(pBot);
	}

	if (pBot->CurrentBotTask->GetTaskState() == done)
	{
		KillCurrentTask(pBot);

		// hack hack... Since its not possible to create a task in a task, we do it here
		if (pBot->TaskAfterTask != TaskNone)
		{
			MakeTask(pBot, pBot->TaskAfterTask);
			pBot->TaskAfterTask = TaskNone;
		}
	}

#ifdef LOGBOTINFO
		clog << "End task perform\n";
#endif
}



#ifdef classstuff

void CBot::MakeTask(ETaskCode TaskCode)
{
#ifdef LOGBOTINFO
	clog << "New task: " << TaskCode << "\n";
#endif

	if (m_pCurrentBotTask)
	{
		if (m_pCurrentBotTask->TaskIdentify() == TaskCode)
			return; // don't make the task when its already made...

		m_ePrevTask = m_pCurrentBotTask->TaskIdentify();
		KillCurrentTask();
	}
#ifdef LOGBOTINFO
	clog << "Making new task: " << TaskCode << "\n";
#endif

/*
	char s[25] = "";
	sprintf(s, "New task: %i\n", int(TaskCode));
	ClientPrint(listenserver_edict, HUD_PRINTTALK, s);
*/
/*
	try{
	switch (TaskCode)
	{
	case TaskPowerup:
		m_pCurrentBotTask = new CPowerUpTask(pBot);
		break;
	case TaskTeleport:
		m_pCurrentBotTask = new CTeleportTask(pBot);
		m_pCurrentBotSubTask = new CAimAtEnemySubTask(pBot);
		break;
	case TaskBlock:
		m_pCurrentBotTask = new CBlockTask(pBot);
		m_pCurrentBotSubTask = new CAimAtEnemySubTask(pBot);
		break;
	case TaskNormalMovement:
		m_pCurrentBotTask = new CNormalMovementTask(pBot);
		break;
	case TaskCombat:
		m_pCurrentBotTask = new CCombatTask(pBot);
		m_pCurrentBotSubTask = new CCombatMovementSubTask(pBot);
		break;
	case TaskChargeAttack:
		m_pCurrentBotTask = new CChargeAttackTask(pBot);
		m_pCurrentBotSubTask = new CCombatMovementSubTask(pBot);
		break;
	case TaskSwoop:
		m_pCurrentBotTask = new CSwoopTask(pBot);
		break;
	case TaskFollowTeamMate:
		m_pCurrentBotTask = new CFollowTeamMateTask(pBot);
		break;
	case TaskControlFriezaDisc:
		m_pCurrentBotTask = new CControlFriezaDiscTask(pBot);
		m_pCurrentBotSubTask = new CCombatMovementSubTask(pBot);
		break;
	case TaskControlBeam:
		m_pCurrentBotTask = new CControlBeamTask(pBot);
		break;
	case TaskAscend:
		m_pCurrentBotTask = new CAscendTask(pBot);
		break;
	case TaskPowerstruggle:
		m_pCurrentBotTask = new CPowerStruggleTask(pBot);
		break;
	case TaskCombo:
		m_pCurrentBotTask = new CComboTask(pBot);
		break;
	case TaskLookAround:
		m_pCurrentBotTask = new CLookAroundTask(pBot);
		break;
	case TaskRecoverHP:
		m_pCurrentBotTask = new CRecoverHPTask(pBot);
		break;
	case TaskMoveDir:
		m_pCurrentBotTask = new CMoveDirTask(pBot);
		m_pCurrentBotSubTask = new CAimAtEnemySubTask(pBot);
		break;
	case TaskBlinded:
		m_pCurrentBotTask = new CBlindedTask(pBot);
		break;
	case TaskFinishMelee:
		m_pCurrentBotTask = new CFinishMeleeTask(pBot);
		break;
	case TaskKickEnemy:
		m_pCurrentBotTask = new CKickEnemyTask(pBot);
		break;
	case TaskCounterMelee:
		m_pCurrentBotTask = new CCounterMeleeTask(pBot);
		break;
	case TaskWallJump:
		m_pCurrentBotTask = new CWallJumpTask(pBot);
		break;
	case TaskKameTorp:
		m_pCurrentBotTask = new CKameTorpTask(pBot);
		break;
	case TaskMeleeStruggle:
		m_pCurrentBotTask = new CMeleeStruggleTask(pBot);
		break;
	case TaskIdle:
		m_pCurrentBotTask = new CIdleTask(pBot);
		break;
	case TaskFlyAfterHit:
		m_pCurrentBotTask = new CFlyAfterHitTask(pBot);
		break;
	case TaskBeamJump:
		m_pCurrentBotTask = new CBeamJumpTask(pBot);
		break;
	case TaskNormalJump:
		m_pCurrentBotTask = new CNormalJumpTask(pBot);
		break;
	case TaskShootCandy:
		m_pCurrentBotTask = new CShootCandyTask(pBot);
		break;

	// exception: tasks get made elsewhere
	case TaskSwoopAway:
		MakeAvoidSwoopTask(pBot);
		break;

	// tasks for super bots
	case TaskShootLaser1:
		m_pCurrentBotTask = new CShootLaser1Task(pBot);
		break;
	case TaskKeepDangerAway:
		m_pCurrentBotTask = new CKeepDangerAwayTask(pBot);
		break;
	case TaskAbsorbHealth:
		m_pCurrentBotTask = new CAbsorbHealthTask(pBot);
		break;

	}
    }
	catch (exception ex)
	{
		lcout << "Encountered exception in MakeTask; " << ex.what() << "\n";
		clog << "Encountered exception in MakeTask; " << ex.what() << "\n";
		m_pCurrentBotTask = NULL;
		m_pCurrentBotSubTask = NULL;
	}
	catch(...)
	{
		lcout << "Error creating task: " << TaskCode << "\n";
		clog << "Error creating task: " << TaskCode << "\n";
		m_pCurrentBotTask = NULL;
		m_pCurrentBotSubTask = NULL;
	}

#ifdef _DEBUG
	if (!m_pCurrentBotTask)
	{
		lcout << "Task " << TaskCode << " not found !!\n";
		clog << "Task " << TaskCode << " not found !!\n";
	}
#endif
*/
#ifdef LOGBOTINFO
	clog << "Done making new task: " << TaskCode << "\n";
#endif
}

void CBot::KillCurrentTask()
{
#ifdef LOGBOTINFO
	clog << "Begin kill task\n";
#endif

	if (m_pCurrentBotTask)
	{
		delete m_pCurrentBotTask;
		m_pCurrentBotTask = NULL;
	}

	if (m_pCurrentBotSubTask)
	{
		delete m_pCurrentBotSubTask;
		m_pCurrentBotSubTask = NULL;
	}

#ifdef LOGBOTINFO
	clog << "End kill task\n";
#endif
}

void CBot::CheckNecessarilyThings()
{
	bool CheckWJump = false;

	if (Blocks())
	{
		bool StopBlock = false;

		if (m_pCurrentBotTask)
		{
			if (m_pCurrentBotTask->TaskIdentify() != TaskBlock)
				StopBlock = true;
		}
		else
			StopBlock = true;

		if (StopBlock)
			FakeClientCommand(pEdict, "-block", NULL, NULL);
	}

	if (BlockStruggles())
	{
		MakeTask(TaskBlock);
		return;
	}

	if (m_pCurrentBotTask)
	{
		if (m_pCurrentBotTask->TaskIdentify() != TaskWallJump)
			CheckWJump = true;
	}
	else
		CheckWJump = true;

	if (CheckWJump)
	{
		//lcout << "Check WJump\n";
		if (CheckWallJump())
		{
			MakeTask(TaskWallJump);
			return;
		}
	}

	// Should the bot behave idle ?
	if ((BotsAreIdle) || (LaysOnGround()) || (pEdict->v.flags & FL_FROZEN))
	{
		MakeTask(TaskIdle);
		return;
	}

	if (GetAutoPunched())
	{
		MakeTask(TaskBlock);
		return;
	}

	// This MUST be before the ps checking, otherwise the bot thinks he's in a ps
	if (MeleeStruggles())
	{
		MakeTask(TaskMeleeStruggle);
		return;
	}

	if (IsAutoPunching())
	{
		MakeTask(TaskFinishMelee);
		return;
	}
	else if ((IsInMeleeMode()) && (pEdict->v.velocity.Length() < 1.0))
	{
		if (m_pCurrentBotTask)
		{
			// only counter melee when we don't charge a combo or throw...
			if (m_pCurrentBotTask->TaskIdentify() != TaskFinishMelee)
				MakeTask(TaskCounterMelee);
		}
		else
			MakeTask(TaskCounterMelee);

		return;
	}

	if (Powerstruggles())
	{
		MakeTask(TaskPowerstruggle);
		return;
	}

	if (Ascends())
	{
		MakeTask(TaskAscend);
		return;
	}

    if (IsBlinded())
	{
		if (m_pBotCombo)
			DeleteCombos();
    
		MakeTask(TaskBlinded);
		return;
	}

	if (Swoops())
	{
		if (m_pCurrentBotTask)
		{
			if (m_pCurrentBotTask->TaskIdentify() != TaskCombo)
			{
				MakeTask(TaskSwoop);
				return;
			}
		}
		else
		{
			MakeTask(TaskSwoop);
			return;
		}
	}

	if (PowerUps())
	{
		if (m_pCurrentBotTask)
		{
			// stop with powerup when we don't want to powerup...
			if ((m_pCurrentBotTask->TaskIdentify() != TaskPowerup) &&
				(m_pCurrentBotTask->TaskIdentify() != TaskCombo))
				FakeClientCommand(pEdict,"-powerup",NULL,NULL);
		}
		else
			FakeClientCommand(pEdict,"-powerup",NULL,NULL);
	}

	if (ControlsAttack())
	{
		if (m_pCurrentBotTask)
		{
			if ((m_pCurrentBotTask->TaskIdentify() != TaskControlBeam) &&
				(m_pCurrentBotTask->TaskIdentify() != TaskBeamJump))
				MakeTask(TaskControlBeam);
		}
		else
			MakeTask(TaskControlBeam);

		return;
	}

   if (FlysAfterHit())
   {
	   MakeTask(TaskFlyAfterHit);
	   return;
   }
}

void CBot::CheckBotInternalStatus()
{
	int StatusChecks = 0;
	bool CheckEverything = false;

	if (m_pCurrentBotTask)
	{
		StatusChecks = m_pCurrentBotTask->GetBotStatusChecks();
	}
	else
		CheckEverything = true;

	if ((StatusChecks & CheckKi) || (CheckEverything))
	{
		if (!ChargesAttack())
		{
			if ((GetKi() < 1000) && (m_fBotPowerUpDelay <= gpGlobals->time) &&
				(PowerUps() == false))
			{
				int MinKiLevel = 750;

				if (m_pBotEnemy)
				{
					switch (m_pPersonality->aggression_level)
					{
					case Aggressive:
						MinKiLevel = 550;
						break;
					case Normal:
						MinKiLevel = 700;
						break;
					case Defensive:
						MinKiLevel = 850;
						break;
					}
				}

				if (m_bBotSkill > 0)
					MinKiLevel -= (m_bBotSkill*80);

				if (m_bDangerLevel)
				{
					switch (m_bDangerLevel)
					{
					case danger_light:
						MinKiLevel -= 100;
						break;
					case danger_medium:
						MinKiLevel = 100; // hurry up
						break;
					case danger_hard:
					case danger_heavy:
						MinKiLevel = 20;
						break;
					}
				}

				if (GetKi() < MinKiLevel)
				{
					if (BotShouldFallDown())
						MakeCombo(fallpowerup);
					else
					{
						MakeTask(TaskPowerup);
						FakeClientCommand(pEdict, "+powerup", NULL, NULL);
						m_fMoveSpeed = 0.0;
					}
					return;
				}
			}
		}
	}

	if ((StatusChecks & CheckFly) || (CheckEverything))
	{
		if (m_fCheckFlyDelay < gpGlobals->time)
		{
			if (RANDOM_LONG(1, 100) <= m_pPersonality->fly_percent)
			{
				m_bShouldFly = true;
			}
			else
			{
				m_bShouldFly = false;
			}

			m_fCheckFlyDelay = gpGlobals->time + RANDOM_FLOAT(15.0, 30.0);
		}

		if (((pEdict->v.gravity > 0.1) && (GetMaxSpeed() > 0) && (m_bShouldFly)) ||
			((pEdict->v.gravity < 0.1) && (m_bShouldFly == false)))
			FakeClientCommand(pEdict,"togglefly",NULL,NULL);		
	}

	if ((StatusChecks & CheckAscend) || (CheckEverything))
	{
		if (m_fLastAscendCheck <= gpGlobals->time)
		{
			m_fLastAscendCheck = gpGlobals->time + 0.2; // don't check every frame
			short AscendPercent = 0;

			if (m_bDangerLevel == 0)
			{	
				AscendPercent = 100;
			}
			else
			{
				switch (m_bDangerLevel)
				{
				case danger_light:
					AscendPercent = 95;
					break;
				case danger_medium:
					AscendPercent = 30;
					break;
				case danger_hard:
				case danger_heavy:
					AscendPercent = 0;
					break;
				}
			}

		    if ((BotCanAscend()) && (RANDOM_LONG(1, 100) <= AscendPercent))
			{
				if (BotShouldFallDown())
				{
					MakeCombo(fallascend);
					return;
				}
				else
           			FakeClientCommand(pEdict, "ascend", NULL, NULL);
			}
		}
	}

	if ((StatusChecks & CheckHealth) || (CheckEverything))
	{
		if ((pEdict->v.playerclass == 1) && (pEdict->v.health <= 75))
		{
			short HealthGainPercent = 100;

			HealthGainPercent -= (m_bDangerLevel * 30);

			HealthGainPercent -= pEdict->v.health;

			if (RANDOM_LONG(1, 100) <= HealthGainPercent)
			{
				MakeTask(TaskRecoverHP);
				return;
			}
		}
	}

	return;
}

void CBot::CheckBotMiscThings()
{
#ifdef LOGBOTINFO
	clog << "Begin MiscChecks\n";
#endif

	int MiscChecks = 0;
	bool CheckEverything = false;

	if (m_pCurrentBotTask)
	{
		MiscChecks = m_pCurrentBotTask->GetBotMiscChecks();
	}
	else
		CheckEverything = true;

	if (((MiscChecks & CheckForEnemys) || (CheckEverything)) && (BotDontSearchEnemy == false))
	{
		m_pBotEnemy = GetBestEnemyFromList();

		if (m_pBotEnemy)
		{
			if (m_pCurrentBotTask)
			{
				if (m_pCurrentBotTask->TaskIdentify() != TaskCombat)
				{
					MakeTask(TaskCombat);
					return;
				}
			}
			else
			{
				MakeTask(TaskCombat);
				return;
			}
		}
	}
	if ((MiscChecks & CheckFollowTeamMate) || (CheckEverything))
	{
		m_pBotUser = BotFindUser();
		if (m_pBotUser)
			MakeTask(TaskFollowTeamMate);
	}
	if ((MiscChecks & CheckKameTorpLeftOvers) || (CheckEverything))
	{
		if (m_fLastKameTorpLeftOverCheck < gpGlobals->time)
		{
			m_fLastKameTorpLeftOverCheck = gpGlobals->time + 0.5;
			// if the bot is a goku and has kame torpedo's selected as current weapon and
			// has a enemy
			if ((GetClass() == 2) && (m_pBotEnemy != NULL) &&
				(m_CurrentWeapon.iId == ESF_WEAPON_KAMETORPEDO))
			{
				extern AttackList_t *pAttackList;
				AttackList_t *pAttackEntry = pAttackList;

				while((pAttackEntry) && (pAttackEntry->pAttack))
				{
					// Is this a kame torpedo and is it owned by the bot?
					if ((FStrEq(STRING(pAttackEntry->pAttack->v.classname), "Kame Torpedo")) &&
						(pAttackEntry->pAttack->v.owner == pEdict))
					{
						//lcout << "Releasing a torpedo\n";
						pEdict->v.button |= IN_ATTACK2; // release it
						break; // since you can only release one torpedo each time
					}
				}
			}
		}
	}
#ifdef LOGBOTINFO
	clog << "End Misc checks\n";
#endif
}

void CBot::BotPerformTasks()
{
#ifdef LOGBOTINFO
		clog << "Begin task perform\n";
#endif

	if (m_pCurrentBotTask == NULL)
		return;

	if (m_pCurrentBotTask->GetTaskDelay() <= gpGlobals->time)
	{
		if (m_pCurrentBotTask->ShouldManageEnemys())
		{
			CClient *pTemp = m_pBotEnemy;
			m_pBotEnemy = GetBestEnemyFromList();

			// Did the bot lost his enemy ?
			if ((m_pBotEnemy == NULL) && (UTIL_IsValidPlayer(pTemp, false)))
			{
				// Add delay
				extern float EnemySearchDelays[5][2];
				float MinDelay = EnemySearchDelays[m_bBotSkill][0];
				float MaxDelay = EnemySearchDelays[m_bBotSkill][1];

				m_fSearchEnemyDelay = gpGlobals->time + RANDOM_FLOAT(MinDelay, MaxDelay);
			}
		}

		if (m_pCurrentBotSubTask)
			m_pCurrentBotSubTask->TaskThink();

#ifdef LOGBOTINFO
		clog << "Begin task think\n";
		clog << "Task is: " << m_pCurrentBotTask->TaskIdentify() << "\n";
#endif

		m_pCurrentBotTask->TaskThink();

#ifdef LOGBOTINFO
		clog << "End task think\n";
#endif

	}
	else
	{
		m_fMoveSpeed = 0.0;
	}

	if (m_pCurrentBotTask->GetTaskState() == done)
	{
		KillCurrentTask();

		// hack hack... Since its not possible to create a task in a task, we do it here
		if (m_eTaskAfterTask != TaskNone)
		{
			MakeTask(m_eTaskAfterTask);
			m_eTaskAfterTask = TaskNone;
		}
	}

#ifdef LOGBOTINFO
		clog << "End task perform\n";
#endif
}

#endif
//////////////////////////////////////////////////////
//				Function code END				    //
//////////////////////////////////////////////////////