#ifndef TASK_H
#define TASK_H

// Status checks
#define CheckFly		(1<<1)
#define CheckKi			(1<<2)
#define CheckAscend		(1<<3)
#define CheckHealth		(1<<4)

// Sense checks
#define CheckSounds					(1<<0)
#define CheckAttacks				(1<<1)
#define CheckSurroundingPlayers		(1<<2)
#define CheckHit					(1<<3)

// Misc checks
#define CheckForEnemys				(1<<1)
#define CheckFollowTeamMate			(1<<2)
#define CheckKameTorpLeftOvers		(1<<3)
#define CheckForChatMsg				(1<<4)
#define CheckDragonBall				(1<<5)

extern float ExplodeBeamDistances[5][2];

enum ETaskCode
{
	TaskPowerup = 0,
	TaskTeleport,
	TaskBlock,
	TaskKickEnemy,
	TaskCounterMelee,
	TaskCombat,
	TaskNormalMovement,
	TaskSwoop,
	TaskFinishMelee,
	TaskPowerstruggle,
	TaskAscend,
	TaskBlinded,
	TaskCombo,
	TaskChargeAttack,
	TaskControlFriezaDisc,
	TaskControlBeam,
	TaskFollowTeamMate,
	TaskLookAround,
	TaskRecoverHP,
	TaskMoveDir,
	TaskWallJump,
	TaskKameTorp,
	TaskSwoopAway,
	TaskMeleeStruggle,
	TaskIdle,
	TaskFlyAfterHit,
	TaskBeamJump,
	TaskNormalJump,
	TaskShootCandy,
	TaskShootScatterShots,
	TaskShootNonChargeble,
	TaskDB,
	TaskGoToVec,
	TaskJumpOffGround,
	TaskHandleSwung,

	// for sbots
	TaskShootLaser1,
	TaskKeepDangerAway,
	TaskAbsorbHealth,

	TaskNone
};

enum ComboTypes;

class CBaseTask
{
public:
	// member variabeles
	int NesseserelyChecks;
	int BotStatusChecks;
	int BotSenseChecks;
	int BotMiscChecks;
	float Delay;
	State TaskState;
	bot_t *pBot;
	edict_t *pEdict;

	CBaseTask(bot_t *pB); // constructor
	virtual ~CBaseTask(void); // destructor

	// member functions
	float GetTaskDelay(void)		{ return Delay; };
	short GetTaskState(void)		{ return TaskState; };
	int GetNesserelyChecks(void)	{ return 0; };
	int GetBotStatusChecks(void)	{ return BotStatusChecks; };
	int GetBotSenseChecks(void)		{ return BotSenseChecks; };
	int GetBotMiscChecks(void)		{ return BotMiscChecks; };

	void MakeCombo(ComboTypes CT);
	void GoToAnPreviousLocation(void);
	void SetGeneralDelay(void);

	virtual void TaskThink(void) = 0;
	virtual bool KillTaskOnDanger(short DangerLevel) = 0;
	virtual ETaskCode TaskIdentify(void) = 0;
	virtual bool ShouldManageEnemys(void) = 0;
	virtual void ManageEnemies(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return true; };
};

class CFocusedTask : public CBaseTask
{
public:
	CFocusedTask(bot_t *pB) : CBaseTask(pB)//, BotStatusChecks(0), BotSenseChecks(0), BotMiscChecks(0)
	{};
	virtual ~CFocusedTask(void) {};

	virtual void Think(void) = 0;
	virtual bool KillTaskOnDanger(short DangerLevel) { return false; };
	virtual void ManageEnemies(void) {};
};

class CPowerUpTask: public CBaseTask
{
public:
	CPowerUpTask(bot_t *pB);
	virtual ~CPowerUpTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CTeleportTask: public CBaseTask
{
public:
	CTeleportTask(bot_t *pB);
	virtual ~CTeleportTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CBlockTask: public CBaseTask
{
public:
	float BlockTime;

	CBlockTask(bot_t *pB);
	virtual ~CBlockTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CKickEnemyTask: public CBaseTask
{
public:
	float LifeTime;

	CKickEnemyTask(bot_t *pB);
	virtual ~CKickEnemyTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return (pDangerInfo->DangerValue >= DANGER_MEDIUM_MIN_VAL); };
};

class CCounterMeleeTask: public CBaseTask
{
	int ComboChargeButton;
	float PunchKickTime;
	float NewKickPunchButtonDelay;

public:

	CCounterMeleeTask(bot_t *pB);
	virtual ~CCounterMeleeTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CCombatTask: public CBaseTask  
{
public:
	CCombatTask(bot_t *pB);
	virtual ~CCombatTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return (pDangerInfo->DangerValue >= DANGER_MEDIUM_MIN_VAL); };

	void CombatThink(void);
	float GetMinDistForAttack(int AttackId);
	float GetMaxDistForAttack(int AttackId);
	void PrepareAttack(int AttackId, EComboUse ComboUse);
	void AttackWithMelee(bool MakeACombo);
	short GetAttackPercent(edict_t *pEnemy, int iId);
	int GetAttackIndex(int AttackId);
	bool CanUseSecondaryFire(int AttackId);
};

class CNormalMovementTask: public CBaseTask  
{
public:
	CNormalMovementTask(bot_t *pB);
	virtual ~CNormalMovementTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
};

class CSwoopTask: public CBaseTask  
{
	bool CheckedSecondary;
	bool HoldSecondary;
	short SwoopOffsetXFactor;
	short SwoopOffsetYFactor;
	float SwoopOffsetXTarget;
	float SwoopOffsetYTarget;

public:

	CSwoopTask(bot_t *pB);
	virtual ~CSwoopTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };

	void CheckSecondaryFire(void);
	void SelectSwoopType(void);
};

class CPowerStruggleTask: public CBaseTask  
{
public:
	CPowerStruggleTask(bot_t *pB);
	virtual ~CPowerStruggleTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CFinishMeleeTask: public CBaseTask  
{
	int ComboChargeButton;
	float NewKickPunchButtonDelay;
	float PunchKickTime;
	float FinishMeleeTime;
	short FinishMeleeType;
	byte PrepunchAmountGoal;

	edict_t *pPlayerToThrowAt;
	Vector ObstacleToThrowAt;
	bool CheckedObstacle;

public:

	CFinishMeleeTask(bot_t *pB);
	virtual ~CFinishMeleeTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };

	void CheckPlayerToThrowAt(void);
	void CheckObstacleToThrowAt(void);
	bool IsValidThrowVictim(edict_t *pPlayer);
};

class CAscendTask: public CBaseTask  
{
public:
	CAscendTask(bot_t *pB);
	virtual ~CAscendTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CBlindedTask: public CBaseTask  
{
public:
	CBlindedTask(bot_t *pB);
	virtual ~CBlindedTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CComboTask: public CBaseTask  
{
public:
	CComboTask(bot_t *pB);
	virtual ~CComboTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CChargeAttackTask: public CBaseTask
{
	float EnemyMovingFastTime; // Time that the enemy is moving very fast
	float EnemyMovingSlowTime; // Time that the enemy is moving very slow

public:
	CChargeAttackTask(bot_t *pB);
	virtual ~CChargeAttackTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CShootNonChargebleTask: public CBaseTask
{
public:
	CShootNonChargebleTask(bot_t *pB);
	virtual ~CShootNonChargebleTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CShootScatterShotsTask: public CBaseTask
{
	float ShootTime;
	float StartTime;
	bool UseScatterCheat; // Not really a cheat, 
					      // more using a bug to let scatter shots fire and home at once :)

public:
	CShootScatterShotsTask(bot_t *pB);
	virtual ~CShootScatterShotsTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};
	
class CControlFriezaDiscTask: public CBaseTask
{
	float EnemyMovingFastTime; // Time that the enemy is moving very fast
	float EnemyMovingSlowTime; // Time that the enemy is moving very slow
	float RealDiscTime;
	float ControlDiscTime;

public:

	CControlFriezaDiscTask(bot_t *pB);
	virtual ~CControlFriezaDiscTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return (pDangerInfo->DangerValue >= DANGER_MEDIUM_MIN_VAL); };
};

class CControlBeamTask: public CBaseTask
{
	float ControlBeamTime;
	bool BotShouldAimAtAttack;

public:
	CControlBeamTask(bot_t *pB);
	virtual ~CControlBeamTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CFollowTeamMateTask: public CBaseTask
{
public:

	CFollowTeamMateTask(bot_t *pB);
	virtual ~CFollowTeamMateTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
};

class CLookAroundTask: public CBaseTask
{
public:
	Vector LookAroundSpots[4];
	short Stage;
	float LookAroundDelay;

	CLookAroundTask(bot_t *pB);
	virtual ~CLookAroundTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
};

// for a buu bot
class CRecoverHPTask: public CBaseTask
{
public:
	CRecoverHPTask(bot_t *pB);
	virtual ~CRecoverHPTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CMoveDirTask: public CBaseTask
{
public:
	CMoveDirTask(bot_t *pB);
	virtual ~CMoveDirTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CWallJumpTask: public CBaseTask
{
	float LifeTime;
	float StartChargingTime;

public:
	CWallJumpTask(bot_t *pB);
	virtual ~CWallJumpTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CKameTorpTask: public CBaseTask
{
	enum EKameTorpState
	{
		STATE_CHARGING,
		STATE_POWERUP,
		STATE_RELEASING,
		STATE_MOVE,
		STATE_LOOK_AROUND
	};

	float MoveAwayTime;
	int MoveAwayDir;
	float LookDelay;
	float MoveAwayDelay;
	bool PickedAnAimSpot;

public:
	EKameTorpState KameTorpState;
	float LifeTime;
	float ShootLifeTime;
	short DesiredTorpCount;
	bool ShootingTorps;

	CKameTorpTask(bot_t *pB);
	virtual ~CKameTorpTask(void);

	void ChargeKameTorps(void);
	bool SetMoveAwayDir(void); // returns true when found a direction to mave at
	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CMeleeStruggleTask: public CBaseTask
{
public:
	float ChargeTime;
	float RestTime;
	short TargetChargeVal;

	CMeleeStruggleTask(bot_t *pB);
	virtual ~CMeleeStruggleTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CIdleTask: public CBaseTask
{
public:
	CIdleTask(bot_t *pB);
	virtual ~CIdleTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CFlyAfterHitTask: public CBaseTask
{
public:
	CFlyAfterHitTask(bot_t *pB);
	virtual ~CFlyAfterHitTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CBeamJumpTask: public CBaseTask
{
	float LifeTime;
	float TraceDelay;
	float CreateTime;

public:
	CBeamJumpTask(bot_t *pB);
	virtual ~CBeamJumpTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo);
};

class CNormalJumpTask: public CBaseTask
{
	float LifeTime;
	float TeleCheckDelay;

public:

	CNormalJumpTask(bot_t *pB);
	virtual ~CNormalJumpTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CShootCandyTask: public CBaseTask
{
	float LifeTime;
	float TeleCheckTime;
	float MoveBackTime;
	float CookieCheckDelay;
	float MoveToCookieTime;
	float EnemyIsInRangeTime;

public:

	CShootCandyTask(bot_t *pB);
	virtual ~CShootCandyTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return (pDangerInfo->DangerValue >= DANGER_MEDIUM_MIN_VAL); };
};

class CDBTask: public CBaseTask
{
	float flDBThrowDelay;

	void HeadToNextDBWP(void);
	bool GoToDBSpot(void);
	bool GoToGatheredSpot(void);
	void GoToDB(void);
	bool CheckIfThrowToDBSpot(void);

public:
	CDBTask(bot_t *pB);
	virtual ~CDBTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
};

class CGoToVecTask: public CBaseTask
{
public:
	CGoToVecTask(bot_t *pB) : CBaseTask(pB) {};
	virtual ~CGoToVecTask(void) {};

	virtual void TaskThink(void) { if (!GoToVec(pBot, pBot->v_GoToVector)) TaskState = done; };
	virtual bool KillTaskOnDanger(short DangerLevel) { return false; };
	virtual ETaskCode TaskIdentify(void) { return TaskGoToVec; };
	virtual bool ShouldManageEnemys(void) { return false; };
};

class CJumpOffGroundTask: public CBaseTask
{
	int MoveDir;

public:

	CJumpOffGroundTask(bot_t *pB);
	virtual ~CJumpOffGroundTask(void) {};

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel) { return false; };
	virtual ETaskCode TaskIdentify(void) { return TaskJumpOffGround; };
	virtual bool ShouldManageEnemys(void) { return false; };
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};

class CHandleSwungTask: public CBaseTask
{
public:
	CHandleSwungTask(bot_t *pB);
	virtual ~CHandleSwungTask(void) {};

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel) { return false; };
	virtual ETaskCode TaskIdentify(void) { return TaskHandleSwung; };
	virtual bool ShouldManageEnemys(void) { return false; };
	virtual bool ReactOnDanger(SDangerInfo *pDangerInfo) { return false; };
};


// following tasks are for super bots

class CShootLaser1Task: public CBaseTask
{
public:
	float ShootTime;
	CShootLaser1Task(bot_t *pB);
	virtual ~CShootLaser1Task(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
};

class CKeepDangerAwayTask: public CBaseTask
{
public:
	float LifeTime;

	CKeepDangerAwayTask(bot_t *pB);
	virtual ~CKeepDangerAwayTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
};

class CAbsorbHealthTask: public CBaseTask
{
public:
	float LifeTime;

	CAbsorbHealthTask(bot_t *pB);
	virtual ~CAbsorbHealthTask(void);

	virtual void TaskThink(void);
	virtual bool KillTaskOnDanger(short DangerLevel);
	virtual ETaskCode TaskIdentify(void);
	virtual bool ShouldManageEnemys(void);
};

// subtasks

class CBaseSubTask
{
public:
	// member variabeles
	bot_t *pBot;
	edict_t *pEdict;

	CBaseSubTask(bot_t *pB); // constructor
	virtual ~CBaseSubTask(void); // destructor

	// member functions
	virtual void TaskThink(void) = 0;
};

class CAimAtEnemySubTask: public CBaseSubTask
{
public:
	CAimAtEnemySubTask(bot_t *pB);
	virtual void TaskThink(void);
};

class CCombatMovementSubTask: public CBaseSubTask
{
public:
	CCombatMovementSubTask(bot_t *pB);
	virtual void TaskThink(void);
};

class CManageEnemySubTask: public CBaseSubTask
{
public:
	CManageEnemySubTask(bot_t *pB);
	virtual void TaskThink(void);
};

// Functions
void MakeTask(bot_t *pBot, ETaskCode TaskCode);
void KillCurrentTask(bot_t *pBot);
void CheckNesseserelyThings(bot_t *pBot);
void CheckBotInternalStatus(bot_t *pBot);
void CheckBotMiscThings(bot_t *pBot);
void BotPerformTasks(bot_t *pBot);

#endif // TASK_H