/***
*
*  Copyright (c) 1999, Valve LLC. All rights reserved.
*
*  This product contains software technology licensed from Id 
*  Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*  All Rights Reserved.
*
*   Use, distribution, and modification of this source code and/or resulting
*   object code is restricted to non-commercial enhancements to products from
*   Valve LLC.  All other use, distribution, or modification is prohibited
*   without written permission from Valve LLC.
*
****/

//
// HPB_bot - botman's High Ping Bastard bot
//
// (http://planethalflife.com/botman/)
//
// util.cpp
//

#include "general_header.h"
#include "studio.h"

extern edict_t *listenserver_edict;
extern int m_spriteTexture;

extern bot_t bots[32];
extern int num_teams;
extern bool b_observer_mode;
extern bool debug_mode;
extern bool DrawTraceLines;
extern bool borg_mode;

extern ClientInfo_t ClientInfo[32];

#ifdef _DEBUG
extern int TraceLineCalls;
extern int FVisibleCalls;
#endif

extern float num_bots_check;

extern bool IsPlayingMessage;

extern bool fill_server;
extern bool fill_half;
extern int fill_team;

int gmsgTextMsg = 0;
int gmsgSayText = 0;
int gmsgShowMenu = 0;
int gmsgMOTD = 0;
int gmsgServerName = 0;

void CmdStart( const edict_t *player, const struct usercmd_s *cmd, unsigned int random_seed );
void CmdEnd ( const edict_t *player );

Vector UTIL_VecToAngles( const Vector &vec )
{
   float rgflVecOut[3];
   VEC_TO_ANGLES(vec, rgflVecOut);
   return Vector(rgflVecOut);
}


void UTIL_DrawBeam(edict_t *pEntity, Vector start, Vector end, int width,
        int noise, int red, int green, int blue, int brightness, int speed, int life)
{
   if (IsPlayingMessage)
   {
	   clog << "Tried to draw a beam when there is already a message active\n";
	   return;
   }

   MESSAGE_BEGIN(MSG_ONE, SVC_TEMPENTITY, NULL, pEntity);
   WRITE_BYTE( TE_BEAMPOINTS);
   WRITE_COORD(start.x);
   WRITE_COORD(start.y);
   WRITE_COORD(start.z);
   WRITE_COORD(end.x);
   WRITE_COORD(end.y);
   WRITE_COORD(end.z);
   WRITE_SHORT( m_spriteTexture );
   WRITE_BYTE( 1 ); // framestart
   WRITE_BYTE( 10 ); // framerate
   WRITE_BYTE( life ); // life in 0.1's
   WRITE_BYTE( width ); // width
   WRITE_BYTE( noise );  // noise

   WRITE_BYTE( red );   // r, g, b
   WRITE_BYTE( green );   // r, g, b
   WRITE_BYTE( blue );   // r, g, b

   WRITE_BYTE( brightness );   // brightness
   WRITE_BYTE( speed );    // speed
   MESSAGE_END();
}

void UTIL_DrawBeamToAll(Vector start, Vector end, int width, int noise, int red, int green, int blue, 
						int brightness, int speed)
{
	edict_t *pPlayer = NULL;

	for (int i=1;i<=gpGlobals->maxClients;i++)
	{
		pPlayer = INDEXENT(i);

		if (IsNull(pPlayer))
			continue;

		if (pPlayer->v.flags & FL_FAKECLIENT)
			continue; // don't bother about bots...

		UTIL_DrawBeam(pPlayer, start, end, width, noise, red, green, blue, brightness, speed);
	}
}

// Overloaded to add IGNORE_GLASS
void UTIL_TraceLine( const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, IGNORE_GLASS ignoreGlass, edict_t *pentIgnore, TraceResult *ptr )
{
   TRACE_LINE( vecStart, vecEnd, (igmon == ignore_monsters ? TRUE : FALSE) | (ignoreGlass?0x100:0), pentIgnore, ptr );

   if (DrawTraceLines)
	   UTIL_DrawBeam(listenserver_edict, vecStart, ptr->vecEndPos, 5, 0, 255, 0, 0, 250, 5, 5);

#ifdef _DEBUG
   TraceLineCalls++;
#endif
}

 
void UTIL_TraceLine( const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, edict_t *pentIgnore, TraceResult *ptr )
{
	TRACE_LINE( vecStart, vecEnd, (igmon == ignore_monsters ? TRUE : FALSE), pentIgnore, ptr );

    if (DrawTraceLines)
		UTIL_DrawBeam(listenserver_edict, vecStart, ptr->vecEndPos, 5, 0, 255, 0, 0, 250, 5, 5);

#ifdef _DEBUG
	TraceLineCalls++;
#endif
}

void UTIL_TraceHull( const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, int hullNumber, edict_t *pentIgnore, TraceResult *ptr )
{
	TRACE_HULL( vecStart, vecEnd, (igmon == ignore_monsters ? TRUE : FALSE), hullNumber, pentIgnore, ptr );

    if (DrawTraceLines)
		UTIL_DrawBeam(listenserver_edict, vecStart, ptr->vecEndPos, 5, 0, 0, 255, 0, 250, 5, 5);

#ifdef _DEBUG
	TraceLineCalls++;
#endif
}

void UTIL_MakeVectors( const Vector &vecAngles )
{
   MAKE_VECTORS( vecAngles );
}


edict_t *UTIL_FindEntityInSphere( edict_t *pentStart, const Vector &vecCenter, float flRadius )
{
   edict_t  *pentEntity;

   pentEntity = FIND_ENTITY_IN_SPHERE( pentStart, vecCenter, flRadius);

   if (!FNullEnt(pentEntity))
      return pentEntity;

   return NULL;
}

 
edict_t *UTIL_FindEntityByString( edict_t *pentStart, const char *szKeyword, const char *szValue )
{
   edict_t *pentEntity;

   pentEntity = FIND_ENTITY_BY_STRING( pentStart, szKeyword, szValue );

   if (!FNullEnt(pentEntity))
      return pentEntity;
   return NULL;
}

edict_t *UTIL_FindEntityByClassname( edict_t *pentStart, const char *szName )
{
   return UTIL_FindEntityByString( pentStart, "classname", szName );
}

edict_t *UTIL_FindEntityByTargetname( edict_t *pentStart, const char *szName )
{
   return UTIL_FindEntityByString( pentStart, "targetname", szName );
}


int UTIL_PointContents( const Vector &vec )
{
   return POINT_CONTENTS(vec);
}


void UTIL_SetSize( entvars_t *pev, const Vector &vecMin, const Vector &vecMax )
{
   SET_SIZE( ENT(pev), vecMin, vecMax );
}


void UTIL_SetOrigin( entvars_t *pev, const Vector &vecOrigin )
{
   SET_ORIGIN(ENT(pev), vecOrigin );
}


void ClientPrint( edict_t *pEntity, int msg_dest, const char *msg_name)
{
   if (IsNull(pEntity))
	   return;

   if (IsPlayingMessage)
   {
	   clog << "Tried to print out this message when there is already one active;\n" <<
			   msg_name << "\n";
	   return;
   }

   if (gmsgTextMsg == 0)
      gmsgTextMsg = REG_USER_MSG( "TextMsg", -1 );

   pfnMessageBegin( MSG_ONE, gmsgTextMsg, NULL, pEntity );

   pfnWriteByte( msg_dest );
   pfnWriteString( msg_name );
   pfnMessageEnd();
}

void UTIL_SayText( const char *pText, edict_t *pEdict )
{
   if (gmsgSayText == 0)
      gmsgSayText = REG_USER_MSG( "SayText", -1 );

   if (IsPlayingMessage)
   {
	   clog << "Tried to print out this message when there is already one active;\n" <<
			   pText << "\n";
	   return;
   }

   pfnMessageBegin( MSG_ONE, gmsgSayText, NULL, pEdict );
      pfnWriteByte( ENTINDEX(pEdict) );
      pfnWriteString( pText );
   pfnMessageEnd();
}


void UTIL_HostSay( edict_t *pEntity, int teamonly, char *message )
{
   int   j;
   char  text[128];
   char *pc;
   int   sender_team, player_team;
   edict_t *client;

   if ((!message) || (!*message))
	   return;

   // make sure the text has content
   for ( pc = message; pc != NULL && *pc != 0; pc++ )
   {
      if ( isprint( *pc ) && !isspace( *pc ) )
      {
         pc = NULL;   // we've found an alphanumeric character,  so text is valid
         break;
      }
   }

   if ( pc != NULL )
      return;  // no character found, so say nothing 

   // turn on color set 2  (color on,  no sound)
   if ( teamonly )
      sprintf( text, "%c(TEAM) %s: ", 2, STRING( pEntity->v.netname ) );
   else
      sprintf( text, "%c%s: ", 2, STRING( pEntity->v.netname ) );

   j = sizeof(text) - 2 - strlen(text);  // -2 for /n and null terminator
   if ( (int)strlen(message) > j )
      message[j] = 0;

   strcat( text, message );
   strcat( text, "\n" );

   // loop through all players
   // Start with the first player.
   // This may return the world in single player if the client types something between levels or during spawn
   // so check it, or it will infinite loop

   if (gmsgSayText == 0)
      gmsgSayText = REG_USER_MSG( "SayText", -1 );

   sender_team = UTIL_GetTeam(pEntity);

   if (IsPlayingMessage)
   {
	   clog << "Tried to print out this message when there is already one active;\n" <<
			   text << "\n";
	   return;
   }

   client = NULL;
   while ( ((client = UTIL_FindEntityByClassname( client, "player" )) != NULL) &&
           (!FNullEnt(client)) ) 
   {
      if ( client == pEntity )  // skip sender of message
         continue;

      player_team = UTIL_GetTeam(client);

      if ( teamonly && (sender_team != player_team) )
         continue;

      pfnMessageBegin( MSG_ONE, gmsgSayText, NULL, client );
         pfnWriteByte( ENTINDEX(pEntity) );
         pfnWriteString( text );
      pfnMessageEnd();
   }

   // print to the sending client
   pfnMessageBegin( MSG_ONE, gmsgSayText, NULL, pEntity );
      pfnWriteByte( ENTINDEX(pEntity) );
      pfnWriteString( text );
   pfnMessageEnd();
   
   // echo to server console
   g_engfuncs.pfnServerPrint( text );
}


#ifdef   DEBUG
edict_t *DBG_EntOfVars( const entvars_t *pev )
{
   if (pev->pContainingEntity != NULL)
      return pev->pContainingEntity;
   ALERT(at_console, "entvars_t pContainingEntity is NULL, calling into engine");
   edict_t* pent = (*g_engfuncs.pfnFindEntityByVars)((entvars_t*)pev);
   if (pent == NULL)
      ALERT(at_console, "DAMN!  Even the engine couldn't FindEntityByVars!");
   ((entvars_t *)pev)->pContainingEntity = pent;
   return pent;
}
#endif //DEBUG


// return team number 0 through 3 based what MOD uses for team numbers
int UTIL_GetTeam(edict_t *pEntity)
{
   return pEntity->v.team - 1;  // ESF teams are 1-2 based
}


// return class number 1 through N
int UTIL_GetClass(edict_t *pEntity)
{
   return pEntity->v.playerclass;
}


int UTIL_GetBotIndex(edict_t *pEdict)
{
   register byte index;

   for (index=0; index < gpGlobals->maxClients; index++)
   {
      if (bots[index].pEdict == pEdict)
      {
         return index;
      }
   }

   return -1;  // return -1 if edict is not a bot
}


bot_t *UTIL_GetBotPointer(edict_t *pEdict)
{
   register byte index;

   for (index=0; index < 32; index++)
   {
      if (bots[index].pEdict == pEdict)
      {
         break;
      }
   }

   if (index < 32)
      return (&bots[index]);

   return NULL;  // return NULL if edict is not a bot
}


bool IsAlive(edict_t *pEdict)
{
   return ((pEdict->v.deadflag == DEAD_NO) &&
           (pEdict->v.health > 0) && !(pEdict->v.flags & FL_NOTARGET) &&
		   (pEdict->v.takedamage != 0) && !(pEdict->v.flags & EF_NODRAW));
}


bool FInViewCone(Vector *pOrigin, edict_t *pEdict)
{
   Vector2D vec2LOS;
   float    flDot;

   UTIL_MakeVectors ( pEdict->v.angles );

   vec2LOS = ( *pOrigin - pEdict->v.origin ).Make2D();
   vec2LOS = vec2LOS.Normalize();

   flDot = DotProduct (vec2LOS , gpGlobals->v_forward.Make2D() );

   if ( flDot > 0.50 )  // 60 degree field of view 
   {
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}


bool FVisible( const Vector &vecOrigin, edict_t *pEdict )
{
   TraceResult tr;
   Vector      vecLookerOrigin;

#ifdef _DEBUG
   FVisibleCalls++;
#endif

   // look through caller's eyes
   vecLookerOrigin = pEdict->v.origin + pEdict->v.view_ofs;

   if (UTIL_VecIsInLava(vecOrigin))
	   return false;

//   int bInWater = (UTIL_PointContents (vecOrigin) == CONTENTS_WATER);
//   int bLookerInWater = (UTIL_PointContents (vecLookerOrigin) == CONTENTS_WATER);

   UTIL_TraceLine(vecLookerOrigin, vecOrigin, ignore_monsters, ignore_glass, pEdict, &tr);

   if (tr.flFraction != 1.0)
   {
      return false;  // Line of sight is not established
   }
   else
   {
      return true;  // line of sight is valid.
   }
}

bool UTIL_VecIsInLava(Vector vec)
{
	return (UTIL_PointContents(vec) == CONTENTS_LAVA);
}

Vector GetGunPosition(edict_t *pEdict)
{
   return (pEdict->v.origin + pEdict->v.view_ofs);
}


void UTIL_SelectItem(edict_t *pEdict, char *item_name)
{
   FakeClientCommand(pEdict, item_name, NULL, NULL);
}


Vector VecBModelOrigin(edict_t *pEdict)
{
   return pEdict->v.absmin + (pEdict->v.size * 0.5);
}

void UTIL_ShowMenu( edict_t *pEdict, int slots, int displaytime, bool needmore, char *pText )
{
   if (IS_DEDICATED_SERVER() == false)
   {
	   if (gmsgShowMenu == 0)
		   gmsgShowMenu = REG_USER_MSG( "ShowMenu", -1 );

	   pfnMessageBegin( MSG_ONE, gmsgShowMenu, NULL, pEdict );
		   pfnWriteShort( slots );
		   pfnWriteChar( displaytime );
	       pfnWriteByte( needmore );
	       pfnWriteString( pText );
	   pfnMessageEnd();
   }
   else
   {
	   if (strlen(pText) > 128)
	   {
		   char MenuText1[130] = "";
		   char MenuText2[130] = "";
		   
		   strncpy(MenuText1,pText,128);
		   strncpy(MenuText2,&pText[129],strlen(pText)-128);
		   UTIL_Printf(MenuText1);
		   UTIL_Printf(MenuText2);
	   }
	   else
		   UTIL_Printf(pText);

	   UTIL_Printf("\nTo make a choice type menuselect X where X stands for the choice number\n");
   }
}

void UTIL_BuildFileName(char *filename, char *arg1, char *arg2)
{
#ifndef __linux__

#ifdef _DEBUG
   strcpy(filename, "esf_inteam\\ESF_Bot\\");
#else
   strcpy(filename, "esf\\ESF_Bot\\");
#endif

#else
   strcpy(filename, "esf/");
#endif

   if ((arg1) && (*arg1) && (arg2) && (*arg2))
   {
      strcat(filename, arg1);

#ifndef __linux__
      strcat(filename, "\\");
#else
      strcat(filename, "/");
#endif

      strcat(filename, arg2);
   }
   else if ((arg1) && (*arg1))
   {
      strcat(filename, arg1);
   }
}

//=========================================================
// UTIL_LogPrintf - Prints a logged message to console.
// Preceded by LOG: ( timestamp ) < message >
//=========================================================
void UTIL_LogPrintf( char *fmt, ... )
{
   va_list        argptr;
   static char    string[1024];
   
   va_start ( argptr, fmt );
   vsprintf ( string, fmt, argptr );
   va_end   ( argptr );

   // Print to server console
   ALERT( at_logged, "%s", string );
}

bool UTIL_IsValidPlayer(edict_t *pEdict, bool CheckAlive, bool CheckObserver)
{
	if (IsNull(pEdict) == false)
	{
		if ((CheckObserver) && (b_observer_mode) && !(pEdict->v.flags & FL_FAKECLIENT))
			return false;

		if (FStrEq(STRING(pEdict->v.netname),""))
			return false;

        if ((CheckAlive) && (!IsAlive(pEdict)))
            return false;

		if ((pEdict->v.flags & FL_CLIENT) == 0)
			return false;

        if (pEdict->v.flags & FL_FAKECLIENT)
		{
			register short BotIndex = UTIL_GetBotIndex(pEdict);
            if ((BotIndex != -1) && (bots[BotIndex].is_used == false))
				return false;
		}

        return true;
	}
    return false;   
}

#ifdef classstuff
bool UTIL_IsValidPlayer(CClient *pClient, bool CheckAlive, bool CheckObserver)
{
	if (!pClient)
		return false;

	if (pClientManager->pClientList->IsInList(pClient) == false)
		return false; // if its not in the list its not valid

	if (IsNull(pClient->pEdict))
		return false;

	if ((pEdict->v.flags & FL_CLIENT) == 0)
		return false;

	if ((CheckObserver) && (b_observer_mode) && 
		!(pClient->pEdict->v.flags & FL_FAKECLIENT))
		return false;
		  
	if (FStrEq(STRING(pClient->pEdict->v.netname),""))
		return false;

	if ((CheckAlive) && (!pClient->IsAlive()))
		return false;

	return true;
}

#endif

bool IsNull (edict_t *pEdict)
{
    // thanx to PMB for this function
	try
	{
		if (pEdict == NULL)
			return true;

		if (pfnEntOffsetOfPEntity (pEdict) == 0)
			return true;
   
		if (pEdict->free)
		    return true;

		if (pEdict->v.flags & FL_KILLME)
			return true;
   }
   catch (...)
   {
#ifdef _DEBUG
		FILE *fp = fopen("bot.txt", "a");
		fprintf(fp, "IsNull() Error\n");
		fclose(fp);
#endif
	   return true; // failed so assuming entity isn't valid...
   }

   return false;
}

float UTIL_Distance(Vector vec1, Vector vec2)
{
    Vector v_dist = (vec2 - vec1);
    return v_dist.Length();
}

referential_t ReferentialOfAngles (Vector &v_angles)
{
   // this function considers a system of angles (most common example, a view angle), and sets
   // a vector referential upon it, in order to have the relative direction of "up", "right" and
   // "forward".

   referential_t referential;

   // the engine has a handy function for this too
   pfnAngleVectors (WrapAngles (v_angles), (float *) &referential.v_forward, (float *) &referential.v_right, (float *) &referential.v_up);

   return (referential); // return the new referential
}

void UTIL_GetModel(edict_t *pEdict, char *model_name)
{
   char *infobuffer;
   char szTemp[256] = "";

   infobuffer = (*g_engfuncs.pfnGetInfoKeyBuffer)( pEdict );
   strcpy(szTemp, (g_engfuncs.pfnInfoKeyValue(infobuffer, "model")));

   if (strlen(szTemp) > 32)
   {
	   strncpy(model_name, szTemp, 31);
	   model_name[31] = 0;
   }
   else
	   strcpy(model_name, szTemp);
}

int UTIL_GetClientIndex(edict_t *pEdict)
{
	if (!pEdict)
		return NULL;

	register byte i;

	for (i=0;i<gpGlobals->maxClients;i++)
	{
		if (ClientInfo[i].pEdict == pEdict)
			return i;
	}
	return -1;
}

ClientInfo_t *UTIL_GetClientPointer(edict_t *pEdict)
{
	if (!pEdict)
		return NULL;

	register byte i;

	for (i=0;i<gpGlobals->maxClients;i++)
	{
		if (ClientInfo[i].pEdict == pEdict)
			return &ClientInfo[i];
	}
	return NULL;
}

bool UTIL_IsValidEnemy(bot_t *pBot, edict_t *pEnemy)
{
	if (UTIL_IsValidPlayer(pEnemy) == false)
		return false;

   	// is team play enabled?
	if (IsTeamPlay())
	{
		int player_team = UTIL_GetTeam(pEnemy);
		int bot_team = UTIL_GetTeam(pBot->pEdict);

		// don't target your teammates...
		if (bot_team == player_team)
		{
			return false;
		}	
	}

	Vector vecEnemy = pEnemy->v.origin + pEnemy->v.view_ofs;

	Vector v_dist = vecEnemy - GetGunPosition(pBot->pEdict);
	v_dist.z = 0;
	float distance = v_dist.Length();

	bool OnScouter = ((distance <= 2300.0) || (pBot->SBotType > 0));
	bool InViewCone = (FInViewCone(&vecEnemy, pBot->pEdict));
	bool JustMadeASound = false;
	int ClInd = UTIL_GetClientIndex(pEnemy);

	distance = UTIL_Distance(pBot->pEdict->v.origin, pEnemy->v.origin);

	if (ClInd > -1)
		JustMadeASound = ((ClientInfo[ClInd].LastSoundTime > (gpGlobals->time - 1.0)) &&
						  (ClientInfo[ClInd].SoundName[0] != 0) && (distance < 3000.0));

	if ((!InViewCone) && (!OnScouter) && (!JustMadeASound))
		return false;

	// Check if there is an player list, saves us a traceline :)
	if (pBot->MadePlayerList)
	{
		return ((pBot->PlayersChecked & (1<<(ENTINDEX(pEnemy) - 1))) != 0);
	}
	else
	{
		// We have to use a traceline :(
		return (FVisible(vecEnemy,pBot->pEdict));
	}

	return false;
}

// yay yet another pmb function
void UTIL_PrintEntity (edict_t *pEntity, const char *szFileName)
{
   int index;

   char temp[128];
   
   if (!szFileName)
	   sprintf(temp,"bot%i.txt",ENTINDEX(pEntity));
   else
	   strcpy(temp, szFileName);

   FILE *fp = fopen(temp,"a");
   fprintf(fp,"\n");

   // now just dump the entity variables
   fprintf(fp, "pEntity->v.classname = %d (\"%s\")\n", pEntity->v.classname, STRING (pEntity->v.classname));
   fprintf(fp, "pEntity->v.globalname = %d (\"%s\")\n", pEntity->v.globalname, STRING (pEntity->v.globalname));
   fprintf(fp, "pEntity->v.origin = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.origin.x, pEntity->v.origin.y, pEntity->v.origin.z);
   fprintf(fp, "pEntity->v.oldorigin = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.oldorigin.x, pEntity->v.oldorigin.y, pEntity->v.oldorigin.z);
   fprintf(fp, "pEntity->v.velocity = Vector (%.1f, %.1f, %.1f) - Length %.3f - Length2D %.3f\n", pEntity->v.velocity.x, pEntity->v.velocity.y, pEntity->v.velocity.z, pEntity->v.velocity.Length (), pEntity->v.velocity.Length2D ());
   fprintf(fp, "pEntity->v.basevelocity = Vector (%.1f, %.1f, %.1f) - Length %.3f - Length2D %.3f\n", pEntity->v.basevelocity.x, pEntity->v.basevelocity.y, pEntity->v.basevelocity.z, pEntity->v.velocity.Length (), pEntity->v.velocity.Length2D ());
   fprintf(fp, "pEntity->v.clbasevelocity = Vector (%.1f, %.1f, %.1f) - Length %.3f - Length2D %.3f\n", pEntity->v.clbasevelocity.x, pEntity->v.clbasevelocity.y, pEntity->v.clbasevelocity.z, pEntity->v.velocity.Length (), pEntity->v.velocity.Length2D ());
   fprintf(fp, "pEntity->v.movedir = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.movedir.x, pEntity->v.movedir.y, pEntity->v.movedir.z);
   fprintf(fp, "pEntity->v.angles = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.angles.x, pEntity->v.angles.y, pEntity->v.angles.z);
   fprintf(fp, "pEntity->v.avelocity = Vector (%.1f, %.1f, %.1f) - Length %.3f - Length2D %.3f\n", pEntity->v.avelocity.x, pEntity->v.avelocity.y, pEntity->v.avelocity.z, pEntity->v.velocity.Length (), pEntity->v.velocity.Length2D ());
   fprintf(fp, "pEntity->v.punchangle = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.punchangle.x, pEntity->v.punchangle.y, pEntity->v.punchangle.z);
   fprintf(fp, "pEntity->v.v_angle = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.v_angle.x, pEntity->v.v_angle.y, pEntity->v.v_angle.z);
   fprintf(fp, "pEntity->v.endpos = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.endpos.x, pEntity->v.endpos.y, pEntity->v.endpos.z);
   fprintf(fp, "pEntity->v.startpos = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.startpos.x, pEntity->v.startpos.y, pEntity->v.startpos.z);
   fprintf(fp, "pEntity->v.impacttime = %.3f\n", pEntity->v.impacttime);
   fprintf(fp, "pEntity->v.starttime = %.3f\n", pEntity->v.starttime);
   fprintf(fp, "pEntity->v.fixangle = %d (%s)\n", pEntity->v.fixangle, (pEntity->v.fixangle == 0 ? "nothing" : (pEntity->v.fixangle == 1 ? "force view angles" : (pEntity->v.fixangle == 2 ? "add velocity" : "???"))));
   fprintf(fp, "pEntity->v.idealpitch = %.3f\n", pEntity->v.idealpitch);
   fprintf(fp, "pEntity->v.pitch_speed = %.3f\n", pEntity->v.pitch_speed);
   fprintf(fp, "pEntity->v.ideal_yaw = %.3f\n", pEntity->v.ideal_yaw);
   fprintf(fp, "pEntity->v.yaw_speed = %.3f\n", pEntity->v.yaw_speed);
   fprintf(fp, "pEntity->v.modelindex = %d\n", pEntity->v.modelindex);
   fprintf(fp, "pEntity->v.model = %d (\"%s\")\n", pEntity->v.model, STRING (pEntity->v.model));
   fprintf(fp, "pEntity->v.viewmodel = %d (\"%s\")\n", pEntity->v.viewmodel, STRING (pEntity->v.viewmodel));
   fprintf(fp, "pEntity->v.weaponmodel = %d (\"%s\")\n", pEntity->v.weaponmodel, STRING (pEntity->v.weaponmodel));
   fprintf(fp, "pEntity->v.absmin = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.absmin.x, pEntity->v.absmin.y, pEntity->v.absmin.z);
   fprintf(fp, "pEntity->v.absmax = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.absmax.x, pEntity->v.absmax.y, pEntity->v.absmax.z);
   fprintf(fp, "pEntity->v.mins = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.mins.x, pEntity->v.mins.y, pEntity->v.mins.z);
   fprintf(fp, "pEntity->v.maxs = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.maxs.x, pEntity->v.maxs.y, pEntity->v.maxs.z);
   fprintf(fp, "pEntity->v.size = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.size.x, pEntity->v.size.y, pEntity->v.size.z);
   fprintf(fp, "pEntity->v.ltime = %.3f (current time is %.3f)\n", pEntity->v.ltime, gpGlobals->time);
   fprintf(fp, "pEntity->v.nextthink = %.3f (current time is %.3f)\n", pEntity->v.nextthink, gpGlobals->time);
   fprintf(fp, "pEntity->v.movetype = %d (%s)\n", pEntity->v.movetype, (pEntity->v.movetype == 0 ? "MOVETYPE_NONE" : (pEntity->v.movetype == MOVETYPE_WALK ? "MOVETYPE_WALK" : (pEntity->v.movetype == MOVETYPE_STEP ? "MOVETYPE_STEP" : (pEntity->v.movetype == MOVETYPE_FLY ? "MOVETYPE_FLY" : (pEntity->v.movetype == MOVETYPE_TOSS ? "MOVETYPE_TOSS" : (pEntity->v.movetype == MOVETYPE_PUSH ? "MOVETYPE_PUSH" : (pEntity->v.movetype == MOVETYPE_NOCLIP ? "MOVETYPE_NOCLIP" : (pEntity->v.movetype == MOVETYPE_FLYMISSILE ? "MOVETYPE_FLYMISSILE" : (pEntity->v.movetype == MOVETYPE_BOUNCE ? "MOVETYPE_BOUNCE" : (pEntity->v.movetype == MOVETYPE_BOUNCEMISSILE ? "MOVETYPE_BOUNCEMISSILE" : (pEntity->v.movetype == MOVETYPE_FOLLOW ? "MOVETYPE_FOLLOW" : (pEntity->v.movetype == MOVETYPE_PUSHSTEP ? "MOVETYPE_PUSHSTEP" : "???")))))))))))));
   fprintf(fp, "pEntity->v.solid = %d (%s)\n", pEntity->v.solid, (pEntity->v.solid == SOLID_NOT ? "SOLID_NOT" : (pEntity->v.solid == SOLID_TRIGGER ? "SOLID_TRIGGER" : (pEntity->v.solid == SOLID_BBOX ? "SOLID_BBOX" : (pEntity->v.solid == SOLID_SLIDEBOX ? "SOLID_SLIDEBOX" : (pEntity->v.solid == SOLID_BSP ? "SOLID_BSP" : "???"))))));
   fprintf(fp, "pEntity->v.skin = %d (\"%s\")\n", pEntity->v.skin, STRING (pEntity->v.skin));
   fprintf(fp, "pEntity->v.body = %d (\"%s\")\n", pEntity->v.body, STRING (pEntity->v.body));
   sprintf (temp, ""); // prepare pattern string
   if (pEntity->v.effects & EF_BRIGHTFIELD) strcat (temp, "EF_BRIGHTFIELD | ");
   if (pEntity->v.effects & EF_MUZZLEFLASH) strcat (temp, "EF_MUZZLEFLASH | ");
   if (pEntity->v.effects & EF_BRIGHTLIGHT) strcat (temp, "EF_BRIGHTLIGHT | ");
   if (pEntity->v.effects & EF_DIMLIGHT) strcat (temp, "EF_DIMLIGHT | ");
   if (pEntity->v.effects & EF_INVLIGHT) strcat (temp, "EF_INVLIGHT | ");
   if (pEntity->v.effects & EF_NOINTERP) strcat (temp, "EF_NOINTERP | ");
   if (pEntity->v.effects & EF_LIGHT) strcat (temp, "EF_LIGHT | ");
   if (pEntity->v.effects & EF_NODRAW) strcat (temp, "EF_NODRAW | ");
   if (strlen (temp) > 2) temp[strlen (temp) - 3] = 0; // remove trailing " | "
   fprintf(fp, "pEntity->v.effects = %d (%s)\n", pEntity->v.effects, temp);
   fprintf(fp, "pEntity->v.gravity = %.3f (fraction /1 of normal)\n", pEntity->v.gravity);
   fprintf(fp, "pEntity->v.friction = %.3f\n", pEntity->v.friction);
   fprintf(fp, "pEntity->v.light_level = %d\n", pEntity->v.light_level);
   fprintf(fp, "pEntity->v.sequence = %d (\"%s\")\n", pEntity->v.sequence, STRING (pEntity->v.sequence));
   fprintf(fp, "pEntity->v.gaitsequence = %d (\"%s\")\n", pEntity->v.gaitsequence, STRING (pEntity->v.gaitsequence));
   fprintf(fp, "pEntity->v.frame = %.3f (position /255 of total in sequence)\n", pEntity->v.frame);
   fprintf(fp, "pEntity->v.animtime = %.3f (current time is %.3f)\n", pEntity->v.animtime, gpGlobals->time);
   fprintf(fp, "pEntity->v.framerate = %.3f (times normal speed)\n", pEntity->v.framerate);
   fprintf(fp, "pEntity->v.controller = {%d, %d, %d, %d}\n", pEntity->v.controller[0], pEntity->v.controller[1], pEntity->v.controller[2], pEntity->v.controller[3]);
   fprintf(fp, "pEntity->v.blending = {%d, %d}\n", pEntity->v.blending[0], pEntity->v.blending[1]);
   fprintf(fp, "pEntity->v.scale = %.3f\n", pEntity->v.scale);
   fprintf(fp, "pEntity->v.rendermode = %d (%s)\n", pEntity->v.rendermode, (pEntity->v.rendermode == kRenderNormal ? "kRenderNormal" : (pEntity->v.rendermode == kRenderTransColor ? "kRenderTransColor" : (pEntity->v.rendermode == kRenderTransTexture ? "kRenderTransTexture" : (pEntity->v.rendermode == kRenderGlow ? "kRenderGlow" : (pEntity->v.rendermode == kRenderTransAlpha ? "kRenderTransAlpha" : (pEntity->v.rendermode == kRenderTransAdd ? "kRenderTransAdd" : "???")))))));
   fprintf(fp, "pEntity->v.renderamt = %.3f\n", pEntity->v.renderamt);
   fprintf(fp, "pEntity->v.rendercolor = Vector (%.1f, %.1f, %.1f) (RGB)\n", pEntity->v.rendercolor.x, pEntity->v.rendercolor.y, pEntity->v.rendercolor.z);
   fprintf(fp, "pEntity->v.renderfx = %d (%s)\n", pEntity->v.renderfx, (pEntity->v.renderfx == kRenderFxNone ? "kRenderFxNone" : (pEntity->v.renderfx == kRenderFxPulseSlow ? "kRenderFxPulseSlow" : (pEntity->v.renderfx == kRenderFxPulseFast ? "kRenderFxPulseFast" : (pEntity->v.renderfx == kRenderFxPulseSlowWide ? "kRenderFxPulseSlowWide" : (pEntity->v.renderfx == kRenderFxPulseFastWide ? "kRenderFxPulseFastWide" : (pEntity->v.renderfx == kRenderFxFadeSlow ? "kRenderFxFadeSlow" : (pEntity->v.renderfx == kRenderFxFadeFast ? "kRenderFxFadeFast" : (pEntity->v.renderfx == kRenderFxSolidSlow ? "kRenderFxSolidSlow" : (pEntity->v.renderfx == kRenderFxSolidFast ? "kRenderFxSolidFast" : (pEntity->v.renderfx == kRenderFxStrobeSlow ? "kRenderFxStrobeSlow" : (pEntity->v.renderfx == kRenderFxStrobeFast ? "kRenderFxStrobeFast" : (pEntity->v.renderfx == kRenderFxStrobeFaster ? "kRenderFxStrobeFaster" : (pEntity->v.renderfx == kRenderFxFlickerSlow ? "kRenderFxFlickerSlow" : (pEntity->v.renderfx == kRenderFxFlickerFast ? "kRenderFxFlickerFast" : (pEntity->v.renderfx == kRenderFxNoDissipation ? "kRenderFxNoDissipation" : (pEntity->v.renderfx == kRenderFxDistort ? "kRenderFxDistort" : (pEntity->v.renderfx == kRenderFxHologram ? "kRenderFxHologram" : (pEntity->v.renderfx == kRenderFxDeadPlayer ? "kRenderFxDeadPlayer" : (pEntity->v.renderfx == kRenderFxExplode ? "kRenderFxExplode" : (pEntity->v.renderfx == kRenderFxGlowShell ? "kRenderFxGlowShell" : (pEntity->v.renderfx == kRenderFxClampMinScale ? "kRenderFxClampMinScale" : "???"))))))))))))))))))))));
   fprintf(fp, "pEntity->v.health = %.3f\n", pEntity->v.health);
   fprintf(fp, "pEntity->v.frags = %.3f\n", pEntity->v.frags);
   sprintf (temp, "00000000000000000000000000000000"); // prepare pattern string
   for (index = 0; index < 32; index++) if (pEntity->v.weapons & (1 << index)) temp[index] = '1';
   fprintf(fp, "pEntity->v.weapons = %d (%s)\n", pEntity->v.weapons, temp);
   fprintf(fp, "pEntity->v.takedamage = %.3f (%s)\n", pEntity->v.takedamage, (pEntity->v.takedamage == DAMAGE_NO ? "DAMAGE_NO" : (pEntity->v.takedamage == DAMAGE_YES ? "DAMAGE_YES" : (pEntity->v.takedamage == DAMAGE_AIM ? "DAMAGE_AIM" : "???"))));
   fprintf(fp, "pEntity->v.deadflag = %d (%s)\n", pEntity->v.deadflag, (pEntity->v.deadflag == DEAD_NO ? "DEAD_NO" : (pEntity->v.deadflag == DEAD_DYING ? "DEAD_DYING" : (pEntity->v.deadflag == DEAD_DEAD ? "DEAD_DEAD" : (pEntity->v.deadflag == DEAD_RESPAWNABLE ? "DEAD_RESPAWNABLE" : (pEntity->v.deadflag == DEAD_DISCARDBODY ? "DEAD_DISCARDBODY" : "???"))))));
   fprintf(fp, "pEntity->v.view_ofs = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.view_ofs.x, pEntity->v.view_ofs.y, pEntity->v.view_ofs.z);
   sprintf (temp, ""); // prepare pattern string
   if (pEntity->v.button & IN_ATTACK) strcat (temp, "IN_ATTACK | ");
   if (pEntity->v.button & IN_JUMP) strcat (temp, "IN_JUMP | ");
   if (pEntity->v.button & IN_DUCK) strcat (temp, "IN_DUCK | ");
   if (pEntity->v.button & IN_FORWARD) strcat (temp, "IN_FORWARD | ");
   if (pEntity->v.button & IN_BACK) strcat (temp, "IN_BACK | ");
   if (pEntity->v.button & IN_USE) strcat (temp, "IN_USE | ");
   if (pEntity->v.button & IN_CANCEL) strcat (temp, "IN_CANCEL | ");
   if (pEntity->v.button & IN_LEFT) strcat (temp, "IN_LEFT | ");
   if (pEntity->v.button & IN_RIGHT) strcat (temp, "IN_RIGHT | ");
   if (pEntity->v.button & IN_MOVELEFT) strcat (temp, "IN_MOVELEFT | ");
   if (pEntity->v.button & IN_MOVERIGHT) strcat (temp, "IN_MOVERIGHT | ");
   if (pEntity->v.button & IN_ATTACK2) strcat (temp, "IN_ATTACK2 | ");
   if (pEntity->v.button & IN_RUN) strcat (temp, "IN_RUN | ");
   if (pEntity->v.button & IN_RELOAD) strcat (temp, "IN_RELOAD | ");
   if (pEntity->v.button & IN_ALT1) strcat (temp, "IN_ALT1 | ");
   if (pEntity->v.button & IN_SCORE) strcat (temp, "IN_SCORE | ");
   if (strlen (temp) > 2) temp[strlen (temp) - 3] = 0; // remove trailing " | "
   fprintf(fp, "pEntity->v.button = %d (%s)\n", pEntity->v.button, temp);
   fprintf(fp, "pEntity->v.impulse = %d\n", pEntity->v.impulse);
   if (pEntity->v.chain != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.chain), STRING (pEntity->v.chain->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.chain = %s\n", temp);
   if (pEntity->v.dmg_inflictor != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.dmg_inflictor), STRING (pEntity->v.dmg_inflictor->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.dmg_inflictor = %s\n", temp);
   if (pEntity->v.enemy != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.enemy), STRING (pEntity->v.enemy->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.enemy = %s\n", temp);
   if (pEntity->v.aiment != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.aiment), STRING (pEntity->v.aiment->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.aiment = %s\n", temp);
   if (pEntity->v.owner != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.owner), STRING (pEntity->v.owner->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.owner = %s\n", temp);
   if (pEntity->v.groundentity != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.groundentity), STRING (pEntity->v.groundentity->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.groundentity = %s\n", temp);
   fprintf(fp, "pEntity->v.spawnflags = %d\n", pEntity->v.spawnflags);
   sprintf (temp, ""); // prepare pattern string
   if (pEntity->v.flags & FL_FLY) strcat (temp, "FL_FLY | ");
   if (pEntity->v.flags & FL_SWIM) strcat (temp, "FL_SWIM | ");
   if (pEntity->v.flags & FL_CONVEYOR) strcat (temp, "FL_CONVEYOR | ");
   if (pEntity->v.flags & FL_CLIENT) strcat (temp, "FL_CLIENT | ");
   if (pEntity->v.flags & FL_INWATER) strcat (temp, "FL_INWATER | ");
   if (pEntity->v.flags & FL_MONSTER) strcat (temp, "FL_MONSTER | ");
   if (pEntity->v.flags & FL_GODMODE) strcat (temp, "FL_GODMODE | ");
   if (pEntity->v.flags & FL_NOTARGET) strcat (temp, "FL_NOTARGET | ");
   if (pEntity->v.flags & FL_SKIPLOCALHOST) strcat (temp, "FL_SKIPLOCALHOST | ");
   if (pEntity->v.flags & FL_ONGROUND) strcat (temp, "FL_ONGROUND | ");
   if (pEntity->v.flags & FL_PARTIALGROUND) strcat (temp, "FL_PARTIALGROUND | ");
   if (pEntity->v.flags & FL_WATERJUMP) strcat (temp, "FL_WATERJUMP | ");
   if (pEntity->v.flags & FL_FROZEN) strcat (temp, "FL_FROZEN | ");
   if (pEntity->v.flags & FL_FAKECLIENT) strcat (temp, "FL_FAKECLIENT | ");
   if (pEntity->v.flags & FL_DUCKING) strcat (temp, "FL_DUCKING | ");
   if (pEntity->v.flags & FL_FLOAT) strcat (temp, "FL_FLOAT | ");
   if (pEntity->v.flags & FL_GRAPHED) strcat (temp, "FL_GRAPHED | ");
   if (pEntity->v.flags & FL_IMMUNE_WATER) strcat (temp, "FL_IMMUNE_WATER | ");
   if (pEntity->v.flags & FL_IMMUNE_SLIME) strcat (temp, "FL_IMMUNE_SLIME | ");
   if (pEntity->v.flags & FL_IMMUNE_LAVA) strcat (temp, "FL_IMMUNE_LAVA | ");
   if (pEntity->v.flags & FL_PROXY) strcat (temp, "FL_PROXY | ");
   if (pEntity->v.flags & FL_ALWAYSTHINK) strcat (temp, "FL_ALWAYSTHINK | ");
   if (pEntity->v.flags & FL_BASEVELOCITY) strcat (temp, "FL_BASEVELOCITY | ");
   if (pEntity->v.flags & FL_MONSTERCLIP) strcat (temp, "FL_MONSTERCLIP | ");
   if (pEntity->v.flags & FL_ONTRAIN) strcat (temp, "FL_ONTRAIN | ");
   if (pEntity->v.flags & FL_WORLDBRUSH) strcat (temp, "FL_WORLDBRUSH | ");
   if (pEntity->v.flags & FL_SPECTATOR) strcat (temp, "FL_SPECTATOR | ");
   if (pEntity->v.flags & FL_CUSTOMENTITY) strcat (temp, "FL_CUSTOMENTITY | ");
   if (pEntity->v.flags & FL_KILLME) strcat (temp, "FL_KILLME | ");
   if (pEntity->v.flags & FL_DORMANT) strcat (temp, "FL_DORMANT | ");
   if (strlen (temp) > 2) temp[strlen (temp) - 3] = 0; // remove trailing " | "
   fprintf(fp, "pEntity->v.flags = %d (%s)\n", pEntity->v.flags, temp);
   fprintf(fp, "pEntity->v.colormap = %d (0x%X)\n", pEntity->v.colormap, pEntity->v.colormap);
   fprintf(fp, "pEntity->v.team = %d\n", pEntity->v.team);
   fprintf(fp, "pEntity->v.max_health = %.3f\n", pEntity->v.max_health);
   fprintf(fp, "pEntity->v.teleport_time = %.3f\n", pEntity->v.teleport_time);
   fprintf(fp, "pEntity->v.armortype = %.3f\n", pEntity->v.armortype);
   fprintf(fp, "pEntity->v.armorvalue = %.3f\n", pEntity->v.armorvalue);
   fprintf(fp, "pEntity->v.waterlevel = %d (%s)\n", pEntity->v.waterlevel, (pEntity->v.waterlevel == 0 ? "not in water" : (pEntity->v.waterlevel == 2 ? "walking in water" : (pEntity->v.waterlevel == 3 ? "swimming in water" : "???"))));
   fprintf(fp, "pEntity->v.watertype = %d\n", pEntity->v.watertype);
   fprintf(fp, "pEntity->v.target = %d (\"%s\")\n", pEntity->v.target, STRING (pEntity->v.target));
   fprintf(fp, "pEntity->v.targetname = %d (\"%s\")\n", pEntity->v.targetname, STRING (pEntity->v.targetname));
   fprintf(fp, "pEntity->v.netname = %d (\"%s\")\n", pEntity->v.netname, STRING (pEntity->v.netname));
   fprintf(fp, "pEntity->v.message = %d (\"%s\")\n", pEntity->v.message, STRING (pEntity->v.message));
   fprintf(fp, "pEntity->v.dmg_take = %.3f\n", pEntity->v.dmg_take);
   fprintf(fp, "pEntity->v.dmg_save = %.3f\n", pEntity->v.dmg_save);
   fprintf(fp, "pEntity->v.dmg = %.3f\n", pEntity->v.dmg);
   fprintf(fp, "pEntity->v.dmgtime = %.3f (current time is %.3f)\n", pEntity->v.dmgtime, gpGlobals->time);
   fprintf(fp, "pEntity->v.noise = %d (\"%s\")\n", pEntity->v.noise, STRING (pEntity->v.noise));
   fprintf(fp, "pEntity->v.noise1 = %d (\"%s\")\n", pEntity->v.noise1, STRING (pEntity->v.noise1));
   fprintf(fp, "pEntity->v.noise2 = %d (\"%s\")\n", pEntity->v.noise2, STRING (pEntity->v.noise2));
   fprintf(fp, "pEntity->v.noise3 = %d (\"%s\")\n", pEntity->v.noise3, STRING (pEntity->v.noise3));
   fprintf(fp, "pEntity->v.speed = %.3f\n", pEntity->v.speed);
   fprintf(fp, "pEntity->v.air_finished = %.3f\n", pEntity->v.air_finished);
   fprintf(fp, "pEntity->v.pain_finished = %.3f\n", pEntity->v.pain_finished);
   fprintf(fp, "pEntity->v.radsuit_finished = %.3f\n", pEntity->v.radsuit_finished);
   if (pEntity->v.pContainingEntity != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.pContainingEntity), STRING (pEntity->v.pContainingEntity->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.pContainingEntity = %s\n", temp);
   fprintf(fp, "pEntity->v.playerclass = %d\n", pEntity->v.playerclass);
   fprintf(fp, "pEntity->v.maxspeed = %.3f\n", pEntity->v.maxspeed);
   fprintf(fp, "pEntity->v.fov = %.3f\n", pEntity->v.fov);
   fprintf(fp, "pEntity->v.weaponanim = %d\n", pEntity->v.weaponanim);
   fprintf(fp, "pEntity->v.pushmsec = %d\n", pEntity->v.pushmsec);
   fprintf(fp, "pEntity->v.bInDuck = %d (%s)\n", pEntity->v.bInDuck, (pEntity->v.bInDuck > 0 ? "TRUE" : "FALSE"));
   fprintf(fp, "pEntity->v.flTimeStepSound = %d (current time is %.3f)\n", pEntity->v.flTimeStepSound, gpGlobals->time);
   fprintf(fp, "pEntity->v.flSwimTime = %d (current time is %.3f)\n", pEntity->v.flSwimTime, gpGlobals->time);
   fprintf(fp, "pEntity->v.flDuckTime = %d (current time is %.3f)\n", pEntity->v.flDuckTime, gpGlobals->time);
   fprintf(fp, "pEntity->v.iStepLeft = %d\n", pEntity->v.iStepLeft);
   fprintf(fp, "pEntity->v.flFallVelocity = %.3f\n", pEntity->v.flFallVelocity);
   fprintf(fp, "pEntity->v.gamestate = %d\n", pEntity->v.gamestate);
   fprintf(fp, "pEntity->v.oldbuttons = %d\n", pEntity->v.oldbuttons);
   fprintf(fp, "pEntity->v.groupinfo = %d\n", pEntity->v.groupinfo);
   fprintf(fp, "pEntity->v.iuser1 = %d\n", pEntity->v.iuser1);
   fprintf(fp, "pEntity->v.iuser2 = %d\n", pEntity->v.iuser2);
   fprintf(fp, "pEntity->v.iuser3 = %d\n", pEntity->v.iuser3);
   fprintf(fp, "pEntity->v.iuser4 = %d\n", pEntity->v.iuser4);
   fprintf(fp, "pEntity->v.fuser1 = %.3f\n", pEntity->v.fuser1);
   fprintf(fp, "pEntity->v.fuser2 = %.3f\n", pEntity->v.fuser2);
   fprintf(fp, "pEntity->v.fuser3 = %.3f\n", pEntity->v.fuser3);
   fprintf(fp, "pEntity->v.fuser4 = %.3f\n", pEntity->v.fuser4);
   fprintf(fp, "pEntity->v.vuser1 = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.vuser1.x, pEntity->v.vuser1.y, pEntity->v.vuser1.z);
   fprintf(fp, "pEntity->v.vuser2 = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.vuser2.x, pEntity->v.vuser2.y, pEntity->v.vuser2.z);
   fprintf(fp, "pEntity->v.vuser3 = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.vuser3.x, pEntity->v.vuser3.y, pEntity->v.vuser3.z);
   fprintf(fp, "pEntity->v.vuser4 = Vector (%.1f, %.1f, %.1f)\n", pEntity->v.vuser4.x, pEntity->v.vuser4.y, pEntity->v.vuser4.z);
   if (pEntity->v.euser1 != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.euser1), STRING (pEntity->v.euser1->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.euser1 = %s\n", temp);
   if (pEntity->v.euser2 != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.euser2), STRING (pEntity->v.euser2->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.euser2 = %s\n", temp);
   if (pEntity->v.euser3 != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.euser3), STRING (pEntity->v.euser3->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.euser3 = %s\n", temp);
   if (pEntity->v.euser4 != NULL) sprintf (temp, "Entity #%d (\"%s\")", ENTINDEX (pEntity->v.euser4), STRING (pEntity->v.euser4->v.classname)); else sprintf (temp, "NULL");
   fprintf(fp, "pEntity->v.euser4 = %s\n", temp);

   fprintf(fp,"\n");
   fclose(fp);
   return;
}

void UTIL_ChangePl(edict_t *pEdict, long NewPl)
{
	*((int *)pEdict->pvPrivateData + PL_OFFSET) = NewPl;
	return;
}

void UTIL_ChangeRealPl(edict_t *pEdict, long NewPl)
{
	*((int *)pEdict->pvPrivateData + ACTUAL_PL_OFFSET) = NewPl;
	return;
}

void UTIL_ChangeKi(edict_t *pEdict, int NewKi)
{
	pEdict->v.fuser4 = float(NewKi);
	return;
}

short UTIL_EntDir(edict_t *pPlayer, edict_t *pEntity)
{
	if ((!pPlayer) || (!pEntity))
		return 0;

	Vector v_src = pPlayer->v.origin;
	Vector v_predict = PredictEntityPos(pEntity, 1.0);

	Vector v_end;

	float Dist1 = 0.0;
	float Dist2 = 0.0;
	short Dir = 0;

	UTIL_MakeVectors(pPlayer->v.angles);

	// Is the entity heading left or right ?
	v_end = v_src - gpGlobals->v_right * 100.0;
	Dist1 = UTIL_Distance(v_predict, v_end);

	v_end = v_src + gpGlobals->v_right * 100.0;
	Dist2 = UTIL_Distance(v_predict, v_end);

	if (Dist1 < Dist2)
		Dir |= dir_left;
	else if (Dist1 > Dist2)
		Dir |= dir_right;

	// Is the entity heading upward or downward ?
	v_end = v_src - gpGlobals->v_up * 100.0;
	Dist1 = UTIL_Distance(v_predict, v_end);
	
	v_end = v_src + gpGlobals->v_up * 100.0;
	Dist2 = UTIL_Distance(v_predict, v_end);

	if (Dist1 < Dist2)
		Dir |= dir_down;
	else if (Dist1 > Dist2)
		Dir |= dir_up;

	// Is the entity heading backward or forward ?
	v_end = v_src - gpGlobals->v_forward * 100.0;
	Dist1 = UTIL_Distance(v_predict, v_end);
	
	v_end = v_src + gpGlobals->v_forward * 100.0;
	Dist2 = UTIL_Distance(v_predict, v_end);

	if (Dist1 < Dist2)
		Dir |= dir_back;
	else if (Dist1 > Dist2)
		Dir |= dir_forward;

	return Dir;
}

// Determine at which side an vector is
int UTIL_VecDir(edict_t *pEdict, Vector v_vec)
{
	Vector v_src = pEdict->v.origin;
	int vecDir = 0;

	Vector v_end1 = g_vecZero;
	Vector v_end2 = g_vecZero;
	float dist1 = 0.0;
	float dist2 = 0.0;

	UTIL_MakeVectors(pEdict->v.angles);

	// Is the vector at the left side or at the right side ?
	v_end1 = v_src - gpGlobals->v_right * 100.0;
	v_end2 = v_src + gpGlobals->v_right * 100.0;

	dist1 = UTIL_Distance(v_vec, v_end1);
	dist2 = UTIL_Distance(v_vec, v_end2);

	if (dist1 > dist2)
		vecDir |= dir_right; // vector is at the right side
	else if (dist2 > dist1)
		vecDir |= dir_left; // vector is at the left side

	// Is the vector lower or higher ?
	if (v_src.z > v_vec.z)
		vecDir |= dir_down; // vector is lower
	else if (v_src.z < v_vec.z)
		vecDir |= dir_up; // vector is higher

	// Is the vector behind or infront ?
	v_end1 = v_src - gpGlobals->v_forward * 100.0;
	v_end2 = v_src + gpGlobals->v_forward * 100.0;

	dist1 = UTIL_Distance(v_vec, v_end1);
	dist2 = UTIL_Distance(v_vec, v_end2);

	if (dist1 > dist2)
		vecDir |= dir_forward; // vector is at the right side
	else if (dist2 > dist1)
		vecDir |= dir_back; // vector is at the left side

	return vecDir;
}
/*
// Determine at which side an vector is
int UTIL_VecDir(Vector v_src, Vector v_aim, Vector v_vec)
{
	Vector v_src = pEdict->v.origin;
	int vecDir = 0;

	Vector v_end1 = g_vecZero;
	Vector v_end2 = g_vecZero;
	float dist1 = 0.0;
	float dist2 = 0.0;

	UTIL_MakeVectors(v_aim);

	// Is the vector at the left side or at the right side ?
	v_end1 = v_src - gpGlobals->v_right * 100.0;
	v_end2 = v_src + gpGlobals->v_right * 100.0;

	dist1 = UTIL_Distance(v_vec, v_end1);
	dist2 = UTIL_Distance(v_vec, v_end2);

	if (dist1 > dist2)
		vecDir |= dir_right; // vector is at the right side
	else if (dist2 > dist1)
		vecDir |= dir_left; // vector is at the left side

	// Is the vector lower or higher ?
	if (v_src.z > v_vec.z)
		vecDir |= dir_down; // vector is lower
	else if (v_src.z < v_vec.z)
		vecDir |= dir_up; // vector is higher

	// Is the vector behind or infront ?
	v_end1 = v_src - gpGlobals->v_forward * 100.0;
	v_end2 = v_src + gpGlobals->v_forward * 100.0;

	dist1 = UTIL_Distance(v_vec, v_end1);
	dist2 = UTIL_Distance(v_vec, v_end2);

	if (dist1 > dist2)
		vecDir |= dir_forward; // vector is at the right side
	else if (dist2 > dist1)
		vecDir |= dir_back; // vector is at the left side

	return vecDir;
}
*/
// Simulates damaging a player by a player with a specified weapon name
void UTIL_DamagePlayer(edict_t *pKiller, edict_t *pVictim, const char *Weapon, float Damage)
{
	static CBaseEntity *pDamageEnt = NULL;

	CBaseEntity *pEnt = (CBaseEntity *)GET_PRIVATE(pVictim);

	if (!pDamageEnt)
		pDamageEnt = UTIL_CreateEnt("info_target",Vector(0,0,0),Vector(0,0,0),	NULL);

	if ((pDamageEnt) && (pEnt))
	{
		UTIL_MakeVectors(WrapAngles(pVictim->v.angles+Vector(180,180,180)));
		pDamageEnt->pev->velocity = pDamageEnt->pev->velocity.Normalize() * 180 + gpGlobals->v_forward * 10;
		pDamageEnt->pev->origin = pKiller->v.origin;
		pDamageEnt->pev->owner = pKiller;
		pDamageEnt->pev->classname = MAKE_STRING(Weapon);
		pEnt->TakeDamage( pDamageEnt->pev, &pKiller->v,Damage,0);
	}
}

// Function code by cheesy
CBaseEntity *UTIL_CreateEnt( char *szName, const Vector &vecOrigin, const Vector &vecAngles, edict_t *pentOwner ) 
{ 
      edict_t *pent; 
      CBaseEntity *pEntity; 

      pent = CREATE_NAMED_ENTITY( MAKE_STRING( szName )); 

      if ( FNullEnt( pent ) ) 
      {
#ifdef _DEBUG
            ALERT ( at_console, "NULL Ent in Create!\n" ); 
#endif
            return NULL; 
      }
 
      if ( !pent ) 
            pent = ENT(0); 

      pEntity = (CBaseEntity *)GET_PRIVATE(pent);  

      pEntity->pev->owner = pentOwner; 
      pEntity->pev->origin = vecOrigin; 
      pEntity->pev->angles = vecAngles; 

      DispatchSpawn( pEntity->edict() ); 

      return pEntity; 
}

int UTIL_LookUpSequence(void *pModel, const char *Label)
{
	studiohdr_t *pstudiohdr;
	
	pstudiohdr = (studiohdr_t *)pModel;
	if (! pstudiohdr)
		return 0;

	mstudioseqdesc_t	*pseqdesc;

	pseqdesc = (mstudioseqdesc_t *)((byte *)pstudiohdr + pstudiohdr->seqindex);

	for (int i = 0; i < pstudiohdr->numseq; i++)
	{
		if (stricmp( pseqdesc[i].label, Label ) == 0)
			return i;
	}

	return -1;
}

void UTIL_Printf(const char *msg, int type, edict_t *pEntity)
{
	if (IsNull(pEntity))
		pEntity = listenserver_edict;

	if (IsNull(pEntity))
	{
		(*g_engfuncs.pfnServerPrint) (const_cast<char*>(msg));
	}
	else
		ClientPrint(pEntity, type, msg);
}

void UTIL_KickBots(short team)
{
	int index;
	if ((team == -1) || (IsTeamPlay() == false))
	{
		// Kick all bots
		for (index = 0; index < 32; index++)
		{
			// is this slot used?
			if ((bots[index].is_used) && (bots[index].pEdict != listenserver_edict))
			{
				char cmd[40];
				 		  
				sprintf(cmd, "kick \"%s\"\n", STRING(bots[index].pEdict->v.netname));
						  
				SERVER_COMMAND(cmd);  // kick the bot using (kick "name")
				num_bots_check = -2;
			}
		}
	}
	else
	{   
		// Kick all bots in specified team
		for (index = 0; index < 32; index++)
		{
			// is this slot used?
			if ((bots[index].is_used) && (IsNull(bots[index].pEdict) == false) &&
				(UTIL_GetTeam(bots[index].pEdict) == (team-1)) && 
				(bots[index].pEdict != listenserver_edict)) 
			{
				char cmd[40];
				
				sprintf(cmd, "kick \"%s\"\n", STRING(bots[index].pEdict->v.netname));
						  
				SERVER_COMMAND(cmd);  // kick the bot using (kick "name")
				num_bots_check = -2;
			}
		}
 	}

	if (fill_server) // stop filling the server(guess the user want to abort it)
	{
		fill_team = 0;
		fill_half = fill_server = false;
	}
}

void UTIL_KickPlayer(edict_t *pEdict)
{
	if (IsNull(pEdict))
		return;

	char cmd[40];
	sprintf(cmd, "kick \"%s\"\n", STRING(pEdict->v.netname));
	SERVER_COMMAND(cmd);  // kick the player using (kick "name")
}

#define chunk_size 100
#define max_size 1536

extern bool PrintingInGameHelp;

// MOTD code ripped from the sdk...
void UTIL_ShowInGameHelpWindow()
{
	static bool Started = false;
	static int length = 0;
	static char_count = 0;
	static char *pFileList;
	static char *aFileList = pFileList = NULL;
	static char OldServerName[256] = "";
	static char chunk[chunk_size+1];

	if (gmsgMOTD == 0)
		gmsgMOTD = REG_USER_MSG("MOTD", -1);

	if (gmsgServerName == 0)
		gmsgServerName = REG_USER_MSG("ServerName", -1);

	if (Started == false)
	{
		PrintingInGameHelp = true;
		length = char_count = 0;
		aFileList = pFileList = (char*)LOAD_FILE_FOR_ME("ESF_Bot\\ingame_help.txt" , &length );
		Started = true;
		strcpy(OldServerName, CVAR_GET_STRING("hostname"));

		if (IS_DEDICATED_SERVER())
			UTIL_Printf("In game bot help\n\n");
		else
		{
			if (IsPlayingMessage)
			{
				clog << "Tried to change server name when there is already a message active\n";
				return;
			}

			// send the server name
			MESSAGE_BEGIN( MSG_ONE, gmsgServerName, NULL, listenserver_edict );
				WRITE_STRING( "In game bot help" );
			MESSAGE_END();
		}
	}

	// Send the message of the day
	// read it chunk-by-chunk,  and send it in parts

	if ( pFileList && *pFileList && char_count < max_size )
	{	
		if ( strlen( pFileList ) < chunk_size )
		{
			strcpy( chunk, pFileList );
		}
		else
		{
			strncpy( chunk, pFileList, chunk_size );
			chunk[chunk_size] = 0;		// strncpy doesn't always append the null terminator
		}

		char_count += strlen( chunk );
		if ( char_count < max_size )
			pFileList = aFileList + char_count; 
		else
			*pFileList = 0;

		if (IS_DEDICATED_SERVER())
			UTIL_Printf(chunk);
		else
		{
			if (IsPlayingMessage)
			{
				clog << "Tried to send ingame help data when there is already a message active\n";
				return;
			}

			MESSAGE_BEGIN( MSG_ONE, gmsgMOTD, NULL, listenserver_edict );
				WRITE_BYTE( *pFileList ? FALSE : TRUE );	// FALSE means there is still more message to come
				WRITE_STRING( chunk );
			MESSAGE_END();
		}
	}

	if ((!pFileList) || (*pFileList == NULL) || (char_count > max_size))
	{
		FREE_FILE( aFileList );

		if (IS_DEDICATED_SERVER() == false)
		{
			if (IsPlayingMessage)
			{
				clog << "Tried to change server name when there is already a message active\n";
				return;
			}

			// Restore server name(I think...)
			MESSAGE_BEGIN( MSG_ONE, gmsgServerName, NULL, listenserver_edict );
				WRITE_STRING( OldServerName );
			MESSAGE_END();
		}

		char_count = length = 0;
		Started = PrintingInGameHelp = false;
		*chunk = NULL;
	}
}

bool IsTeamPlay()
{
	// Always team play when someone chose a team :-)
	if (TeamMenuShowedUp)
		return true;

	if (ClassMenuShowedUp) // Only class menu showed up, so no team play
		return false;

	// See what the game mode is(0 = freefight; 1=teamplay; 2=cptdb mode)
	if (CVAR_GET_FLOAT("mp_gamemode") > 0.0)
		return true;

	return false; // Guess its free fight
}

short UTIL_GetHeadingPercentage(edict_t *pEnt1, edict_t *pEnt2)
{
	if ((!pEnt1) || (!pEnt2))
		return 0;

	if (pEnt2->v.velocity.Length() == 0.0f)
		return 0;

	Vector v_predict = PredictEntityPos(pEnt2, 0.01f, false);

	float DistEnt1ToEnt2 = UTIL_Distance(pEnt1->v.origin, pEnt2->v.origin);
	float PredictDist = UTIL_Distance(pEnt1->v.origin, v_predict);

	if (PredictDist > DistEnt1ToEnt2)// && (DistEnt1ToEnt2 > pEnt2->v.velocity.Length()))
		return 0;

	float MaxDistCloser = pEnt2->v.velocity.Length() * 0.01f;
	float RealDistCloser = DistEnt1ToEnt2 - PredictDist;

	return ((RealDistCloser / MaxDistCloser) * 100.0f);
}

byte UTIL_GetNumPlayers()
{
	byte NumPlayers = 0;
	byte b;

	for (b=1;b<=gpGlobals->maxClients;b++)
	{
		if (UTIL_IsValidPlayer(INDEXENT(b), false, false))
			NumPlayers++;
	}

	return NumPlayers;
}

#ifdef useclassshit
bool UTIL_IsBot(CClient *pClient)
{
	if (!pClient)
		return false;

	return (pClient->IsBot());
}
#endif
